/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Compatibility for dirent.h */

/*
 opendir
 closedir
 readdir
*/



#ifndef __CORAL_COMPAT_DIRENT_H
# define __CORAL_COMPAT_DIRENT_H



#include <coral/def.h>



/* dirent.h supported compilers */

#if defined(__UNIXC__) || ( defined(__TURBOC__) && __TURBOC__>=0x200 )

# include <dirent.h>

#else /* dirent.h unsupported compilers */



#ifdef __WATCOMC__

/* Watcom C bug fix, incompatible action of the function opendir */
#include <direct.h>
#include <dos.h>

#else /* !__WATCOMC__ */



#if defined(__WIN32__)

# define _find_t_name(find) (find).cFileName

#ifndef WIN32_DIR_DEFINED

struct dirent {
  char __F* d_name; /* directory name reference */
};

typedef struct {
  int reserved;
} DIR;

#endif

#else /* __WIN32__ */

#include <dos.h>
#ifdef __TURBOC__
# include <dir.h>
#else
# define _find_t_attrib(find)  (find).attrib
# define _find_t_wr_time(find) (find).wr_time
# define _find_t_wr_data(find) (find).wr_date
# define _find_t_size(find)    (find).size
# define _find_t_name(find)    (find).name
#endif

typedef struct
{
 char __F*      path;
 struct dirent  dirent; /* posix dirent object */
 struct _find_t find;   /* dos _find_t object */
 char           handle; /* indication flag to represent findfirst was processed */
} DIR;

#endif /* __WIN32__ */



_API( void, compat_closedir, ( DIR __F* dir ) );
_API( struct dirent __F*, compat_readdir, ( DIR __F* dir ) );

#define closedir compat_closedir
#define readdir compat_readdir



#endif /* __WATCOMC__ */


_API( DIR __F*, compat_opendir, ( const char __F* name ) );
#define opendir compat_opendir



#endif  /* dirent.h supported compilers */



#endif /* __CORAL_COMPAT_DIRENT_H */
