/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* stdio.h compatibility */


#ifndef __CORAL_COMPAT_STDIO_H
# define __CORAL_COMPAT_STDIO_H



#include <stdarg.h>
#include <coral/def.h>



#ifdef __PALMOS__

#include <StdIOPalm.h>

#else /* !__PALMOS__ */

#include <stdio.h>

#endif /* __PALMOS__ */



#ifdef _MSC_VER
#	define FMT_LL "I64"
#else
#	define FMT_LL "ll"
#endif



/* _STDIOLOCKBUFFER */

struct _STDIOLOCKBUFFER {
	FILE __F* in, __F* out;
	void ( __API *prompt)( struct _STDIOLOCKBUFFER __F* stdiolock );
	void __F* context;
	int state;
	int bufsize, bufpos, buflen;
	_integer1 buf[1];
};



/* _STDIOLOCK, state */
/* IO_END */
#define STDIO_BREAK -1
#define STDIO_NORMAL 1



__CBEGIN__



#ifdef _MSC_VER
/* stdio.h functions which are not fully supported in msvc */
_CAPI( int, vsnprintf, ( char __F* buf, int len, const char __F* fmt, va_list arg ) );
_CAPI( int, snprintf, ( char __F* buf, int len, const char __F* fmt, ... ) );
_CAPI( int, vasprintf, ( char __F*__F* buf, const char __F* fmt, va_list arg ) );
_CAPI( int, asprintf, ( char __F*__F* buf, const char __F* fmt, ... ) );
#endif



/* some extensions */

_CAPI( int, vbnprintf, ( char __F* buf, int len, int __F* bufpos, const char __F* fmt, va_list arg ) );
_CAPI( int, bnprintf, ( char __F* buf, int len, int __F* bufpos, const char __F* fmt, ... ) );
_CAPI( int, vabprintf, ( char __F*__F* buf, int __F* bufpos, const char __F* fmt, va_list arg ) );
_CAPI( int, abprintf, ( char __F*__F* buf, int* bufpos, const char* fmt, ... ) );



_API( _integer1 __F*, stdioLockBuffer, ( struct _STDIOLOCKBUFFER __F* stdiolock, int __F* locked, int size, unsigned flag ) );
_API( void, stdioUnlockBuffer, ( struct _STDIOLOCKBUFFER __F* stdiolock, _integer1 __F* current ) ); 
_API( int, stdioOpen, ( struct _IO __F* io, const char __F* filename, unsigned flag ) );
_API( int, stdioClose, ( struct _IO __F* io ) );
_API( long, stdioRead, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, stdioWrite, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, stdioSeek, ( struct _IO __F* io, long offset, long __F* offset_high, int seekpos ) );



__CEND__



#endif /* __CORAL_COMPAT_STDIO_H */

