/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Coral library Definitions

 < Platform Specific Macros >

 __UNIX__         , unix
 __BEOS__         , beos - untested yet
 __OS2__          , os2 - untested yet
 __MACOS__        , macos
 __PALMOS__       , palmos
 __WINDOWS__      , windows
 __WIN32__        , 32 bit windows
 __WIN16__        , 16 bit windows
 __DOS__          , dos
 __DOS16__        , 16 bit dos
 __DOS32__        , 32 bit dos
 __16__           , 16 bit architecture
 __ARCH__ , default pointer size of a platform
 __INTEGERX__     , compiler supports extended sized integer
 __UNIXC__        , a c compiler which follows unix rules

 __BIG_ENDIAN__, for big endian machines
 __LITTLE_ENDIAN__, for little endian machines
 __PDP_ENDIAN__, for pdp endian machines

  Some other platform specific macro names contained in this source
 can be variable.

 < Code Generation Switches >

 __STRICT__      , use strict type, this option will affect windows.h
 __EXPORT__      , export functions without def file if possible
 __L77__         , settings are switched to use L77
 __MODERNPARAM__ , use modern c or c++ parameter definition rules
 __CLASSICPARAM__, use classic c parameter definition rules */



#ifndef __CORAL_DEF_H
#	define __CORAL_DEF_H



/* version 1.0.0 */

#define __CORALVERSION__ 0x00010000



/* C section */

#ifdef __cplusplus
#	define __CBEGIN__ extern "C" {
#	define __CEND__ }
#	ifndef __STRICT__
#		define __STRICT__
#	endif
#else
#	define __CBEGIN__
#	define __CEND__
#endif /* __cplusplus */



/*-------------------------------------------------------------------------*
   Platform Specific Macros
 *-------------------------------------------------------------------------*/



/* UNIX */

#if ( defined(__unix) || defined(__unix__) ) && \
	!defined(DJGPP) && !defined(__CYGWIN__)

#	ifndef __UNIX__
#		define __UNIX__
#	endif
#	define FLAG_PATH_NAME 0

/* End of UNIX */



#else /* !__UNIX__ */



/* MACOS */

#ifdef macintosh

#	define __MACOS__

/* End of MACOS */



#else /* !MAC */



/* PalmOS */

#ifdef __palmos__

#	define __PALMOS__

/* End of PalmOS */



#else /* !__palmos__ */



/* WINDOWS */

#if defined(_Windows) && !defined(__WINDOWS__)
#	define __WINDOWS__
#endif /* _Windows */


#if defined(WIN32) || defined(_WIN32) || defined(__CYGWIN__)
#	ifndef __WINDOWS__
#		define __WINDOWS__
#	endif
  /* borland c automatically defines __WIN32__ */
#	ifndef __WIN32__
#		define __WIN32__
#	endif
#endif /* WIN32 || _WIN32 */



#ifdef __WINDOWS__

#	if defined(__STRICT__) && !defined(STRICT)
#		define STRICT
#	endif

/* @cmt0001 */
#	if defined (__EXPORT__)
#		if defined(__TURBOC__)
#			define EXPORT_KEYWORD __export
#			define EXPORT_AFTER_RETURN
#		else
#			define EXPORT_KEYWORD __declspec(dllexport)
#			define EXPORT_BEFORE_RETURN
#		endif
#	endif /* __EXPORT__ */

#	ifdef __WIN32__

		/* WINAPI is standard calling convention */
#		ifdef __GNUC__
#			define __API __attribute__((stdcall))
			/* maybe it was a bug in old version ( the exact version is
				unclear ) cygwin gcc that __attribute__((stdcall)) cannot be
				used for the function pointer casting */
#			if __GNUC__<0 && __GNUC_MINOR__<0 && __GNUC_PATCHLEVEL__<0
#				define __APICAST
#			endif

#		else
#			define __API _stdcall
#		endif /* __GNUC__ */

#	else /* !__WIN32__ */

#		define __16__
#		define __WIN16__
#		define __API __F pascal

#	endif /* __WIN32__ */

/* End of WINDOWS */



#else /* !__WINDOWS__ */



/* DOS */

#if defined(MSDOS) || defined(__MSDOS__) || defined(__FLAT__)

#	ifndef __DOS__
#		define __DOS__
#	endif

/* detect dos16 or dos32 */
#	ifdef __FLAT__
#		define __DOS32__
#	else /* !__FLAT__ */
#	ifdef _MSC_VER
#		if !defined(M_I86TM) && !defined(M_I86SM) && !defined(M_I86CM) && \
		   !defined(M_I86MM) && !defined(M_I86LM) && !defined(M_I86HM)
#			define __DOS32__
#		endif
#	else /* !_MSC_VER */

#		if !defined(__TINY__) && !defined(__SMALL__) && !defined(__MEDIUM__) && \
		   !defined(__LARGE__) && !defined(__HUGE__) && !defined(__COMPACT__)
#			define __DOS32__
#		endif

#	endif /* _MSC_VER */
#	endif /* __FLAT__ */

#	ifndef __DOS32__
#		define __16__
#		define __DOS16__
#	endif

#	define __API __F pascal

/* End of DOS */



/* Other platforms */

#else /* !( defined(MSDOS) || defined(__MSDOS__) || defined(__FLAT__) ) */

... /* error */

#endif /* defined(MSDOS) || defined(__MSDOS__) || defined(__FLAT__) */

/* End of other platforms */



#endif /* __WINDOWS__ */

#endif /* __palmos__ */

#endif /* MAC */

#endif /* ( __unix || __unix__ ) && !DJGPP && !__CYGWIN__ */



/* to write 'internal' interface for L77, define L77_INTERFACE previously
   : 32 bit Borland C/C++ is needed to support Lahey F77 */
#if defined(L77_INTERFACE) || defined(__L77__)

#	if defined(__TURBOC__) && defined(__WIN32__)

 /* override platform specifiers */
#		undef __WINDOWS__
#		undef __WIN32__
#		undef __EXPORT__
#		undef __API
#		define __DOS__
#		define __DOS32__
#		define __API
#		ifdef L77_INTERFACE
#			define API_NAME(name) _##name
#		endif

#	else /* !__TURBOC__ || !__WIN32__ */

... /* error */

#	endif /* __TURBOC__ && __WIN32__ */

#else /* !L77_INTERFACE && !__L77__ */

#ifdef _EiC
#	undef __EXPORT__
#	undef __API
#endif

#endif /* L77_INTERFACE || __L77__ */



/* byte order definitions */
#if defined(mc68000) || defined(__mc68000__) || defined (__sparc__) || defined (__PPC__) \
 || (defined(__mips__) && (defined(MIPSEB) || defined (__MIPSEB__)))
#	ifndef __BIG_ENDIAN__
#		define __BIG_ENDIAN__
#	endif
#	define __ENDIAN__ 4321
#else /* !big endian platforms */
#if defined(PDP_ENDIAN_PLATFORM)
#	ifndef __PDP_ENDIAN__
#		define __PDP_ENDIAN__
#	endif
#	define __ENDIAN__ 3412
#else /* !pdp endian platforms */
#	ifndef __LITTLE_ENDIAN__
#		define __LITTL_EENDIAN__
#	endif
#	define __ENDIAN__ 1234
#endif /* pdp endian platforms */
#endif /* big endian platforms */



/*------------------------------------------------------------------------*
   Definitions for functions
 *------------------------------------------------------------------------*/



/* __UNIXC__ for unix-like c compilers */
#if defined(__UNIX__) || defined(__GNUC__)
#	define __UNIXC__
#endif



#if !defined(__STDC__) && defined(__UNIXC__)

/* some symbols K&R C doesn't recognize */
#	define signed
#	define const

/* classic parameter is default if modern parameter was not forced */
#	ifndef __MODERNPARAM__
#		define __CLASSICPARAM__
#	endif

#else /* defined(__STDC__) || !defined(__UNIXC__) */

/* modern parameter is default if classic parameter was not forced */
#	ifndef __CLASSICPARAM__
#		define __MODERNPARAM__
#	endif

#endif /* !defined(__STDC__) && defined(__UNIXC__) */


#ifndef __API
#	define __API
#endif

#ifndef __APICAST
#	define __APICAST __API
#endif



/* Mixed usage of modern function prototype and classic function definition
 is restricted because automatic type promotion rule is quite complex
 problem to be clearly solved by some macro definitions.
 PARAM_BEGIN, PARAM_1ST, PARAM_AND, PARAM_END, PARAM_VOID or
 P_B, P_1, P_N, P_E, P_V are available */

#if defined(__MODERNPARAM__) || defined(__cplusplus)
#	define P_A(param) param
#	define P_B(param) (
#	define P_E )
#	define P_X ,...)
#	define P_N ,
#	define P_V ( void )
#else
#	define P_A(param) ()
#	define P_B(param) param
#	define P_E ;
#	define P_X ; va_dcl
#	define P_N ;
#	define P_V ()
#endif /* defined(__MODERNPARAM__) || defined(__cplusplus) */

#define P_1
#define PARAM_API   P_A
#define PARAM_BEGIN P_B
#define PARAM_END   P_E
#define PARAM_VAR   P_X
#define PARAM_AND   P_N
#define PARAM_VOID  P_V
#define PARAM_1ST   P_1



/* export specification */
#ifdef __EXPORT__
#	ifdef EXPORT_BEFORE_RETURN
#		define API_RETURN(ret) EXPORT_KEYWORD ret
#		define API_NORETURN EXPORT_KEYWORD
#	else
#	ifdef EXPORT_AFTER_RETURN
#		define API_RETURN(ret) ret EXPORT_KEYWORD
#		define API_NORETURN EXPORT_KEYWORD
#	endif /* EXPORT_AFTER_RETURN */
#	endif /* EXPORT_BEFORE_RETURN */
#endif /* __EXPORT__ */

#ifndef API_RETURN
#	define API_RETURN(ret) ret
#	define API_NORETURN
#endif /* __EXPORT__ */

#ifndef API_NAME
#	define API_NAME(name) name
#endif

#define _CAPI(ret,name,param) API_RETURN(ret) API_NAME(name) P_A(param)
#define _CAPI_EXT(ret,prefix,postfix,name,param) API_RETURN(ret) ( prefix ( API_NAME(name) P_A(param) ) ) postfix
#define _CAPI_NORETURN(name,param) API_NORETURN API_NAME(name) P_A(param)
#define _API(ret,name,param) _CAPI(ret __API,name,param)
#define _API_EXT(ret,prefix,postfix,name,param) _CAPI_EXT(ret,prefix __API,postfix,name,param)
#define _API_NORETURN(name,param) API_NORETURN __API API_NAME(name) P_A(param)



/*------------------------------------------------------------------------*
   Generic Definitions
 *------------------------------------------------------------------------*/



/* Module specifiers */

#ifdef __PREPROCESS__
#	define __MODULEBEGIN__(name) __MODULEBEGIN name
#	define __MODULEEND__         __MODULEEND
#else /* !__PREPROCESS__ */
#	define __MODULEBEGIN__(name)
#	define __MODULEEND__
#endif /* __PREPROCESS__ */



/* Generic context templates */

#define DECLARE_CONTEXT(name) typedef struct { int reserved; } __F* name;
#define DECLARE_END }



/* Coral linked list templates */

#define DECLARE_STACK       DECLARE_LIST
#define DECLARE_LIST(name)  struct name { struct name __F* next;
#define DECLARE_DLIST(name) struct name { struct name __F* next, __F* prev;



/* Coral binary tree templates */

#define DECLARE_BINTREE(name) struct name { struct name __F* left, __F* right;
#ifdef __16__
#	define DECLARE_BALANCED_BINTREE(name) DECLARE_BINTREE(name) _integer1 bf;
#else
#	define DECLARE_BALANCED_BINTREE(name) DECLARE_BINTREE(name) int bf; /* not to remain some padded bytes */
#endif

/* Generic constants */

#define SUCCESS_EXIST     1     /* an object is already present */
#define SUCCESS_NORMAL    0     /* normal result */
#define FAIL_IO          -1     /* io operation error */
#define FAIL_ENTRY       -2     /* file or directory entry node was not found */
#define FAIL_MEMORY      -3     /* memory allocation error */
#define FAIL_BUFFER      -4     /* the given buffer is full and no space to store data */
#define FAIL_PARAM       -5     /* invalid parameter is present */
#define FAIL_UNAVAILABLE -6     /* the resource is unavailable */
#define FAIL_DUPLICATED  -7     /* the resource is duplicated ( the resource should be unique ) */
#define FAIL_UNKNOWN     -8     /* the reason was unknown */
#define FAIL_PARSE       -9     /* parse error */
#define FAIL_RESERVED    -1024  /* reserved error codes */


#define MASK_COMBINATED_SIZE  0x00ff  /* to get combinated size */
#define MASK_BYTE_ORDER       0x3000  /* to get byte ordering flags */
#define FLAG_LITTLE_ENDIAN    0x1000
#define FLAG_BIG_ENDIAN       0x2000
#define FLAG_PDP_ENDIAN       0x3000
#define FLAG_NETWORK          FLAG_BIGENDIAN
#define FLAG_TEMPORARY        0x0100
#define FLAG_DEFAULT          0x0000
#define FLAG_CLEAR            0x0001

#define FLAG_ROW_VARIABLE     0x4000
#define FLAG_COL_VARIABLE     0x8000

#define FLAG_STRING_POINTER     0x1000  /* string pointer in a class */
#define FLAG_CASE_INSENSITIVE   0x2000  /* case insensitive string comparison */

#ifndef FLAG_PATH_NAME
#	define FLAG_PATH_NAME FLAG_CASE_INSENSITIVE
#endif



#ifndef NULL
#   ifdef __cplusplus
#       define NULL 0
#   else
#       ifdef __16__
#           define NULL ((void __F*)0)
#       else
#           define NULL ((void*)0)
#       endif
#   endif
#endif



/* Integer value limits */

#define OP_MAXBITINT_EXT(_1,cnt)  (-(OP_MINBITINT_EXT(_1,cnt)+1))
#define OP_MINBITINT_EXT(_1,cnt)  ((-(_1))<<((cnt)-1))
#define OP_MAXBITINT(cnt)  OP_MAXBITINT_EXT(1L,cnt)
#define OP_MINBITINT(cnt)  OP_MINBITINT_EXT(1L,cnt)

#define OP_MAXBITUINT_EXT(_1u,cnt) (((_1u)<<(cnt))-1)
#define OP_MAXBITUINT(cnt) OP_MAXBITUINT_EXT(1UL,cnt)



/* Explicit size integers */

typedef signed char    _integer1;
typedef unsigned char  _unsigned1;
typedef short          _integer2;
typedef unsigned short _unsigned2;

#ifdef __16__ /* __16__ is the special symbol only for 16 bit platform */

#	define __F far
#	define __H huge
	typedef long _integer4;
	typedef unsigned long _unsigned4;
#	define __ARCH__ 16

#else /* !__16__ */

#	define __F
#	define __H

#	ifdef __TURBOC__

#		define __ARCH__ (sizeof(long)*8) /* borland c allows to use sizeof in preprocessor */

#	else /* !__TURBOC__ */

#		include <limits.h>
#		if INT_MAX==OP_MAXBITINT(32)
#			define __ARCH__ 32
			typedef long _integer4;
			typedef unsigned long _unsigned4;
#		else
#			if LONG_MAX==OP_MAXBITINT(64)
#				define __ARCH__ 64
				typedef int _integer4;
				typedef unsigned int _unsigned4;
				typedef long _integer8;
				typedef unsigned long _unsigned8;
#			define __INTEGER8__
#		else
			... /* #error Unsupported architecture. */
#		endif
#	endif

#	endif /* __TURBOC__ */

#endif /* __16__ */


#ifdef __GNUC__

	typedef long long __integerx;
	typedef unsigned long long __unsignedx;
#	define __INTEGERX__

#else /* !__GNUC__ */
#ifdef _MSC_VER
#	if __ARCH__==32
		typedef __int64 __integerx;
		typedef unsigned __int64 __unsignedx;
#		define __INTEGERX__
#	endif
#endif /* _MSC_VER */
#endif /* __GNUC__ */



#ifndef __INTEGER8__
#	ifdef __INTEGERX__
	typedef __integerx _integer8;
	typedef __unsignedx _unsigned8;
#	define __INTEGER8__
#else
	typedef struct { _integer4 low, high; } _integer8;
	typedef struct { _unsigned4 low, high; } _unsigned8;
#endif
#endif



/* Other types */

typedef unsigned short _wchar;
typedef int _bool;
typedef _integer4 _fix;
DECLARE_CONTEXT(ANYCONTEXT);



/* Coral I/O */

struct _IO {
	long (__API* read) P_A(( struct _IO __F* io, void __F* p, long size ));   /* read */
	long (__API* write) P_A(( struct _IO __F* io, void __F* p, long size ));  /* write */
	long (__API* seek) P_A(( struct _IO __F* io, long offset, long __F* offset_high, int seekpos )); /* seek */
	void __F* handle; /* io handle */
	struct _io_buffer __F* buffer; /* coral specific intermediate buffer */
	unsigned  op;     /* the current operation */
};



/* _DIOHANDLE */

typedef _integer8 _DIOHANDLE;



/* Coral dynamic I/O */

struct _DIO {
	struct _IO __F* io;
	_bool (__API* seteof) P_A(( struct _IO __F* io )); /* set eof */
	long      offset;    /* initial offset */
	long      align;     /* size to be aligned ( including internal record ) : it should be power of 2 */
	long      minsize;   /* minimum size */
	unsigned  byteorder; /* byte order flag */
};



/* Coral I/O buffer */
typedef _integer1 __F* (__API* _IOLOCKBUFFER) P_A(( ANYCONTEXT context, int __F* locked, int size, unsigned flag ));
typedef void (__API* _IOUNLOCKBUFFER) P_A(( ANYCONTEXT context, _integer1 __F* current ));

struct _IOBUFFER {
	ANYCONTEXT context;     /* io buffer context */
	_IOLOCKBUFFER lock;     /* function pointer to lock buffer */
	_IOUNLOCKBUFFER unlock; /* function pointer to unlock buffer */
};



#endif /* __CORAL_DEF_H */

