/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Coral Dynamic I/O */



#ifndef __CORAL_DIO_H
#	define __CORAL_DIO_H



#include <coral/def.h>



#define DIO_ALIGN(size,align) \
	(OP_ALIGN(size+sizeof(_DIOHANDLE[2]),align)-sizeof(_DIOHANDLE[2]))

#define DIO_MIN_SIZE 32


typedef _integer4 _diosize;



__CBEGIN__



_API( int, dioInitFreeLink, ( struct _DIO __F* dio ) );
_API( int, dioAlloc, ( struct _DIO __F* dio, _DIOHANDLE __F* handle, long size ) );
_API( int, dioFree, ( struct _DIO __F* dio, const _DIOHANDLE __F* handle ) );
_API( int, dioSeek, ( struct _DIO __F* dio, const _DIOHANDLE __F* handle ) );
_API( int, dioRead, ( struct _DIO __F* dio, const _DIOHANDLE __F* handle, void __F* p, long n, unsigned flag ) );
_API( int, dioWrite, ( struct _DIO __F* dio, const _DIOHANDLE __F* handle, void __F* p, long n, unsigned flag ) );
_API( int, dioWalk, ( struct _DIO __F* dio, long __F* p_size, long used ) );
_API( int, dioWalkFree, ( struct _DIO __F* dio, int (__API* walk )( long size, void __F* param ), void __F* param ) );

_API( int, dioListPush, ( struct _DIO __F* dio, _DIOHANDLE __F* head, unsigned long size ) );
_API( void, dioListPop, ( struct _DIO __F* dio, _DIOHANDLE __F* head ) );
_API( _DIOHANDLE, dioListCreate, ( struct _DIO __F* dio, _DIOHANDLE __F* head, _DIOHANDLE where, unsigned long size ) );
_API( _DIOHANDLE, dioListRemove, ( struct _DIO __F* dio, _DIOHANDLE __F* head, _DIOHANDLE what, _DIOHANDLE prev ) );
_API( _DIOHANDLE, dioListClear, ( struct _DIO __F* dio, _DIOHANDLE __F* head, void (__API* free)( struct _DIO __F* dio, _DIOHANDLE handle ) ) );

_API( int, dioListDNAdd, ( struct _DIO __F* dio, _DIOHANDLE __F* head, const _DIOHANDLE __F* where, const _DIOHANDLE __F* what, unsigned flag ) );
_API( int, dioListDNCut, ( struct _DIO __F* dio, _DIOHANDLE __F* head, const _DIOHANDLE __F* what ) );
_API( int, dioListDNCreate, ( struct _DIO __F* dio, _DIOHANDLE __F* handle, _DIOHANDLE __F* head, const _DIOHANDLE __F* where, long size, unsigned flag ) );
_API( int, dioListDNRemove, ( struct _DIO __F* dio, _DIOHANDLE __F* head, const _DIOHANDLE __F* what ) );



__CEND__



#endif /* __CORAL_DIO_H */

