/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Compatibility for the dynamic loader of a shared library */
/*
 DL_UNSUPPORTED is defined if dynamic loading is not supported.
*/



#ifndef __CORAL_DL_H
# define __CORAL_DL_H

#include <coral/io.h>



#if defined(__WIN32__) && defined(__EXPORT__)
#	define DL_API_SYMBOL(name,paramsize) "_" name "@" paramsize
#else
#	define DL_API_SYMBOL(name,paramsize) name
#endif



/* dlOpen, flags */

#define DL_FLAG_IMMEDIATE  0x0001  /* immediate binding */
#define DL_FLAG_PURE_NAME  0x0100  /* the given name is purely defined and modification is not needed */



/* dlGetSymbolTable */
#define DL_INVALID_SYMBOL (FAIL_RESERVED-1) /* an invalid symbol is found */
#define DL_INVALID_TABLE  (FAIL_RESERVED-2) /* invalid table was constructed by mismatched entry counter */



/* dlLoadSymbolTable */
/* DL_INVALID_SYMBOL, DL_INVALID_TABLE */
#define DL_INVALID_LIBRARY (FAIL_RESERVED-3) /* invalid shared library was supplied */



__CBEGIN__



_API( void __F*, dlOpen, ( const char __F* name, unsigned flag ) );
_API( void, dlClose, ( void __F* handle ) );
_API( void __F*, dlSymbol, ( void __F* handle, const char __F* name ) );

_API( _bool, dlInitialize, ( void ) );
_API( void, dlFinalize, ( void ) );
_API( void, dlBegin, ( void ) );
_API( void, dlEnd, ( void ) );


__CEND__



#endif /* __CORAL_DL_H */

