/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Redefinition of platform specific thread API
   Cooperative thread implementation

_THREAD_INLINE_ to use inline thread macro

threadFork(proc,data)
threadExit(retval)
threadYield()

for platform free thread implementation, use following API

define local variables of the current thread procedure :
DECLARE_THREAD_LOCAL
DECLARE_THREAD_END

use a local variable :
_(name)

to give a chance another thread can be executed
threadCooperate */



#ifndef __CORAL_EXT_PROCESS_H
# define __CORAL_EXT_PROCESS_H



#include <coral/def.h>



/* thread flags */
#define THREAD_FLAG_WAIT      0x0001  /* a thread which can be waited or threadDetach need to be called to clear resources assigned for a thread */
#define MUTEX_FLAG_TRYLOCK    0x0001  /* support TryLock */
#define MUTEX_FLAG_RECURSIVE  0x0002  /* support recursive lock */



/* thread priority */
#define THREAD_IDLE    0
#define THREAD_LOWEST  1
#define THREAD_LOW     2
#define THREAD_NORMAL  3
#define THREAD_HIGH    4
#define THREAD_HIGHEST 5
#define THREAD_URGENT  6



#ifndef __INTERNAL__
	DECLARE_CONTEXT(PROCLISTCONTEXT);
	DECLARE_CONTEXT(THREADCONTEXT);
	DECLARE_CONTEXT(THREADCOND);
	DECLARE_CONTEXT(PROCESSMUTEX);
#ifndef __UNIX__ // XXX
	DECLARE_CONTEXT(THREADMUTEX);
#endif
#else /* !__INTERNAL__ */
	DECLARE_CONTEXT(THREADCOND);
#endif /* __INTERNAL__ */



#ifdef __UNIX__
	/* -lpthread is needed */
#	include <pthread.h>
#	define THREADCONTEXT pthread_t
//#	define THREADMUTEX   pthread_mutex_t
typedef pthread_mutex_t __F* THREADMUTEX;
#endif /* __UNIX__ */



DECLARE_CONTEXT(THREADKEY);
DECLARE_CONTEXT(PROCESSKEY);
DECLARE_CONTEXT(PROCESS_CONTEXT);
DECLARE_CONTEXT(PROCESS_SHARED_MEMORY);



struct _PROCINFO {
  char dummy;
};



/* processFork, errors */
/* ERROR_MEM */

/* processWait, flags */
#define PROCESS_FLAG_INFINITE 0x0000  /* wait the process infinitely */
#define PROCESS_FLAG_NOHANG   0x0001  /* return immediately after checking the given process is died */


/* threadWait, error codes */
#define THREAD_DETACHED       1
#define THREAD_DEADLOCK       2


/* threadCreateMutex, error codes */
#define MUTEX_ERROR   2


/* threadTryLockMutex, error codes */
#define MUTEX_BUSY    1



__CBEGIN__


/*
_API( PROCLISTCONTEXT, processOpen, ( void ) );
_API( void, processClose, ( PROCLISTCONTEXT handle ) );
_API( unsigned, processRead, ( PROCLISTCONTEXT handle, struct _PROCINFO __F* info, unsigned flag ) );

_API( int, processFork, ( PROCESSCONTEXT __F* process, const char __F* cmd, const char __F* const __F* arg, const char __F* const __F* env ) );
_API( int, processWait, ( PROCESSCONTEXT process, int __F* status, unsigned flag ) );
_API( void, processExit, ( int status ) );

_API( PROCESSSHARED, processCreateSharedMemory, ( const char __F* name, unsigned long size, unsigned flag ) );
_API( void, processOpenSharedMemory, ( const char __F* name ) );
_API( void, processRemoveSharedMemory, ( PROCESSSHARED shm ) );
_API( void __F*, processAttachSharedMemory, ( PROCESSSHARED shared, unsigned flag ) );
_API( void, processDetachSharedMemory, ( void __F* p ) );

_API( int, processCreateMutex, ( PROCESSMUTEX __F* mutex, PROCESSKEY key, int n, unsigned flag ) );
_API( void, processRemoveMutex, ( PROCESSMUTEX mutex ) );
_API( int, processTryLockMutex, ( PROCESSMUTEX mutex, int i ) );
_API( void, processLockMutex, ( PROCESSMUTEX mutex, int i ) );
_API( void, processUnlockMutex, ( PROCESSMUTEX mutex, int i ) );
*/



_API( _bool, threadInitialize, ( void ) );
_API( void, threadFinalize, ( void ) );

_API( void, threadDetach, ( THREADCONTEXT thread ) );
_API( void __F*, threadSelf, ( void ) );
_API( _bool, threadEqual, ( THREADCONTEXT thread1, THREADCONTEXT __F* thread2 ) );
_API( int, threadWait, ( THREADCONTEXT thread, void __F*__F* result ) );
_API( int, threadCancel, ( THREADCONTEXT thread ) );
_API( void, threadPriority, ( THREADCONTEXT thread, int priority ) );
_API( void, threadYield, ( void ) );
_API( void, threadCooperate, ( void ) );
_API( void, threadExit, ( void __F* result ) );

_API( int, threadCreateKey, ( THREADKEY __F* key ) );
_API( void, threadRemoveKey, ( THREADKEY key ) );
_API( void, threadSetSpecific, ( THREADKEY key, const void __F* value ) );
_API( void __F*, threadGetSpecific, ( THREADKEY key ) );

_API( int, threadCreateMutex, ( THREADMUTEX __F* mutex, unsigned flag ) );
_API( void, threadRemoveMutex, ( THREADMUTEX mutex ) );
_API( int, threadTryLockMutex, ( THREADMUTEX mutex ) );
_API( void, threadLockMutex, ( THREADMUTEX mutex ) );
_API( void, threadUnlockMutex, ( THREADMUTEX mutex ) );

_API( int, threadCreateCond, ( THREADCOND __F* cond ) );
_API( void, threadRemoveCond, ( THREADCOND __F* cond ) );
_API( void, threadSignalCond, ( THREADCOND __F* cond ) );
_API( void, threadBroadcastCond, ( THREADCOND __F* cond ) );
_API( void, threadWaitCond, ( THREADCOND __F* cond, THREADMUTEX __F* mutex ) );


/* more consideration is needed for mutex =====*/
/*_API( int, mutexCreate, ( MUTEXCONTEXT __F* mutex, const char __F* name, int n, unsigned flag ) );
_API( void, mutexRemove, ( MUTEXCONTEXT mutex ) );
_API( void, mutexLock, ( MUTEXCONTEXT mutex, int n ) );
_API( void, mutexTryLock, ( MUTEXCONTEXT mutex, int n ) );
_API( void, mutexUnlock, ( MUTEXCONTEXT mutex, int n ) );*/



__CEND__



#ifdef __UNIX__



#define threadFork(thread,attr,start,param) pthread_create(thread,attr,start,param)
#define threadExit(ret) pthread_exit(ret)
#define threadYield() pthread_yield()
#define threadSelf() pthread_self()
#define threadWait(thread,ret) pthread_join(thread,(void**)ret);

#define threadTryLockMutex pthread_mutex_trylock
#define threadLockMutex pthread_mutex_lock
#define threadUnlockMutex pthread_mutex_unlock

//#define DEFINE_MUTEX(name) pthread_mutex_t name
//#define EXTERN_MUTEX(name) extern pthread_mutex_t name
//#define mutexInit(m,attr) pthread_mutex_init(m,attr)
//#define mutexClear(m)     pthread_mutex_destroy(m)
//#define mutexLock(m)      pthread_mutex_lock(m)
//#define mutexUnlock(m)    pthread_mutex_unlock(m)



#else /* !__UNIX__ */
#ifdef __WIN32__

# ifndef WINVER
  void __F* __API GetCurrentThread P_V;
  void __API Sleep P_A(( unsigned long msec ));
  void __API ExitThread P_A(( unsigned long exitcode ));
  _bool __API TlsFree P_A(( unsigned long key ));
  _bool __API TlsSetValue P_A(( unsigned long key, void __F* value ));
  void __F* __API TlsGetValue P_A(( unsigned long key ));
# endif /* WINVER */

# define threadEqual(thread1,thread2) ((thread1)==(thread2))
# define threadYield() Sleep(0)
# define threadCooperate()
# define threadRemoveKey(key) TlsFree((unsigned long)key)
# define threadSetSpecific(key,value) TlsSetValue((unsigned long)key,value)
# define threadGetSpecific(key) TlsGetValue((unsigned long)key)



/* mutex */
/*#define MUTEX_SUPPORTED
#define DEFINE_MUTEX(name) CRITICAL_SECTION name;
#define EXTERN_MUTEX(name) extern CRITICAL_SECTION name;
typedef CRITICAL_SECTION _MUTEX;
#define mutexInit(m)   InitializeCriticalSection(m)
#define mutexClear(m)  DeleteCriticalSection(m)
#define mutexLock(m)   EnterCriticalSection(m)
#define mutexUnlock(m) LeaveCriticalSection(m)
*/


#else /* __WIN32__ */



/* cooperative thread implementation */


#define mutexInit(m)
#define mutexClear(m)
#define mutexLock(m)
#define mutexUnlock(m)



#endif /* __WIN32__ */
#endif /* __UNIX__ */



#endif /* __CORAL_EXT_PROCESS_H */
