/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Coral File Operations */


#ifndef __CORAL_FILE_H
# define __CORAL_FILE_H

#include <coral/io.h>




/*** fileOpen */

/* opening modes */
/* IO_FLAG_READ, reading mode
   IO_FLAG_WRITE, writing mode */
#define FILE_FLAG_OVERWRITE 0x0100  /* overwrite the existing file */
#define FILE_FLAG_TRUNCATE  0x0200  /* truncate file for overwriting mode */
#define FILE_FLAG_EXISTENCE 0x0400  /* file existence test should be done for writing mode : FILE_FLAG_OVERWRITE is meaningless because overwriting is default for this option */
/* From 0x1000 to 0x8000 will never be used by fileOpen */



/* fileOpen, error codes
ERROR_DUPLICATED, the given file is already
ERROR_PARAM, the given flag is invalid - io mode is unknown
ERROR_IO, io error - file was not opened  */






/*** fileCopy */
/* all error codes returned by ioTransfer */
#define FILE_COPY_SRC_NOT_OPENED   (ERROR_RESERVED-11) /* the source file was not opened */
#define FILE_COPY_DEST_NOT_OPENED  (ERROR_RESERVED-12) /* the destination file was not opened */

/* FILE_FLAG_OVERWRITE, overwrite existing file
   FILE_FLAG_TRUNCATE, don't append data */
#define FILE_FLAG_OFFSET 0x1000 /* the space to store 2 long values which contains destination and source offsets, is reserved at the beginning of the supplied progression context */



__CBEGIN__



_API( int, fileOpen, ( struct _IO __F* io, const char __F* filename, unsigned flag ) );
_API( int, fileClose, ( struct _IO __F* io ) );
_API( long, fileRead, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, fileWrite, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, fileSeek, ( struct _IO __F* io, long offset, long __F* offset_high, int seekpos ) );
_API( int, fileSetEOF, ( struct _IO __F* io ) );

_API( int, fileOpenNative, ( struct _IO __F* io, const char __F* filename, unsigned flag ) );
_API( int, fileCloseNative, ( struct _IO __F* io ) );
_API( long, fileReadNative, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, fileWriteNative, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, fileSeekNative, ( struct _IO __F* io, long offset, long __F* offset_high, int seekpos ) );
_API( int, fileSetEOFNative, ( struct _IO __F* io ) );

_API( char __F*, fileLoad, ( const char __F* filename, long offset, long __F* offset_high, _integer1 __F* buf, long __F* p_size ) );
_API( _bool, fileSave, ( const char __F* filename, long offset, long __F* offset_high, _integer1 __F* buf, long size, unsigned flag ) );
_API( int, fileCopy, ( const char __F* destname, const char __F* srcname, long destoffset, long destoffset_high, long srcoffset, long srcoffset_high, long totalsize, _integer1 __F* buf, int bufsize, unsigned flag, int (__API* progress)(void __F* context,int size), void __F* context ) ); /* XXX not implemented yet */



__CEND__



#ifndef __WIN32__
# define fileOpenNative fileOpen
# define fileCloseNative fileClose
# define fileReadNative fileRead
# define fileWriteNative fileWrite
# define fileSeekNative fileSeek
#endif /* !__WIN32__ */



#endif /* __CORAL_FILE_H */

