/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Integer math */

#ifndef __CORAL_INTEGER_H
# define __CORAL_INTEGER_H

#include <coral/def.h>



/* x_string */
#define INTX_FLAG_UNSIGNED  0x1000 /* for radix==10, given value is unsigned */
#define INTX_FLAG_UPPERCASE 0x2000 /* uppercase character is used if radix>10 */



__CBEGIN__


_API( long, intGCM, ( long a, long b ) );
_API( long, intLCM, ( long a, long b ) );


/*
 Do not use x_const and x_short as an initializer constant for a global variable.
 Sample usage for a=b*c+d, x_store( a, x_add( x_mul(b,c), d ) ).
*/

#ifdef __INTEGERX__

 /* arithmetic */
# define x_sign(a) (-(a))
# define x_add(a,b) ((a)+(b))
# define x_adds(a,b) ((a)+=(b))
# define x_inc(a) ((a)++)
# define x_sub(a,b) ((a)-(b))
# define x_subs(a,b) (a-=(b))
# define x_dec(a) ((a)--)
# define x_mul(a,b) ((a)*(b))
# define x_muls(a,b) ((a)*=(b))
# define x_div(a,b) ((a)/(b))
# define x_divs(a,b) ((a)/=(b))
# define x_udiv(a,b) ((__unsignedx(a)/(__unsignedx)(b))
# define x_udivs(a,b) ((__unsignedx)(a)/=(__unsignedx)(b))
# define x_mod(a,b) ((a)%(b))
# define x_mods(a,b) ((a)%=(b))
# define x_umod(a,b) ((__unsignedx)(a)%(__unsignedx)(b))
# define x_umods(a,b) ((__unsignedx)(a)%=(__unsignedx)(b))

 /* bit */
# define x_and(a,b) ((a)&(b))
# define x_ands(a,b) ((a)&=(b))
# define x_or(a,b) ((a)|(b))
# define x_ors(a,b) ((a)|=(b))
# define x_xor(a,b) ((a)^(b))
# define x_xors(a,b) ((a)^=(b))
# define x_not(a) (~(a))
# define x_shl(a,cnt) ((a)<<(cnt))
# define x_shls(a,cnt) ((a)<<=(cnt))
# define x_shr(a,cnt) ((a)>>(cnt))
# define x_shrs(a,cnt) ((a)>>=(cnt))
# define x_ushr ((__unsignedx)(a)>>(cnt))
# define x_ushrs ((__unsignedx)(a)>>=(cnt))

 /* logical */
# define x_zero(a) (!(a))
# define x_nonzero(a) (a)
# define x_eq(a,b) ((a)==(b))
# define x_ne(a,b) ((a)!=(b))
# define x_gt(a,b) ((a)>(b))
# define x_ge(a,b) ((a)>=(b))
# define x_lt(a,b) ((a)<(b))
# define x_le(a,b) ((a)<=(b))
# define x_ugt(a,b) ((__unsignedx)(a)>(__unsignedx)(b))
# define x_uge(a,b) ((__unsignedx)(a)>=(__unsignedx)(b))
# define x_ult(a,b) ((__unsignedx)(a)<(__unsignedx)(b))
# define x_ule(a,b) ((__unsignedx)(a)<=(__unsignedx)(b))

 /* miscellaneous */
# define x_const(high,low) ((__integerx)((((__integerx)(high))<<(sizeof(__integerx)*4))|(unsigned long)(low)))
# define x_high(a) ((long)((a)>>(sizeof(__integerx)*4)))
# define x_low(a) ((long)(a))
# define x_short(low) (low)
# define x_conv(type,a) ((type)(a))

#else /* !__LONGLONG__ */

# ifdef DEF_INTX
#  undef DEF_INTX
# else
  typedef struct { long a, b; } __integerx, __unsignedx;
# endif

 /* internal flags used in x_divide */
# define INTX_FLAG_DIVISION 0x0000
# define INTX_FLAG_MODULUS  0x0002
 /* INTX_FLAG_UNSIGNED */

#ifdef __TURBOC__
//# pragma warn -stv /* turn off 'structure passed by value */
#endif

 /* arithmetic */
 _API( __integerx, x_sign, ( __integerx a) );
 _API( __integerx, x_add, ( __integerx a, __integerx b ) );
# define x_adds(a,b) x_store(a,x_add(a,b))
 _API( __integerx, x_add1, ( __integerx a ) );
# define x_inc(a) x_store(a,x_add1(a))
 _API( __integerx, x_sub, ( __integerx a, __integerx b ) );
# define x_subs(a,b) x_store(a,x_sub(a,b))
 _API( __integerx, x_sub1, ( __integerx a ) );
# define x_dec(a) x_store(a,x_sub1(a))
 _API( __integerx, x_mul, ( __integerx a, __integerx b ) );
# define x_muls(a,b) x_store(a,x_mul(a,b))
 _API( __integerx, x_divide, ( __integerx a, __integerx b, unsigned flag ) );
# define x_div(a,b) x_divide(a,b,INTX_FLAG_DIVISION)
# define x_divs(a,b) x_store(a,x_div(a,b))
# define x_udiv(a,b) x_divide(a,b,INTX_FLAG_DIVISION|INTX_FLAG_UNSIGNED)
# define x_udivs(a,b) x_store(a,x_udiv(a,b))
# define x_mod(a,b) x_divide(a,b,INTX_FLAG_MODULUS)
# define x_mods(a,b) x_store(a,x_mod(a,b))
# define x_umod(a,b) x_divide(a,b,INTX_FLAG_MODULUS|INTX_FLAG_UNSIGNED)
# define x_umods(a,b) x_store(a,x_umod(a,b))

 /* bit */
 _API( __integerx, x_and, ( __integerx a, __integerx b ) );
# define x_ands(a,b) x_store(a,x_and(a,b))
 _API( __integerx, x_or, ( __integerx a, __integerx b ) );
# define x_ors(a,b) x_store(a,x_or(a,b))
 _API( __integerx, x_xor, ( __integerx a, __integerx b ) );
# define x_xors(a,b) x_store(a,x_xor(a,b))
 _API( __integerx, x_not, ( __integerx a ) );
 _API( __integerx, x_shl, ( __integerx a, unsigned cnt ) );
# define x_shls(a,cnt) x_store(a,x_shl(a,cnt))
 _API( __integerx, x_shr, ( __integerx a, unsigned cnt ) );
# define x_shrs(a,cnt) x_store(a,x_shr(a,cnt))
 _API( __integerx, x_ushr, ( __integerx a, unsigned cnt ) );
# define x_ushrs(a,cnt) x_store(a,x_ushr(a,cnt))

 /* logical */
 _API( _bool, x_zero, ( __integerx a ) );
 _API( _bool, x_nonzero, ( __integerx a ) );
 _API( _bool, x_eq, ( __integerx a, __integerx b ) );
 _API( _bool, x_ne, ( __integerx a, __integerx b ) );
 _API( _bool, x_gt, ( __integerx a, __integerx b ) );
 _API( _bool, x_ge, ( __integerx a, __integerx b ) );
 _API( _bool, x_lt, ( __integerx a, __integerx b ) );
 _API( _bool, x_le, ( __integerx a, __integerx b ) );
 _API( _bool, x_ugt, ( __unsignedx a, __unsignedx b ) );
 _API( _bool, x_uge, ( __unsignedx a, __unsignedx b ) );
 _API( _bool, x_ult, ( __unsignedx a, __unsignedx b ) );
 _API( _bool, x_ule, ( __unsignedx a, __unsignedx b ) );

 /* miscellaneous */
 _API( __integerx, x_const, ( long high, long low ) );
 _API( __integerx, x_short, ( long low ) );
 _API( long, x_conv_x, ( __integerx a ) );
# define x_conv(type,a) ((type)x_conv_x(a))

#endif /* __LONGLONG__ */

#define x_umul  x_mul
#define x_umuls x_muls
#define x_ushl x_shl
#define x_ushls x_shls
#define x_store(a,b) ((a)=(b))

_API( unsigned, x_string, ( __integerx val, char __F* s, unsigned n, unsigned radix ) );
_API( __integerx, x_value, ( const char __F* s, unsigned radix ) );


__CEND__



/* operator support for c++
  This class is supported to write extended integer operation codes in c++
 using its operators without consideration of a compiler which doesn't support
 long long integer. By this reason, no extensions which cannot be done by
 long long are not present. 
  For long long supported compilers, this class is not defined */

#if defined(__cplusplus) && !defined(__LONGLONG__) && defined(LLL)

#include <iostream.h>

 class _integerx;

 class _unsignedx
 {
  friend class _integerx; /* you are my friend */

  protected:

  __unsignedx value;

  inline _unsignedx(){}

  public:

  /* arithmetic */

  inline _unsignedx operator + ()
  {
   return *this;
  }

  inline _unsignedx operator - ()
  {
   return x_sign( value );
  }

  friend _unsignedx operator + ( _unsignedx __F& a, _unsignedx __F& b )
  {
   return x_add( a.value, b.value );
  }

  inline _unsignedx __F& operator += ( _unsignedx __F& a )
  {
   x_adds( value, a.value );
   return *this;
  }

  inline _unsignedx __F& operator ++ ()
  {
   x_inc( value );
   return *this;
  }

  friend _unsignedx operator - ( _unsignedx __F& a, _unsignedx __F& b )
  {
   return x_sub( a.value, b.value);
  }

  inline _unsignedx __F& operator -= ( _unsignedx __F& a )
  {
   x_subs( value, a.value );
   return *this;
  }

  inline _unsignedx __F& operator -- ()
  {
   x_dec( value );
   return *this;
  }

  friend _unsignedx operator * ( _unsignedx __F& a, _unsignedx __F& b )
  {
   return x_umul( a.value, b.value );
  }

  inline _unsignedx __F& operator *= ( _unsignedx __F& a )
  {
   x_umuls( value, a.value );
   return *this;
  }

  friend _unsignedx operator / ( const _unsignedx __F& a, const _unsignedx __F& b )
  {
   return x_udiv( a.value, b.value );
  }

  inline _unsignedx __F& operator /= ( const _unsignedx __F& a )
  {
   x_udivs( value, a.value );
   return *this;
  }

  friend _unsignedx operator % ( const _unsignedx __F& a, const _unsignedx __F& b )
  {
   return x_umod( a.value, b.value );
  }

  inline _unsignedx __F& operator %= ( const _unsignedx __F& a )
  {
   x_umods( value, a.value );
   return *this;
  }

  friend _unsignedx operator & ( const _unsignedx __F& a, const _unsignedx __F& b )
  {
   return x_and( a.value, b.value );
  }

  inline _unsignedx __F& operator &= ( const _unsignedx __F& a )
  {
   x_ands( value, a.value );
   return *this;
  }

  friend _unsignedx __F operator | ( const _unsignedx __F& a, const _unsignedx __F& b )
  {
   return x_or( a.value, b.value );
  }

  inline _unsignedx __F& operator |= ( const _unsignedx __F& a )
  {
   x_ors( value, a.value );
   return *this;
  }

  friend _unsignedx operator ^ ( const _unsignedx __F& a, const _unsignedx __F& b )
  {
   return x_xor( a.value, b.value );
  }

  inline _unsignedx __F& operator ^= ( const _unsignedx __F& a )
  {
   x_xors( value, a.value );
   return *this;
  }

  inline _unsignedx operator ~ ()
  {
   return x_not( value );
  }

  inline _unsignedx operator << ( unsigned cnt )
  {
   return x_shl( value, cnt );
  }

  inline _unsignedx __F& operator <<= ( unsigned cnt )
  {
   x_shls( value, cnt );
   return *this;
  }

  inline _unsignedx operator >> ( unsigned cnt )
  {
   return x_ushr( value, cnt );
  }

  inline _unsignedx __F& operator >>= ( unsigned cnt )
  {
   x_ushrs( value, cnt );
   return *this;
  }

  inline _bool operator ! ()
  {
   return x_zero(value);
  }

/*  inline _bool operator () ()
  {
   return x_nonzero(value);
  } */

  inline _bool operator == ( _unsignedx __F& a )
  {
   return x_eq( value, a.value );
  }

  inline _bool operator != ( _unsignedx __F& a )
  {
   return x_ne( value, a.value );
  }

  friend _bool operator > ( _unsignedx __F& a, _unsignedx __F& b )
  {
   return x_ugt( a.value, b.value );
  }

  friend _bool operator >= ( _unsignedx __F& a, _unsignedx __F& b )
  {
   return x_uge( a.value, b.value );
  }

  friend _bool operator < ( _unsignedx __F& a, _unsignedx __F& b )
  {
   return x_ult( a.value, b.value );
  }

  friend _bool operator <= ( _unsignedx __F& a, _unsignedx __F& b )
  {
   return x_ule( a.value, b.value );
  }

  inline _unsignedx __F& operator = ( const _unsignedx __F& a )
  {
   x_store( value, a.value );
   return *this;
  }



  inline _unsignedx( __unsignedx a )
  {
   x_store(value,a);
  }

  friend ostream __F& operator<<( ostream __F& os, const _unsignedx __F& a )
  {
   char buf[sizeof(__unsignedx)*8+1];
   x_string( a.value, buf, sizeof(buf), 10 );
   os << buf;
   return os;
  }

  inline operator _integerx()
  {
   return *this;
  }
 };



 class _integerx : public _unsignedx
 {
  private :

  inline _integerx() {}


  public :

  /* arithmetic */

  inline _integerx operator + ()
  {
   return *this;
  }

  inline _integerx operator - ()
  {
   return x_sign( value );
  }

  inline _integerx operator + ( _integerx __F& a )
  {
   return x_add( value, a.value );
  }

  inline _integerx __F& operator += ( _integerx __F& a )
  {
   x_adds( value, a.value );
   return *this;
  }

  inline _integerx __F& operator ++ ()
  {
   x_inc( value );
   return *this;
  }

  inline _integerx operator - ( _integerx __F& a )
  {
   return x_sub( value, a.value);
  }

  inline _integerx __F& operator -= ( _integerx __F& a )
  {
   x_subs( value, a.value );
   return *this;
  }

  inline _unsignedx __F& operator -- ()
  {
   x_dec( value );
   return *this;
  }

  inline _integerx operator * ( _unsignedx __F& a )
  {
   return x_mul( value, a.value );
  }

  inline _integerx __F& operator *= ( const _integerx __F& a )
  {
   x_muls( value, a.value );
   return *this;
  }

  inline _integerx operator / ( const _integerx __F& a )
  {
   return x_div( value, a.value );
  }

  inline _integerx __F& operator /= ( const _integerx __F& a )
  {
   x_divs( value, a.value );
   return *this;
  }

  inline _integerx operator % ( const _unsignedx __F& a )
  {
   return x_mod( value, a.value );
  }

  inline _integerx __F& operator %= ( const _integerx __F& a )
  {
   x_mods( value, a.value );
   return *this;
  }

  inline _integerx operator & ( const _integerx __F& a )
  {
   return x_and( value, a.value );
  }

  inline _integerx __F& operator &= ( const _integerx __F& a )
  {
   x_ands( value, a.value );
   return *this;
  }

  inline _integerx __F operator | ( const _integerx __F& a )
  {
   return x_or( value, a.value );
  }

  inline _integerx __F& operator |= ( const _integerx __F& a )
  {
   x_ors( value, a.value );
   return *this;
  }

  inline _integerx operator ^ ( const _unsignedx __F& a )
  {
   return x_xor( value, a.value );
  }

  inline _integerx __F& operator ^= ( const _integerx __F& a )
  {
   x_xors( value, a.value );
   return *this;
  }

  inline _unsignedx operator ~ ()
  {
   return x_not( value );
  }

  inline _integerx operator << ( unsigned cnt )
  {
   return x_shl( value, cnt );
  }

  inline _integerx __F& operator <<= ( unsigned cnt )
  {
   x_shls( value, cnt );
   return *this;
  }

  inline _integerx operator >> ( unsigned cnt )
  {
   return x_shr( value, cnt );
  }

  inline _integerx __F& operator >>= ( unsigned cnt )
  {
   x_shrs( value, cnt );
   return *this;
  }


  inline _bool operator > ( _integerx __F& a )
  {
   return x_gt( value, a.value );
  }

  inline _bool operator >= ( _integerx __F& a )
  {
   return x_ge( value, a.value );
  }

  inline _bool operator < ( _integerx __F& a )
  {
   return x_lt( value, a.value );
  }

  inline _bool operator <= ( _integerx __F& a )
  {
   return x_le( value, a.value );
  }

  inline _integerx __F& operator = ( _integerx __F& a )
  {
   x_store( value, a.value );
   return *this;
  }

  inline _integerx __F& operator = ( _unsignedx __F& a )
  {
   x_store( value, a.value );
   return *this;
  }


  inline _integerx( __integerx a )
  {
   x_store(value,a);
  }

 };



#else /* !__cplusplus || __LONGLONG__ */

 typedef __integerx _integerx;
 typedef __unsignedx _unsignedx;

#endif /* __cplusplus && !__LONGLONG__ */



#endif /* __CORAL_INTEGER_H */
