/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Coral Open I/O */



#ifndef __CORAL_IO_H
# define __CORAL_IO_H

#include <coral/def.h>



#define IO_END            0    /* io stream ended */
#define IO_SEEK_BEGIN     0
#define IO_SEEK_CURRENT   1
#define IO_SEEK_END       2
#define IOTEXT_LINE_END  '\n' /* line feed is the standard line end character */



/* ioLockBuffer */
/* following two flags must not be combinated for ioLockBuffer */
#define IO_FLAG_READ      0x0001  /* lock buffer for reading */
#define IO_FLAG_WRITE     0x0002  /* lock buffer for writing */
/* following three flags cannot be combinated */
//#define IO_FLAG_AVAILABLE 0x0010  /* lock available size, size parameter is integer pointer to receive locked size */
//#define IO_FLAG_REQUESTED 0x0020  /* lock requested size, size parameter is integer value to request locked buffer size */
//#define IO_FLAG_MINIMUM   0x0040  /* lock at least the given size, size parameter is a pointer of the initialized integer to receive locked size */


/* ioTransfer, return values */
/* 0, if this function succeeds
   FAIL_CANCEL, progression procedure returns ERROR_CANCEL
   FAIL_IO, io error occurred
   FAIL_MEM, buffer allocation problem
   FAIL_BUFFER, the given buffer length is too short */
#define IO_INSUFFICIENT_SOURCE (FAIL_RESERVED-1) /* source stream is insufficient to accept the specified total size to be transferred */



#define IO_MASK_SUBTREE_SIZE_FIELD        0x07 /* 3 bits */
#	define IO_SUBTREE_SIZE_FIELD_0        0x00
#	define IO_SUBTREE_SIZE_FIELD_1        0x01
#	define IO_SUBTREE_SIZE_FIELD_2        0x02
#	define IO_SUBTREE_SIZE_FIELD_4        0x03

#define IO_MASK_KEY_LENGTH_FIELD          0x18 /* 2 bits, internally determined */
#	define IO_KEY_LENGTH_FIELD_0          0x00
#	define IO_KEY_LENGTH_FIELD_1          0x08
#	define IO_KEY_LENGTH_FIELD_2          0x10

#define IO_MASK_NODE_PROPERTY             0x60 /* 2 bits */
#	define IO_TERMINAL_NODE               0x00
#	define IO_ENTER_INTERNAL_NODE         0x20
#	define IO_LEAVE_INTERNAL_NODE         0x40

#define IO_FLAG_EXTRA_SUBTREE_INFO        0x80


/* extra subtree informations */

#define IO_MASK_EXTERNAL_FILE_NAME_FIELD  0x00c0
#	define IO_EXTERNAL_FILE_NAME_FIELD_0  0x0000
#	define IO_EXTERNAL_FILE_NAME_FIELD_1  0x0040
#	define IO_EXTERNAL_FILE_NAME_FIELD_2  0x0080

#define IO_MASK_SUBTREE_STATE             0x0038
#	define IO_COLLAPSE_SUBTREE            0x0000
#	define IO_OPEN_SUBTREE                0x0008
#	define IO_OPEN_ALL_SUBTREE            0x0010



/* _IO_HIERARCHY_STACK */

struct _IO_HIERARCHY_STACK {
	unsigned      size_field;   /* width of the size field */
	unsigned long size;         /* the total size of the subtree */
	unsigned long offset;       /* file offset for the current hierarchy node */
	unsigned long offset_high;
};



/* _IOTEXT */

struct _IOTEXT {
	unsigned long line;      /* the current line number */
	unsigned      col;       /* the current column number */
	int llinecmt;            /* line comment length, zero if not present */
	int lblockcmt1;          /* leading block comment length, zero if not present */
	int lblockcmt2;          /* trailing block comment length */
	const char __F* linecmt, __F* blockcmt1, __F* blockcmt2; /* commemt specifiers */
	int (__API* event) P_A(( struct _IOBUFFER __F* iob,  struct _IOTEXT __F* iot, int code, void __F* param ) ); /* event procedure */
	void __F*   eventparam; /* user data to be supplied to the callback procedure */
	unsigned    eventmask; /* event mask */

	/* results reported by iotextRead */
	char          lastchar;   /* termination character */
	unsigned long invalidcmt; /* line number where the invalid comment was begun */
	unsigned      lastcol;    /* the last meaningful column number before the current column number is reset */
	int           length;     /* read length */
};


/* return values for dispatch procedure */
/* IOTEXT_DISPATCH_FAIL */
#define IOTEXT_DISPATCH_SUCCESS 1 /* procedure was normally returned */
#define IOTEXT_SKIP_CHAR        2 /* skip the current character, valid for IOTEXT_CHAR event */

/* events
	Buffer lock-unlock cannot be used for 4 character events.
	The order for comment events should be 'begin', 'end', and 'char'. */
#define IOTEXT_CHAR                0 /* a character was read, param will be the pointer of the current character */
#define IOTEXT_DELIMITER           1 /* a delimiter was found */
#define IOTEXT_BEGIN_LINE_COMMENT  2 /* line comment was begun, param will be the specifier */
#define IOTEXT_END_LINE_COMMENT    3 /* line comment was ended, param will be the specifier */
#define IOTEXT_LINE_COMMENT_CHAR   4 /* a character in a line comment was read, param will be the pointer of the current character */
#define IOTEXT_BEGIN_BLOCK_COMMENT 5 /* block comment was begun, param will be the specifier */
#define IOTEXT_END_BLOCK_COMMENT   6 /* block comment was ended, param will be the specifier */
#define IOTEXT_BLOCK_COMMENT_CHAR  7 /* a character in a block comment was read, param will be the pointer of the current character */



/* event masks
 to reduce overhead caused by event procedure call, following event masks are used */
#define IOTEXT_FLAG_CHAR               0x0001 /* the character event is wanted */
#define IOTEXT_FLAG_DELIMITER          0x0002 /* the delimiter event is wanted */
#define IOTEXT_FLAG_LINE_COMMENT_CHAR  0x0004 /* the line comment character event is wanted */
#define IOTEXT_FLAG_BLOCK_COMMENT_CHAR 0x0008 /* the block comment character event is wanted */
#define IOTEXT_FLAG_LINE_COMMENT       0x0010 /* the line comment event is wanted */
#define IOTEXT_FLAG_BLOCK_COMMENT      0x0020 /* the block comment event is wanted */



/*** iotextRead */



/* terminated reasons
 FAIL_IO, an io error was occurred
 FAIL_BUFFER, buffer is full
 IO_END , stream ended
 IOTEXT_DELIMITER, a delimiter found and 'lastchar' member will be the found delimiter */
#define IOTEXT_COMMENT          2 /* a comment was found */
#define IOTEXT_INVALID_COMMENT (FAIL_RESERVED-1) /* an invalid block comment found */
//???? #define IOTEXT_INVALID_VALUE   (FAIL_RESERVED-2) /* invalid value was found */
#define IOTEXT_DISPATCH_FAIL   (FAIL_RESERVED-3) /* dispatch procedure returns zero */



/* option flags */
#define IOTEXT_FLAG_EXCLUSIVE           0x0001 /* delimiters are tested exclusively */
#define IOTEXT_FLAG_FORCE               0x0002 /* force to find a delimiter though the buffer is full */
#define IOTEXT_FLAG_DISABLE_DELIMITER   0x0004 /* disable delimiter */
#define IOTEXT_FLAG_DISABLE_COMMENT     0x0008 /* disable comment */
#define IOTEXT_FLAG_KEEP_LEAD_BLANK     0x0010 /* keep leading blanks ( ignore initial skipping ) */
#define IOTEXT_FLAG_KEEP_LEAD_DELIMITER 0x0020 /* keep leading delimiters ( exclude white characters in supplied delimiters for initial skipping ) */
#define IOTEXT_FLAG_KEEP_TRAIL_BLANK    0x0040 /* keep trailing blanks */
#define IOTEXT_FLAG_KEEP_LEAD_COMMENT   0x0080 /* keep leading comment to return IOTEXT_COMMENT */
#define IOTEXT_FLAG_PRESERVE_LINE_END   0x0100 /* don't force unix line end */



/*** iotextReadString */

/* option flags */
#define IOTEXT_FLAG_RESERVE_CHAR_ESC 0x0001 /* reserve c style character escape sequence */
/* IOTEXT_FLAG_FORCE : force to find a delimiter though the buffer is full */
#define IOTEXT_FLAG_NO_LINE_CONT     0x0004 /* don't use c style line continuation character */
//#define IOTEXT_FLAG_NO_CHAR_         0x0008 /* ignore c style character escape sequence */
#define IOTEXT_FLAG_ALLOW_LINE_END   0x0010 /* line end character can be found within a string */
#define IOTEXT_FLAG_SKIP_LEAD_BLANK  0x0020 /* ignore leading blanks */
#define IOTEXT_FLAG_SKIP_TRAIL_BLANK 0x0040 /* skip trailing blanks */
/* IOTEXT_FLAG_PRESERVE_LINE_END : don't force unix line end */
#define IOTEXT_FLAG_ENABLE_COMMENT   0x8000 /* enable comment between string specifiers */



/* error codes */
/* FAIL_IO, io error was occurred
   FAIL_BUFFER, buffer is full
   IOTEXT_GARBAGE_FOUND, a garbage was found while searching leading string specifier */
#define IOTEXT_INVALID_STRING   (FAIL_RESERVED-4) /* invalid string */



/*** iotextWait */
/* additional return value */
#define IOTEXT_GARBAGE_FOUND -100 /* garbage found before founding a specified delimiter */



/*** iotextConv, option flags */
/*** iotextFormat, option flags */
/*** iotextWrite, option flags */
#define IOTEXT_LF      0x01             /* line feed */
#define IOTEXT_CR      0x02             /* carriage return */
#define IOTEXT_CRLF    (0x01|0x02)      /* cr/lf combi */
#define IOTEXT_LFCR    (0x01|0x02|0x04) /* lf/cr combi */
#define IOTEXT_UNIX    IOTEXT_LF       /* unix line ends */
#define IOTEXT_MAC     IOTEXT_CR       /* macintosh line ends */
#define IOTEXT_DOS     IOTEXT_CRLF     /* dos line ends */
#define IOTEXT_UNKNOWN 0x08             /* text format cannot be detected */



/* _IOLOCKMEMORY */

struct _IOLOCKMEMORY {
	int size, current;  /* the size and the current position */
	_integer1 __F* p;   /* pointer to lock */
};



/* _IOLOCKCONSOLE */

struct _IOLOCKCONSOLE {
	void __F* context;  /* context */
	int (__API* prompt) P_A(( struct _IOLOCKCONSOLE __F* ioc ));
	int (__API* read) P_A(( struct _IOLOCKCONSOLE __F* ioc, _integer1 __F* buf, int bufsize ));
	int (__API* write) P_A(( struct _IOLOCKCONSOLE __F* ioc, _integer1 __F* buf, int bufsize ));
	unsigned flag;      /* general purpose flag */
	unsigned lockflag;  /* the current lock flag */
	int state;          /* the current state */
	int inbufsize;      /* input buffer size */
	int inbufpos;       /* the current input buffer position */
	int inbuflen;       /* effective input buffer length */
	int outbufsize;     /* output buffer size */
	int outbufpos;      /* the current output buffer position */
	_integer1 buf[1];   /* buffer ( inbufsize + outbufsize ) */
};



/* _IOLOCKCONSOLE, flags */
#define IO_FLAG_PROMPT 0x0001 /* use prompt */



/* _IOLOCKCONSOLE, states */

/* IO_END */
#define IO_BREAK  -1
#define IO_NORMAL 1



/* _IOOFFSET */

struct _IOOFFSET {
	struct _IO io;          /* actual I/O context */
	_integer8 start_offset; /* absolute offset */
	_integer8 limit;        /* limit */
	_integer8 size;         /* size */
	_integer8 offset;       /* relative offset */
};



__CBEGIN__



_API( void, ioEmpty, ( struct _IO __F* io ) );
_API( int, ioOpen, ( struct _IO __F* io, unsigned size ) );
_API( void __F*, ioClose, ( struct _IO __F* io, int __F* result ) );
_API( long, ioEnhance, ( struct _IO __F* io, void __F* p, long size, _unsigned2 (__API* ioproc)( struct _IO __F* io, void __F* p, _unsigned2 size ) ) );
_API( void, ioResetBuffer, ( struct _IO __F* io ) );
_API( _bool, ioAllocBuffer, ( struct _IO __F* io, int size ) );
_API( _bool, ioFreeBuffer, ( struct _IO __F* io ) );
_API( int, ioGetBufferSize, ( struct _IO __F* io ) );
_API( void, ioSetReservedBufferSize, ( struct _IO __F* io, int reserved ) );
_API( _bool, ioFlushBuffer, ( struct _IO __F* io ) );
_API( _bool, ioSyncBuffer, ( struct _IO __F* io ) );
_API( _bool, ioTruncateBuffer, ( struct _IO __F* io, long cp, long size ) );
_API( _integer1 __F*, ioLockBuffer, ( struct _IO __F* io, int __F* locked, int size, unsigned flag ) );
_API( void, ioUnlockBuffer, ( struct _IO __F* io, _integer1 __F* current ) );
_API( void, ioUnlockRewindBuffer, ( struct _IO __F* io, _integer1 __F* current, int offset ) );
_API( void, ioInitializeBuffer, ( struct _IOBUFFER __F* iob, struct _IO __F* io ) );
_API( _integer1 __F*, ioLockMemory, ( struct _IOLOCKMEMORY __F* iom, int __F* locked, int size, unsigned flag ) );
_API( void, ioUnlockMemory, ( struct _IOLOCKMEMORY __F* iom, _integer1 __F* current ) );
_API( void, ioInitializeMemoryBuffer, ( struct _IOBUFFER __F* iob, struct _IOLOCKMEMORY __F* iom, _integer1 __F* buf, int bufsize ) );
_API( _integer1 __F*, ioLockConsole, ( struct _IOLOCKCONSOLE __F* ioc, int __F* locked, int size, unsigned lockflag ) );
_API( void, ioUnlockConsole, ( struct _IOLOCKCONSOLE __F* ioc, _integer1 __F* current ) ); 
_API( void, ioFlushConsole, ( struct _IOLOCKCONSOLE __F* ioc ) );
_API( struct _IOLOCKCONSOLE __F* , ioInitializeConsoleBuffer, ( struct _IOLOCKCONSOLE __F* ioc, void __F* context, int (__API* prompt)( struct _IOLOCKCONSOLE __F* ioc ), int (__API* read)( struct _IOLOCKCONSOLE __F* ioc, _integer1 __F* buf, int bufsize ), int (__API* write)( struct _IOLOCKCONSOLE __F* ioc, _integer1 __F* buf, int bufsize ), int inbufsize, int outbufsize, unsigned flag ) );
_API( int, ioGetConsoleState, ( struct _IOLOCKCONSOLE __F* ioc ) );
_API( void, ioResetConsoleState, ( struct _IOLOCKCONSOLE __F* ioc ) );
_API( long, ioRead, ( struct _IO __F* io, void __F* p, long n, unsigned byteorder ) );
_API( long, ioWrite, ( struct _IO __F* io, void __F* p, long n, unsigned byteorder ) );
_API( long, ioSeek, ( struct _IO __F* io, long offset, long __F* offset_high, int seekpos ) );
_API( int, ioOffsetStart, ( struct _IOOFFSET __F* iooffset ) );
_API( long, ioOffsetRead, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, ioOffsetWrite, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, ioOffsetSeek, ( struct _IO __F* io, long offset, long __F* p_offset_high, int pos ) );
_API( _bool, ioSkip, ( struct _IO __F* io, long size ) );
_API( _bool, ioSkipBuffer, ( struct _IOBUFFER __F* iob, int size ) );
_API( _bool, ioFullRead, ( struct _IO __F* io, void __F* p, long n, unsigned byteorder ) );
_API( _bool, ioFullWrite, ( struct _IO __F* io, void __F* p, long n, unsigned byteorder ) );
_API( _bool, ioWriteInt, ( struct _IO __F* io, long value, unsigned flag ) );
_API( _bool, ioReadRecord, ( struct _IO __F* io, void __F* p, const void __F* ri, unsigned flag ) );
_API( _bool, ioWriteRecord, ( struct _IO __F* io, void __F* p, const void __F* ri, unsigned flag ) );
_API( _bool, ioRead2D, (struct _IO __F* io, void __F* p, long row, unsigned short item, const void __F* colrow, const void __F* colcol, _unsigned2 variable, _unsigned2 byteorder ) );
_API( _bool, ioWrite2D, (struct _IO __F* io, void __F* p, long row, unsigned short item, const void __F* colrow, const void __F* colcol, _unsigned2 variable, _unsigned2 byteorder ) );

_API( _bool, ioFill, ( struct _IO __F* io, void __F* buf, unsigned nbuf, unsigned long n, unsigned byteorder ) );
_API( char __F*, ioReadString, ( struct _IO __F* io, char __F* buf, int bufsize, unsigned byteorder ) );
_API( _bool, ioWriteString, ( struct _IO __F* io, const char __F* s, int len, unsigned byteorder ) );
_API( int, ioTransfer, ( struct _IO __F* dest, struct _IO __F* src, long totalsize, _integer1 __F* buf, int bufsize, int (__API* progress)(void __F* context,int size), void __F* context ) );
_API( int, ioLinearSearch, ( struct _IO __F* io, unsigned long n, unsigned size, const void __F* key, int (__API* compare)( struct _IO __F* io, const void __F* key, void __F* param, int __F* cl ), void __F* param, unsigned long __F* index ) ); /* XXX not implemented */
_API( int, ioBinarySearch, ( struct _IO __F* io, unsigned long n, unsigned size, const void __F* key, int (__API* compare)( struct _IO __F* io, const void __F* key, void __F* param, int __F* cl ), void __F* param, unsigned long __F* index ) );
_API( int, ioReadHierarchy, ( struct _IO __F* io, char __F*__F* excl, int nexcl, char __F*__F* current_key ) );
_API( int, ioEnterHierarchy, ( struct _IO __F* io, unsigned property, unsigned long __F* size, const char __F* key, struct _IO_HIERARCHY_STACK __F* stk ) );
_API( int, ioLeaveHierarchy, ( struct _IO __F* io, struct _IO_HIERARCHY_STACK __F* stk, unsigned byte_order ) );

_API( int, iotextRead, ( struct _IOBUFFER __F* iob, const char __F* delimiter, char __F* buf, int bufsize, unsigned flag, struct _IOTEXT __F* iot ) );
_API( int, iotextReadString, ( struct _IOBUFFER __F* iob, char __F* bufptr, int bufsize, char leadchar, char trailchar, char lastchar, unsigned flag, struct _IOTEXT __F* iot ) );
_API( int, iotextProbeString, ( struct _IOBUFFER __F* iob, const char __F* s, int len, char lastchar, struct _IOTEXT __F* iot ) );
_API( int, iotextReadStringExt, ( struct _IOBUFFER __F* iob, char* bufptr, int bufsize, const char __F* trail, int trail_len, unsigned flag, struct _IOTEXT __F* iot ) );
_API( int, iotextSleep, ( struct _IOBUFFER __F* iob, const char __F* delimiter, struct _IOTEXT __F* iot ) );
_API( int, iotextWait, ( struct _IOBUFFER __F* iob, const char __F* delimiter, struct _IOTEXT __F* iot ) );
_API( int, iotextObserve, ( struct _IOBUFFER __F* iob, const char __F* delimiter, struct _IOTEXT __F* iot ) );
_API( int, iotextSkipLine, ( struct _IOBUFFER __F* iob, unsigned long n, struct _IOTEXT __F* iot ) );
_API( unsigned, iotextConvert, ( struct _IO __F* dest, struct _IO __F* src, unsigned endline ) );
_API( unsigned, iotextFormat, ( struct _IO __F* src, long __F* readbytes ) );
_API( _bool, iotextWrite, ( struct _IO __F* io, const char __F* buf, unsigned endline ) );
_API( _bool, iotextWriteString, ( struct _IO __F* io, const char __F* s ) );



__CEND__



#endif /* __CORAL_IO_H */


