/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */



/* Coral Linked List */



#ifndef __CORAL_LIST_H
# define __CORAL_LIST_H

#include <coral/def.h>



/* Link node by index */
#define LIST_NEXT(l,prev) (((struct _LIST __F*__F*)(l))[prev])



/* List transformation : circular to linear */
#define LIST_BEGIN_LINEAR(head) ( (head)->prev->next = NULL )
#define LIST_END_LINEAR(head)   ( (head)->prev->next = head )



/* List transformation : linear to circular */
#define LIST_BEGIN_CIRCULAR LIST_END_LINEAR
#define LIST_END_CIRCULAR   LIST_BEGIN_LINEAR



/* listDNAdd */
#define LIST_FLAG_LINEAR    0x0000
#define LIST_FLAG_CIRCULAR  0x0002
#define LIST_FLAG_ADD_NEXT  0x0000 /* slightly faster than LIST_FLAG_ADD_PREV */
#define LIST_FLAG_ADD_PREV  0x0004
#define LIST_FLAG_QUEUE     0x0008 /* parameter 'where' must be head node or NULL and LIST_FLAG_CIRCULAR|LIST_FLAG_ADD_PREV must be specified */



/* listDNCreate, flags
 FLAG_CLEAR, clear created node
 LIST_FLAG_LINEAR
 LIST_FLAG_CIRCULAR
 LIST_FLAG_ADD_NEXT
 LIST_FLAG_ADD_PREV */



/* listInsertionNode */
#define LIST_FLAG_INVALID_INDEX  0x8000 /* the given index is invalid */


/* listFindBit */
#define LIST_FLAG_ALL 0x0001 /* all specified bits must be turned on */
#define LIST_FLAG_NOT 0x0002 /* not operation is performed */



/* List base class */
DECLARE_DLIST( _LIST )
DECLARE_END;



__CBEGIN__



/* memAlloc based list node allocation and deallocation */

_API( struct _LIST __F*, listPush, ( struct _LIST __F*__F* head, long size ) );
_API( void, listPop, ( struct _LIST __F*__F* head ) );
_API( struct _LIST __F*, listCreate, ( struct _LIST __F*__F* head, struct _LIST __F* where, long size ) );
_API( void, listRemove, ( struct _LIST __F*__F* head, struct _LIST __F* what, struct _LIST __F* prev ) );
_API( void, listClear, ( struct _LIST __F*__F* head ) );
_API( struct _LIST __F*, listDNCreate, ( struct _LIST __F*__F* head, struct _LIST __F* where, long size, unsigned flag ) );
_API( void, listDNRemove, ( struct _LIST __F*__F* head, struct _LIST __F* what ) );
_API( void, listCNClear, ( struct _LIST __F*__F* head ) );



/* add & cut operations */
_API( void, listAdd, ( struct _LIST __F*__F* head, struct _LIST __F* where, struct _LIST __F* what ) );
_API( void, listCut, ( struct _LIST __F*__F* head, struct _LIST __F* what, struct _LIST __F* prev ) );
_API( void, listDNAdd, ( struct _LIST __F*__F* head, struct _LIST __F* where, struct _LIST __F* what, unsigned flag ) );
_API( void, listDNCut, ( struct _LIST __F*__F* head, struct _LIST __F* what ) );


/* linear list iterations */

_API( unsigned long, listCount, ( struct _LIST __F* head ) );
_API( struct _LIST __F*, listNode, ( struct _LIST __F* head, long index ) );
_API( long, listIndex, ( struct _LIST __F* head, struct _LIST __F* node ) );
_API( struct _LIST __F*, listInsertionNode, ( struct _LIST __F* head, long index, unsigned __F* flag ) );
_API( _bool, listExist, ( struct _LIST __F* head, struct _LIST __F* node ) );
_API( struct _LIST __F*, listFindTail, ( struct _LIST __F* head ) );
_API( struct _LIST __F*, listFindPrev, ( struct _LIST __F* head, struct _LIST __F* node ) );
_API( struct _LIST __F*, listFindBit, ( struct _LIST __F* head, unsigned long mask, int offset, unsigned size, unsigned flag ) );
_API( struct _LIST __F*, listFindInt1, ( struct _LIST __F* head, _integer1 value, int offset ) );
_API( struct _LIST __F*, listFindInt2, ( struct _LIST __F* head, _integer2 value, int offset ) );
_API( struct _LIST __F*, listFindInt4, ( struct _LIST __F* head, _integer4 value, int offset ) );
_API( struct _LIST __F*, listFindInt, ( struct _LIST __F* head, int value, int offset ) );
_API( struct _LIST __F*, listFindLong, ( struct _LIST __F* head, long value, int offset ) );
_API( struct _LIST __F*, listFindStr, ( struct _LIST __F* head, const char __F* s, int offset, unsigned flag ) );
_API( struct _LIST __F*, listMaxInt1, ( struct _LIST __F* head, _integer1 __F* value, int offset ) );
_API( struct _LIST __F*, listMaxInt2, ( struct _LIST __F* head, _integer2 __F* value, int offset ) );
_API( struct _LIST __F*, listMaxInt4, ( struct _LIST __F* head, _integer4 __F* value, int offset ) );
_API( struct _LIST __F*, listMaxInt, ( struct _LIST __F* head, int __F* value, int offset ) );
_API( struct _LIST __F*, listMaxLong, ( struct _LIST __F* head, long __F* value, int offset ) );




/* circular list iterations */

_API( unsigned long, listCNCount, ( struct _LIST __F* head ) );
_API( struct _LIST __F*, listCNNode, ( struct _LIST __F* head, long index ) );
_API( long, listCNIndex, ( struct _LIST __F* head, struct _LIST __F* node ) );
_API( struct _LIST __F*, listCNInsertionNode, ( struct _LIST __F* head, long index, unsigned __F* flag ) );
_API( _bool, listCNExist, ( struct _LIST __F* head, struct _LIST __F* node ) );
_API( struct _LIST __F*, listCNFindBit, ( struct _LIST __F* head, unsigned long mask, int offset, unsigned size, unsigned flag ) );
_API( struct _LIST __F*, listCNFindInt1, ( struct _LIST __F* head, _integer1 value, int offset ) );
_API( struct _LIST __F*, listCNFindInt2, ( struct _LIST __F* head, _integer2 value, int offset ) );
_API( struct _LIST __F*, listCNFindInt4, ( struct _LIST  __F* head, _integer4 value, int offset ) );

_API( struct _LIST __F*, listCNFindInt, ( struct _LIST __F* head, int value, int offset ) );
_API( struct _LIST __F*, listCNFindLong, ( struct _LIST __F* head, long value, int offset ) );
_API( struct _LIST __F*, listCNFindStr, ( struct _LIST __F* head, const char __F* s, int offset, unsigned flag ) );
_API( struct _LIST __F*, listCNMaxInt1, ( struct _LIST __F* head, _integer1 __F* value, int offset ) );
_API( struct _LIST __F*, listCNMaxInt2, ( struct _LIST __F* head, _integer2 __F* value, int offset ) );
_API( struct _LIST __F*, listCNMaxInt4, ( struct _LIST __F* head, _integer4 __F* value, int offset ) );
_API( struct _LIST __F*, listCNMaxInt, ( struct _LIST __F* head, int __F* value, int offset ) );
_API( struct _LIST __F*, listCNMaxLong, ( struct _LIST __F* head, long __F* value, int offset ) );



/* resolve listFindInt, listFindLong, listMaxInt and listMaxLong */

#if __ARCH__<=32

# define listFindLong(head,value,offset) listFindInt4(head,(_integer4)(value),offset)
# define listMaxLong(head,value,offset) listMaxInt4(head,(_integer4 __F*)(value),offset)
# define listCNFindLong(head,value,offset) listCNFindInt4(head,(_integer4)(value),offset)
# define listCNMaxLong(head,value,offset) listCNMaxInt4(head,(_integer4 __F*)(value),offset)

# if __ARCH__<=16
#  define listFindInt(head,value,offset) listFindInt2(head,(_integer2)(value),offset)
#  define listMaxInt(head,value,offset) listMaxInt2(head,(_integer2 __F*)(value),offset)
#  define listCNFindInt(head,value,offset) listCNFindInt2(head,(_integer2)(value),offset)
#  define listCNMaxInt(head,value,offset) listCNMaxInt2(head,(_integer2 __F*)(value),offset)
# else /* __ARCH__>16 */
#  define listFindInt listFindLong
#  define listMaxInt listMaxLong
#  define listCNFindInt listCNFindLong
#  define listCNMaxInt listCNMaxLong
# endif /* __ARCH__<=16 */

#else /* __ARCH__>32 */

#if __ARCH__==64 /* for lp64 model */
# define listFindInt(head,value,offset) listFindInt4(head,(_integer4)(value),offset)
# define listMaxInt(head,value,offset) listMaxInt4(head,(_integer4 __F*)(value),offset)
# define listCNFindInt(head,value,offset) listCNFindInt4(head,(_integer4)(value),offset)
# define listCNMaxInt(head,value,offset) listCNMaxInt4(head,(_integer4 __F*)(value),offset)
#else
... /* #error unknown architecture */
#endif

#endif /* __ARCH__<=32 */



#define listFindChar(head,value,offset) listFindInt1(head,(_integer1)(value),offset)
#define listFindShort(head,value,offset) listFindInt2(head,(_integer2)(value),offset)
#define listFindPtr(head,ptr,offset) listFindLong(head,(long)ptr,offset)
#define listMaxChar(head,value,offset) listMaxInt1(head,(_integer1 __F*)(value),offset)
#define listMaxShort(head,value,offset) listMaxInt2(head,(_integer2 __F*)(value),offset)

#define listCNFindChar(head,value,offset) listCNFindInt1(head,(_integer1)(value),offset)
#define listCNFindShort(head,value,offset) listCNFindInt2(head,(_integer2)(value),offset)
#define listCNFindPtr(head,ptr,offset) listCNFindLong(head,(long)ptr,offset)
#define listCNMaxChar(head,value,offset) listCNMaxInt1(head,(_integer1 __F*)(value),offset)
#define listCNMaxShort(head,value,offset) lisCNtMaxInt2(head,(_integer2 __F*)(value),offset)



__CEND__



#endif /* __CORAL_LIST_H */
