/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Coral Memory Operations */



#ifndef __CORAL_MEMORY_H
# define __CORAL_MEMORY_H



#include <coral/def.h>


typedef unsigned long _memmagic;


DECLARE_DLIST( _MEMDEBUG )
	_memmagic     magic; /* the magic number */
	unsigned long size;  /* allocated size */
DECLARE_END;



#ifndef __INTERNAL__
struct _SHMCONTEXT { int dummy; };
#endif



/* memByteOrder */
#define MEM_INVERT 1 /* inversion is needed */
#define MEM_X1     2
#define MEM_X2     3

/* memReorder
 FLAG_DEFAULT
 FLAG_EXCLUSIVE
 FLAG_BIGENDIAN
 FLAG_LITTLEENDIAN */



#define memFreeRect memFree2D



/* memSourceTable */
#define MEM_SPECIAL_CHAR 0xff /* value to specify the special character */



/* memConvert, return values
	FAIL_UNAVAILABLE : source length is zero or insufficient source */
#define MEM_SPECIAL_CHAR_FOUND  0x8000 /* special character is found, special character can be retrieved by masking low byte of the return value */



/* memConvert, flags */
#define MEM_FLAG_ALLOW_REMAINED     0x0100 /* source bits can be remained */
#define MEM_FLAG_ALIGN              0x0200 /* align result by appending the pad character which is combinated in low byte of the flag */
#define MEM_FLAG_SPECIAL_CHAR_INDEX 0x0400 /* store the index where the special character was found into len[1] */
#define MEM_FLAG_LOW_TO_HIGH        0x0800 /* transfer bits from low bit to high bit */



/*  cl : classification code
    m : middle position */


#define MEM_BEGIN_BSEARCH(n) \
	if( n>0 ) {\
		unsigned l=0, m=0;\
		n--;

#define MEM_ENTER_BSEARCH_LOOP(n) \
		while( l<=n ) {\
			m = (l+n)>>1; \
			{

#define MEM_GET_CURRENT_BSEARCH_ELEM(p,size) ( (char __F*)(p) + (m)*(size) )

#define MEM_BSEARCH_LESS          { l = m + 1; }
#define MEM_BSEARCH_GREATER(n)    { if( m<=0 ) break; n = m - 1; }

#define MEM_LEAVE_BSEARCH_LOOP \
			}\
		}

#define MEM_END_BSEARCH \
	}



__CBEGIN__



_API( struct _MEMDEBUG __F*, memGetDebug, ( struct _MEMDEBUG __F* current ) );
_API( _memmagic, memGetMagic, ( void ) );
_API( void, memSetMagic, ( _memmagic __F* table, unsigned n, void (__API* callback)( _memmagic found ) ) );

_API( void __F*, memAlloc, ( long size ) );
_API( void __F*, memResize, ( void __F* p, long size ) );
_API( void __F*, memCAlloc, ( long size ) );
_API( void, memFree, ( void __F* p ) );
_API( unsigned, memGetExtraSize, ( unsigned extra ) );
_API( void __F*, memAllocExtra, ( long size, unsigned extra ) );
_API( void __F*, memResizeExtra, ( void __F* p, long size ) );
_API( void, memFreeExtra, ( void __F* p ) );
_API( void __F*, memGetExtra, ( void __F* p ) );
_API( void __F*, memAllocName, ( long size, const char __F* name, int __F* status ) );
_API( void, memFreeName, ( const char __F* name ) );
_API( void __F*, memGetNamePtr, ( const char __F* name ) );
_API( const char __F*, memGetName, ( void __F* p ) );
_API( void, memSet, ( void __F* p, _unsigned1 val, long size ) );
_API( void, memSetExt, ( void __F* p, unsigned long n, unsigned size, unsigned __F* param ) ); 
_API( void, memCopy, ( void __F* dest, const void __F* src, long size ) );
_API( void, memConvCopy, ( void __F* dest, void __F* src, void (__API* convert) P_A(( void __F* dest, void __F* src )), long n, unsigned destsize, unsigned srcsize ) );
_API( void, memInvert, ( void __F* p, unsigned size ) );
_API( void, memInvertB, ( void __F* p, long n, unsigned size ) );
_API( void, memInvertA, ( void __F* p, void __F* buf, long n, unsigned size ) );
_API( void, memSourceTable, ( char __F* src_table, const char __F* dest_table, unsigned n ) );
_API( unsigned, memSourceAlignSize, ( unsigned src_nbit ) );
_API( unsigned, memDestBufferSize, ( unsigned dest_nbit, unsigned src_nbit, unsigned len ) );
_API( int, memConvert, ( unsigned char __F* dest, const unsigned char __F* src, unsigned __F* len, const unsigned char __F* dest_table, const unsigned char __F* src_table, unsigned dest_nbit, unsigned src_nbit, unsigned flag ) );
_API( long, memAlign, ( long size, int boundary ) );
_API( unsigned, memByteOrder, ( void ) );
_API( unsigned, memByteOrderHint, ( unsigned flag ) );
_API( _bool, memReorder, ( void __F* p, long n, unsigned byteorder ) );



/*XXX memAllocCol2D and tests are needed */
_API( long, memSize2D, ( long row, unsigned size, const void __F* colrow, const void __F* colcol, unsigned variable ) );
_API( void, memInit2D, ( void __F* p, void __F* column, long row, unsigned size, const void __F* colrow, const void __F* colcol, unsigned variable ) );
_API( void, memInitRect, ( void __F* p, void __F* column, long row, long colsize ) );
_API( void __F*, memAlloc2D, ( long row, unsigned size, const void __F* colrow, const void __F* colcol, unsigned variable ) );
_API( void __F*, memAllocRect, ( long row, long colsize ) );
_API( void, memCopy2D, ( void __F* dest, void __F* src, long row, unsigned size, const void __F* colrow, const void __F* colcol, unsigned variable ) );
_API( void, memSet2D, ( void __F* dest, _unsigned1 val, long row, unsigned size, const void __F* colrow, const void __F* colcol, unsigned variable ) );
_API( void, memFreeCol2D, ( void __F* p ) );
_API( void, memResetCol2D, ( void __F* p ) );
_API( void, memFree2D, ( void __F* p ) );
_API( void, memReset2D, ( void __F*__F* p ) );
_API( void __F*, memAlloc2DExt, ( long row, unsigned size, const void __F* colrow, const void __F* colcol, unsigned variable ) );
_API( void, memFreeCol2DExt, ( void __F* p, long row ) );
_API( void, memResetCol2DExt, ( void __F* p, long row ) );
_API( void, memFree2DExt, ( void __F* p, long row ) );
_API( void, memReset2DExt, ( void __F*__F* p, long row ) );


_API( unsigned, memCountIdentity, ( const void __F* p, unsigned n, unsigned width, long value ) );
_API( unsigned, memCountIdentityExt, ( const void __F* key, const void __F* p, unsigned n, unsigned width, int (__API* compare) P_A(( const void __F* a, const void __F* b )) ) );
_API( void __F*, memLinearSearch, ( const void __F* p, unsigned n, unsigned size, const void __F* key, int (__API* compare) P_A(( const void __F* p, const void __F* key )), unsigned __F* index ) );
_API( void __F*, memBinarySearch, ( const void __F* p, unsigned n, unsigned size, const void __F* key, int (__API* compare) P_A(( const void __F* p, const void __F* key )), unsigned __F* index ) );



_API( void, memReduceInt, ( long __F* value, unsigned size ) );
_API( long, memRestoreInt, ( void __F* value, unsigned size ) );


_API( _bool, memOpen, ( struct _IO __F* io, long initial, long extend, long limit, void __F*__F* p ) );
_API( void, memClose, ( struct _IO __F* io ) );
_API( void, memTruncate, ( struct _IO __F* io, long newsize ) );
_API( long, memRead, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, memWrite, ( struct _IO __F* io, void __F* p, long size ) );
_API( long, memSeek, ( struct _IO __F* io, long offset, int seekpos ) );


/*XXX prototype is incomplete */
#define MEM_FLAG_
_API( struct _SHMCONTEXT __F*, shmCreateContext, ( void ) );
_API( void, shmRemoveContext, ( struct _SHMCONTEXT __F* shmc ) );
_API( void, shmBegin, ( struct _SHMCONTEXT __F* shmc ) );
_API( void, shmEnd, ( struct _SHMCONTEXT __F* shmc ) );
_API( void __F*, shmAlloc, ( struct _SHMCONTEXT __F* shmc, const char __F* name, unsigned long size, unsigned flag ) );
_API( void __F*, shmOpen, ( struct _SHMCONTEXT __F* shmc, const char __F* name ) );
_API( unsigned long, shmGetSize, ( struct _SHMCONTEXT __F* shmc, void __F* shm ) );
_API( void, shmFree, ( struct _SHMCONTEXT __F* shmc, void __F* p ) );
_API( void , shmInstallCallback, ( struct _SHMCONTEXT __F* shmc, void (__API* callback)( const char __F* name, unsigned action, void __F* param ), void __F* param ) );
_API( void, shmUninstallCallback, ( struct _SHMCONTEXT __F* shmc, void (__API* callback)( const char __F* name, unsigned action, void __F* param ) ) );



__CEND__



#include <coral/compat/memory.h>



#ifdef __WIN16__

# ifndef WINVER
  extern void __API hmemcpy P_A(( void _huge* dest, const void _huge* src, long size ));
# endif
# define mem_short_copy hmemcpy
# define mem_short_set  _fmemset
# define mem_short_comp _fmemcmp
# define memCopy        hmemcpy

#else /* !__WIN16 */
#if __ARCH__<=16

# define  mem_short_copy _fmemcpy
# define  mem_short_set  _fmemset
# define  mem_short_comp _fmemcmp

#else /* __ARCH__>16 */

# define mem_short_copy memcpy
# define mem_short_set  memset
# define mem_short_comp memcmp
# define memCopy        memcpy
# define memSet         memset

#endif /* __ARCH__<=16 */
#endif /* __WIN16__ */



#endif /* __CORAL_MEMORY_H */

