/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Operation ( Undo & Redo ) log */


#ifndef __CORAL_OPLOG_H
# define __CORAL_OPLOG_H



#include <coral/def.h>



#ifndef __INTERNAL__
	DECLARE_CONTEXT( OPLOGCONTEXT )
#endif



#define DECLARE_OPLOG(name) DECLARE_DLIST(name) unsigned flag; union _OPLOGCODE code;



/* _OPLOGCODE */

union _OPLOGCODE {
	char __F* s;
	unsigned  u;
};



/* _OPLOG */

DECLARE_OPLOG ( _OPLOG )
DECLARE_END;





/* modes */

#define OPLOG_UNDO 0
#define OPLOG_REDO 1



/*--- Action codes for op log callback function ---*/

/* Common return values for callback function
	0, success
	negative value, error code */

#define OPLOG_EXECUTE 0  /* execute op log */

struct _OPLOG_EXECUTE_PARAM {
	struct _OPLOG __F* oplog; /* op log to be executed */
};

/* return value for OPLOG_EXECUTE */
#	define OPLOG_FLAG_RESERVE  0x0001  /* reserve the current log */
#	define OPLOG_FLAG_SKIP     0x0002  /* XXX */



#define OPLOG_CLEAR  1  /* clear resources allocated by the given log */

struct _OPLOG_CLEAR_PARAM {
	struct _OPLOG __F* oplog; /* op log to be cleared */
};

/* return value for OPLOG_CLEAR
	OPLOG_FLAG_RESERVE, reserve the current log
	OPLOG_FLAG_SKIP, skip the current log */




#define OPLOG_CLEAR_IO 2 /* clear resources allocated by the given log io */

struct _OPLOG_CLEAR_IO_PARAM {
	_DIOHANDLE       iolog; /* op log to be cleared */
};



#define OPLOG_WRITE   3  /* write log */
/* getting serialized size - exact, extendible */


#define OPLOG_READ    4  /* read log */



union _OPLOG_PARAM {
	struct _OPLOG_EXECUTE_PARAM execute;
	struct _OPLOG_CLEAR_PARAM  remove;
};



typedef int (__API* OPLOGCALLBACK) P_A(( OPLOGCONTEXT oplc, unsigned action, union _OPLOG_PARAM __F* param ));



__CBEGIN__



_API( OPLOGCONTEXT, oplogCreate, ( struct _DIO __F* dio, OPLOGCALLBACK callback, void __F* param ) );
_API( void, oplogRemove, ( OPLOGCONTEXT oplc ) );
_API( void, oplogClear, ( OPLOGCONTEXT oplc, unsigned mode ) );
_API( void, oplogSetState, ( OPLOGCONTEXT oplc, unsigned mode, _bool enable ) );
_API( void, oplogSetCount, ( OPLOGCONTEXT oplc, unsigned cnt ) );
_API( struct _OPLOG __F*, oplogGetRoot, ( OPLOGCONTEXT oplc, unsigned mode ) );
_API( void __F*, oplogGetParam, ( OPLOGCONTEXT oplc ) );
_API( struct _OPLOG __F*, oplogAdd, ( OPLOGCONTEXT oplc, unsigned mode, unsigned size, unsigned ucode, const char __F* scode ) );
_API( void, oplogBegin, ( OPLOGCONTEXT oplc ) );
_API( void, oplogEnd, ( OPLOGCONTEXT oplc ) );
_API( int, oplogExecute, ( OPLOGCONTEXT oplc, unsigned mode ) );
_API( int, oplogReserve, ( OPLOGCONTEXT oplc, unsigned mode ) );



__CEND__



#endif /* __CORAL_OPLOG_H */

