/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* String */


#ifndef __CORAL_STRING_H
# define __CORAL_STRING_H


#include <coral/def.h>



#define strGetDirectorySeparatorSize() 1
#define STR_WHITE_CHAR " \t\r\n"



/* strGetRelativePath */
#define STR_FLAG_ALLOCATE       0x0004 /* allocate the result string */



/* strToken */
#define STR_FLAG_EXCLUSIVE       0x0001 /* test delimiters exclusively */
#define STR_FLAG_FOUND_DELIMITER 0x0002 /* the pointer where the found delimiter is placed is returned and token will not be null terminated */ 
#define STR_FLAG_CHAR_ESC        0x0004 /* use the first character in the specified delimiter as the character escape sequence */


/* strBackwardToken */
/* STR_FLAG_EXCLUSIVE, test delimiters exclusively
   STR_FLAG_FOUND_TOKEN, the found delimiter is reserved and token string will be returned */
#define STR_FLAG_END_OF_STRING 0x0010 /* the end of string pointer and exact string length are supplied */


/* strTokenIndex */
/* STR_FLAG_EXCLUSIVE */


/* strBackwardTokenIndex */
/* STR_FLAG_EXCLUSIVE, STR_FLAG_END_OF_STRING */


/* strFind */
/* FLAG_CASE_INSENSITIVE */


/* stroptOption */

/* option flags */
#define STROPT_FLAG_COMBINATION       0x0010 /* support combinated option, i.e -x -z can be -xz */
#define STROPT_FLAG_ALWAYS            0x0020 /* always parse options, i.e -xz can be xz */
#define STROPT_FLAG_FIRST             0x0040 /* alway parse options only for the first argument, used with STR_FLAG_OPTION_ALWAYS */
#define STROPT_FLAG_DEFAULT_INDICATOR 0x0080 /* use default indicator '-' */
#define STROPT_FLAG_DEFAULT_DELIMITER 0x0100 /* use default delimiter '=' */
#define STROPT_FLAG_DOUBLE_INDICATOR  0x0200 /* assume string option if double indicator found, otherwise character option */
#define STROPT_FLAG_INDICATOR_ONLY    0x0400 /* if only an indicator was found, return STR_OPTION_CHARACTER|0 ( == 0 ) instead of STR_OPTION_UNKNOWN */
#define STROPT_FLAG_ALLOW_BLANK       0x0800 /* allow blank between string option and indicator or indicator with appended string */
#define STROPT_FLAG_BINARY_SEARCH     0x1000 /* the given string table is sorted so binary search can be used - 'delimiter' must not be NULL or STROPT_FLAG_DEFAULT_DELIMITER flag should be supplied */

/* internally used flags */
#define STROPT_FLAG_BLANK_ALLOWED   0x2000 /* blank was allowed by STR_OPTION_FLAG_ALLOW_BLANK flag */
#define STROPT_FLAG_ENDED           0x4000 /* there is no more string option : internally used when strOptionLinear is used */
#define STROPT_FLAG_COMBINATED      0x8000 /* character option was combinated : internally used */

/* return values */
#define STROPT_CHARACTER     0x00000000L /* a character option was found, low word is the found character */
#define STROPT_STRING        0x10000000L /* a string option was found, low word is the found index */
#define STROPT_RESERVED      0x20000000L /* current argument is not an option */
#define STROPT_IGNORED       0x30000000L /* parsing is ignored ( an option indicator appended after the space is found ), low word is found indicator */
#define STROPT_TERMINATED    0x40000000L /* argument array was terminated */
#define STROPT_UNKNOWN       0x50000000L /* unknown option was found */
#define STROPT_INVALID_PARAM 0x60000000L /* invalid parameter for the string option was found */


struct _STROPT {
	/* following members need to be initialized externally */
	const char __F* (__API* getarg) P_A(( struct _STROPT __F* stropt )); /* a routine to get the next argument */
	unsigned  flag;    /* option flags */
	void __F* context; /* arbitrary context for getarg function */
	const char __F* append;  /* appended string of the current argument, it should be initially NULL and it mustn't be manually modified */
	const char __F* nextarg; /* the next argument, initially NULL and it mustn't be manually modified */

	/* the following member must not be modified externally */
	const char __F* arg;     /* the current argument */
};



struct _STROPT_LONG {
	const char __F* name;  /* long option name */
	int             value; /* long option value */
};



__CBEGIN__



_API( int, strCompare, ( const char __F* a, const char __F* b, unsigned flag ) );
_API( char __F*, strCopy, ( char __F* dest, const char __F* src, int size ) );
_API( char __F*, strUnion, ( char __F* dest, const char __F* src1, const char __F* src2, int size ) );
_API( char __F*, strAppend, ( char __F* dest, const char __F* src, int size, int __F* resize ) );
_API( char __F*, strInsert, ( char __F* dest, const char __F* src, int i, int size, int __F* resize ) );
_API( unsigned, strGetCompatiblePathLength, ( const char __F* path ) );
_API( char __F*, strGetCompatiblePath, ( char __F* path ) );
_API( int, strStoreDirectorySeparator, ( char __F* name, int index ) ); 
_API( int, strAppendDirectorySeparator, ( char __F* name ) );
_API( const char __F*, strGetDirectorySeparator, ( void ) );
_API( const char __F*, strGetFileExtension, ( const char __F* path ));
_API( char __F*, strReplaceFileExtension, ( char __F* path, const char __F* ext ) );
_API( const char __F*, strGetFileName, ( const char __F* path ));
_API( char __F*, strGetPathName, ( char __F* path ) );
_API( void, strSplitPath, ( char __F*__F* dir, char __F*__F* name, char __F*__F* ext, char __F* path ) );
_API( _bool, strAbsolutePath, ( const char __F* path ) );
_API( char __F*, strGetRelativePath, ( const char __F* base, char __F* path, unsigned flag ) );
_API( char __F*, strGetAbsolutePath, ( const char __F* base, char __F* path, unsigned flag ) );
_API( char __F*, strCanonicalPath, ( char __F* result, const char __F* current, char __F* relative ) );

_API( char __F*, strToken, ( char __F* str, const char __F* delimiter, unsigned flag ) );
_API( char __F*, strBackwardToken, ( char __F* s, int len, const char __F* delimiter, unsigned flag ) );
_API( int, strTokenIndex, ( const char __F* s, const char __F* delimiter, unsigned flag ) );
_API( int, strBackwardTokenIndex, ( const char __F* s, int len, const char __F* delimiter, unsigned flag ) );
_API( char __F*, strReplaceChar, ( char __F* str, const char __F* oldchr, const char __F* newchr, int n ) );
_API( unsigned, strCountChar, ( const char __F* str, const char __F* chr ) );
_API( char __F*, strReduceChar, ( char __F* str, const char __F* chr ) );
_API( char __F*, strReduceZero, ( char __F* s, unsigned level ) );
_API( int, strFind, ( const char __F* s, const char __F* key, unsigned flag ) );

_API( const char __F*, stroptArgv, ( struct _STROPT __F* stropt ) );
_API( _unsigned4, stroptGetOption, ( struct _STROPT __F* stropt, const char __F* indicator, const char __F* delimiter, const struct _STROPT_LONG __F* lopt, unsigned nlopt ) );
_API( const char __F*, stroptGetString, ( struct _STROPT __F* stropt ) );



__CEND__



#endif /* __CORAL_STRING_H */

