/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */



#ifndef __CORAL_TYPE_H
#define __CORAL_TYPE_H



/* TYPE CODES */


/* type bits */

#define TC_FLAG_INTEGER  0x1000  /* integer value */
#define TC_FLAG_UNSIGNED 0x2000  /* unsigned value */
#define TC_FLAG_REAL     0x4000  /* real value */
#define TC_FLAG_CLASS    0x8000  /* the kind of class */



/* types codes */

#define TC_VOID       0x0000  /* void */
#define TC_STRING     0x0001  /* string */
#define TC_FUNCTION   0x0002  /* function */
#define TC_CLASS      0x8003  /* class */
#define TC_STRUCTURE  0x8004  /* structure */
#define TC_UNION      0x8005  /* union */
#define TC_DEFINITION 0x0006  /* definition */
#define TC_I1         0x1107  /* 1 byte signed integer */
#define TC_I2         0x1208  /* 2 bytes signed integer */
#define TC_I4         0x1409  /* 4 bytes signed integer */
#define TC_I8         0x180a  /* 8 bytes signed integer */
#define TC_U1         0x310b  /* 1 byte unsigned integer */
#define TC_U2         0x320c  /* 2 bytes unsigned integer */
#define TC_U4         0x340d  /* 4 bytes unsigned integer */
#define TC_U8         0x380e  /* 8 bytes unsigned integer */
#define TC_R4         0x440f  /* 4 bytes single precision floating point */
#define TC_R8         0x4810  /* 8 bytes double precision floating point */



/* type masks */

#define TC_MASK_ID       0x00ff  /* mask to get type identifier */
#define TC_MASK_SIZE     0x0f00  /* integral size */
#define TC_SHIFT_SIZE    8       /* number of bits to be shifted to get integral size */
#define TC_MASK_FLAG     0x7000  /* type mask to get value flags */



/* type property */

#define TP_POINTER    0
#define TP_ARRAY      1 /* modifier is number of elements */



/* _TYPE_PROPERTY_ELEM */

struct _TYPE_PROPERTY_ELEM {
	unsigned      code; /* property code */
	unsigned long mod;  /* modifier */
};



/* _TYPE_PROPERTY */

struct _TYPE_PROPERTY {
	int nelem;   /* number of elements */
	struct _TYPE_PROPERTY_ELEM elem[1]; /* elements */
};



/* _TYPE_BASE */

union _TYPE_BASE {
	struct _TYPE_CLASS __F*   cl;  /* if _TYPE_SYMBOL::code is TC_CLASS,  this member if valid */
	struct _TYPE_SYMBOL __F*  symbol; /* if _TYPE_SYMBOL::code is TC_DEFINITION, this member is valid */
};



/* _TYPE_SYMBOL */

struct _TYPE_SYMBOL {
	char __F*  name; /* symbol name */
	unsigned  code;  /* type code */
	unsigned  mod;   /* modifier */
	union _TYPE_BASE base;  /* base type */
	struct _TYPE_PROPERTY __F* prop;  /* type property */
	void __F* param; /* custom parameter */
};



#define TM_FLAG_CONSTANT 0x0001 /* type modifier flag */



/* _TYPE_CLASS_MEMBER */

struct _TYPE_CLASS_MEMBER {
	int nelem;
	struct _TYPE_SYMBOL elem[1];
};



/* _TYPE_PARENT_CLASS */

struct _TYPE_PARENT_CLASS {
	int nelem;
	struct _TYPE_CLASS __F* elem[1];
};



/* _TYPE_CLASS */

struct _TYPE_CLASS {
	char __F*  name;  /* name of this class */
	unsigned   code;  /* type code to specify class, struct, or union */
	struct _TYPE_CLASS_MEMBER __F* member; /* members */
	struct _TYPE_PARENT_CLASS __F* parent; /* parent classes */
};



/* typeTraverseClass, actions */
#define TYPE_CLASS_MEMBER 0 /* a member of a class */
#define TYPE_TRACE_CLASS  1 /* tracing class */



__CBEGIN__



/* XXX incomplete function */
_API( int, typeTraverseClass, ( struct _TYPE_CLASS __F* cl, int (__API* traverse)( struct _TYPE_CLASS __F* cl, struct _TYPE_SYMBOL __F* symbol, unsigned action, void __F* param ), void __F* param ) );



__CEND__



#endif /* __CORAL_TYPE_H */

