/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Routines for the Compatibility

 1997.01.   File I/O routines
 1997.02    Memory routines
 1997.10.   Buffered I/O ( currently moved to open I/O buffer API )
 1998.02.   I/O routines are reviewed
 1999.08.12 Module creation
 2000.05.15 POSIX opendir compatibility routines
 2000.07.19 The compatible routine for getcwd extension

 mem_win_iter, win16
 _memCopy, dos16
 _memSet, dos16, win16

 Compile switches :
 __ALIGNED__, memBestAlign returns the result of memAlign
 __PACKED__, memBestAlign returns the given size */


/* XXX : THIS SOURCE CODES SHOULD BE CLEANED :-( */


/*
#define _INCLUDE_XOPEN_SOURCE
#define _INCLUDE_HPUX_SOURCE */

#ifdef __hpux
#define _HPUX_SOURCE
#endif

#include <errno.h>
#include <coralsrc/memory.h>
#include <coralsrc/string.h>
#include <coralsrc/io.h>
#include <coral/inline.h>
#include <coral/compat/unistd.h>


#if !defined(__MACOS__) && !defined(__PALMOS__)

//#define __COMPAT__
//#include <coral/compat/unistd.h>
//#include <coral/ext/thread.h>


/*??? 'open' standard campaign */
//# include <fcntl.h>
//# include <coral/compat/stat.h>
/*???*/

#ifdef __WINDOWS__
#	include <windows.h>
#endif


#ifdef _MSC_VER
/*??? it is temporary... fileTrace will integrate them */
struct dirent {
  char __F* d_name;
};

typedef struct
{
 char* path;
 struct dirent   dirent; /* POSIX dirent object */
 WIN32_FIND_DATA find;   /* finddata */
 HANDLE          handle; /* win32 search handle */
} DIR;
#	define WIN32_DIR_DEFINED
#else /* !_MSC_VER */
#	include <fcntl.h>
#	include <coral/compat/stat.h>
#endif /* _MSC_VER */

#include <coral/compat/dirent.h>


/*                 *\

   Memory Routines

\*                 */



#if __ARCH__<=16 && !defined(__TURBOC__)



__MODULEBEGIN__( "alloc.c" )
/* memAlloc */

void __F* __API memAlloc
P_B(( size ))
P_1  long size
P_E
{

}

__MODULEEND__



__MODULEBEGIN__( "free.c" )
/* memFree */

void __API memFree


__MODULEEND__



__MODULEBEGIN__( "ralloc.c" )
/* memReAlloc */




#endif /* __ARCH__<=16 && !defined(__TURBOC__) */






#ifdef XX
/* memBestAlign

 This method aligns the given size as the best for the running platform */

__MODULEBEGIN__( "mbstalgn.c" )

//#include <signal.h>
//#include <offset.

#ifdef __TURBOC__
# pragma argsused
#endif

long __API memBestAlign
P_B(( size, mrb ))
P_1  long size
P_N  int mrb
{
 /* though gnu c/c++ supports __alignof__, following routine can be applied
  for all c compilers */
 struct _mrb8
 {
  char a;
  double b;
 };

 if( OP_OFFSET(struct _mrb8,b)==1 || mrb<=1 )
 {
  /* alignment is not needed so return the best size */
  return size;
 }
 else if( 


long double or long long or __int64
 }

#if defined(__PACKED__)

 return size;

#else /* !__PACKED__ */
#ifdef __GNUC__ /* gnu c specific */

 /* if __alignof__(double) is less than or equal to 4, the most restrictive
  boundary of the used gcc compiler is 4. evidently, __alignof__(double)<=4
  can be evaluated at compile time but __alignof__ cannot be used in the
  preprocessor line */

/* actually, several compilers like gcc which support multiple platforms
 can define SIGBUS though the current running platform doesn't have bus
 error exception, but typical alignment rules are applied */
 return
  memAlign( size,
   mrb>4 && __alignof__(double)<=4 ? 4 : mrb );

4  ==4  4
   <=4

#else /* !__GNUC__ */
#ifdef __ALIGNED__

 return memAlign( size, mrb );

#else /* !__ALIGNED__ */
#endif /* __ALIGNED__ */

#endif /* __GNUC__ */
#endif /* __PACKED__ */
}
__MODULEEND__
#endif



/*              *\

   I/O Routines

\*              */



__MODULEBEGIN__( "ftrunc.c" )
/* fileTruncate

 Using WIN32 kernel function :
  dos or unix file truncation functions don't modify the current
 file position but win32 need to do that because SetEndOfFile
 makes current file position to the end of a file. If this
 incompatibility can cause some problems for the common case,
 win32 api cannot be used.

 chsize implementation of MSVC++ 5.0, chsize.c ( 2000.08.01 )
  - reserve the current offset
  - set offset to the end of file which will have the given new size
  - use kernel function SetFilePointer to truncate size or
    append null bytes to extend size
  - recover the current offset

 ??? ftruncate implementation of gcc need to be considered
*/

_bool __API fileTruncate
P_B(( io, newsize ))
P_1  struct _IO __F* io
P_N  long newsize
P_E
{
#ifdef __WIN32__

 return SetFilePointer
  ( (HANDLE)io->handle, newsize, NULL, IO_SEEK_BEGIN ) ?
  SetEndOfFile( (HANDLE)io->handle ) : 0;

#else /* !__WIN32__ */

 return
# ifdef __UNIXC__
  ftruncate
# else
  chsize
# endif
   ( (int)io->handle, newsize )==FAIL_IO ? 0 : 1;

#endif /* __WIN32__ */
}
__MODULEEND__







/*                                      *\

   POSIX opendir compatibility routines

\*                                      */



#if !( defined(__UNIXC__) || ( defined(__TURBOC__) && __TURBOC__>=0x200 ) )



#ifdef __WATCOMC__



/* compat_opendir

  compat_opendir is needed for watcom c because a file can
 be opened by its opendir method */

DIR __F* __API compat_opendir
P_B(( name ))
P_1  const char __F* name
P_E
{
 struct find_t find;

 /* pass only for a directory */
 return
   _dos_findfirst( name, _A_SUBDIR, &find )==0 &&
   (find.attrib&_A_SUBDIR) ?
  opendir( name ) : NULL;

}



#else /* __WATCOMC__ */



/* compat_opendir */

DIR __F* __API compat_opendir
P_B(( name ))
P_1  const char __F* name
P_E
{
 DIR __F* dir;

 /* check given path is a directory */
 #if defined(__WIN32__)
 _unsigned4 attr = GetFileAttributes(name);
 if( attr!=0xffffffff && (attr&FILE_ATTRIBUTE_DIRECTORY) )
 {
 #else

 struct _find_t find;
 int length = strlen(name);

 /* current or root directory of a drive was given */
 if( ( length==3 && name[1]==':' && (name[2]=='\\'||name[2]=='/') ) ||
     ( length==2 && name[1]==':' ) )
 {
  char buf[256];

# ifdef __TURBOC__
  if( getcurdir(name[0]-'a'+1,buf)==0 )
# else
  if( _getdcwd(name[0]-'a'+1,buf,sizeof(buf))!=NULL )
# endif
   goto progress;

  return NULL;
 }

 if( _dos_findfirst( name, _A_SUBDIR, &find )==0 &&
     ( _find_t_attrib(find)&_A_SUBDIR ) )
 {
  progress:
 #endif
  dir = (DIR __F*) memCAlloc( sizeof(DIR) );
  if( dir )
  {
   dir->path = (char*)memAlloc( strlen(name)+5 );
   if( dir->path )
   {
    /* just link to the real buffer */
    dir->dirent.d_name = _find_t_name(dir->find);
    strcpy( dir->path, name );
    strAppendDirectorySeparator( dir->path );
    strcat( dir->path, "*.*" );
    return dir;
   }
   else memFree( dir );
  }
 }
 return NULL;
}



/* compat_closedir */


/* compat_readdir */




void __API compat_closedir
P_B(( dir ))
P_1  DIR __F* dir
P_E
{
 if( dir )
 {
# ifdef __WIN32__

  /* bug fixed on 2000.08.22, immediate call of closedir after calling
   opendir will make dir->fh to NULL and FindClose posts warning
   message for the chance of access violation in debug mode vc++ */
  if( dir->handle )
   FindClose( dir->handle );
# endif /* __WIN32__ */

  memFree( dir->path );
  memFree( dir );
 }
}

struct dirent __F* __API compat_readdir
P_B(( dir ))
P_1  DIR __F* dir
P_E
{
 /* if indication flag was not initialized, the first execution */
 if( !dir->handle )
 {
# ifdef __WIN32__

  if( (dir->handle=FindFirstFile
   (dir->path,&dir->find))==INVALID_HANDLE_VALUE )

# else

  dir->handle = 1;
  if( _dos_findfirst( dir->path,
       _A_RDONLY|_A_HIDDEN|_A_SYSTEM|_A_SUBDIR|_A_ARCH,
       &dir->find )!=0 )

# endif

  {
   return NULL;
  }
 }
 else
 {
# ifdef __WIN32__
  if( !FindNextFile( dir->handle, &dir->find ) )
# else
  if( _dos_findnext( &dir->find )!=0 )
# endif

  {
   return NULL;
  }
 }

 return &dir->dirent;
}



#endif /* __WATCOMC__ */



#endif /* COMPAT_DIRENT */



/*                 *\

   unistd routines

\*                 */



#ifdef compat_getcwd



__MODULEBEGIN__( "getcwd.c" )
/* compat_getcwd */

char* __API compat_getcwd
P_B(( buf, size ))
P_1  char* buf
P_N  int size
P_E
{
 /* test the current coral/compat/unistd.h couldn't detect getcwd extension
  supported compiler correctly */
 char* cwd = getcwd(buf,size);
 if( !cwd && buf==NULL )
 {

  /* no way to supply optimal buffer size */
  buf = (char*)malloc( PATH_MAX );

  if( buf )
  {
   /* try again */
   cwd = getcwd( buf, PATH_MAX );

   if( !cwd )
   {
    free( buf );
    return NULL;
   }
  }
  else errno = ENOMEM;

  return buf;
 }
 return cwd;
}
__MODULEEND__



#endif /* compat_getcwd */


#endif /* __MACOS__ */



/* end of file coralsrc/src/compat.c */

