/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <coralsrc/io.h>
#include <coralsrc/dio.h>
#include <coralsrc/integer.h>
#include <coral/list.h>
#include <coral/inline.h>



/* XXX failure recovery */


/*------------------------------------------------------------------------*
	Free link functions
 *------------------------------------------------------------------------*/

/*
dio_add_free_node
dio_cut_free_node
: return value should be 1 if the free link header should be modifed */



/* dio_search_freenode

         ___ free link maintained in free node list
        |
 | size | link info | ... | size |


XXX currently linked list is used for freelink */

static int
dio_search_free_link
	P_B(( dio, p_size, handle ))
	P_1		struct _DIO __F* dio
	P_N		const _DIOHANDLE __F* p_freelink
	P_N		long __F* p_size
	P_N		_DIOHANDLE __F* handle
	P_E
{
	long size;
	_DIOHANDLE freelink;
	struct _IO __F* io = dio->io;
	unsigned byteorder = dio->byteorder;


	/* first-fit */
	size = *p_size;
	x_store( freelink, *p_freelink );

	while( 1 ) {
		_DIOHANDLE allocated;
		_diosize nodesize;

		if( x_zero(freelink) ) break;

		/* allocated = freelink - sizeof(_diosize) */
		x_store( allocated, x_sub( freelink, x_short(sizeof(_diosize) ) ) );

		if( dioRead( dio, &allocated,
				&nodesize, 1, byteorder|sizeof(_diosize) )!=0 )
			return FAIL_IO;

		if( (nodesize=-nodesize)>=size ) {
			*p_size = nodesize;
			x_store( *handle, freelink );
			return 1;
		}

		if( ioRead( io, &freelink, 1, byteorder|sizeof(_DIOHANDLE) )!=1 )
			return FAIL_IO;
	}

	return 0;
}



/* dio_add_free_node

XXX linked list is used currently */

static int
dio_add_free_node
	P_B(( dio, freelink, freeh ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* freelink
	P_N		const _DIOHANDLE __F* freeh
	P_E
{
	int result;
	_bool modified = x_zero(*freelink);
	result = dioListDNAdd( dio, freelink, NULL, freeh, LIST_FLAG_ADD_NEXT );
	if( result==0 && modified ) return 1;
	return result;
}



/* dio_cut_free_node

XXX linked list is used currently */

static int
dio_cut_free_node
	P_B(( dio, freelink, freeh ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* freelink
	P_N		const _DIOHANDLE __F* freeh
	P_E
{
	_bool modified = x_eq(*freelink,*freeh);
	int result = dioListDNCut( dio, freelink, freeh );
	if( result==0 && modified ) return 1;
	return result;
}



/* dio_get_next_free_link
XXX linked list is used */

static int
dio_get_next_free_link
	P_B(( dio, freelink ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* freelink
	P_E
{
	if( ioRead( dio->io, freelink, 1, dio->byteorder|sizeof(_DIOHANDLE) )==1 )
		return 0;
	return -1;
}



/*------------------------------------------------------------------------*
	Internal functions
 *------------------------------------------------------------------------*/



/* dio_get_free_link */


int
dio_get_free_link
	P_B(( dio, freelink ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* freelink
	P_E
{
	register struct _IO __F* io = dio->io;
	if( ioSeek( io, dio->offset, NULL, IO_SEEK_BEGIN )==FAIL_IO
		|| ioRead( io, freelink, 1, dio->byteorder|sizeof(_DIOHANDLE) )!=1 )
		return FAIL_IO;
	return 0;
}



/* dio_set_free_link */

int
dio_set_free_link
	P_B(( dio, freelink ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* freelink
	P_E
{
	register struct _IO __F* io = dio->io;
	if( ioSeek( io, dio->offset, NULL, IO_SEEK_BEGIN )==FAIL_IO
		|| ioWrite( io, freelink, 1, dio->byteorder|sizeof(_DIOHANDLE) )!=1 )
		return FAIL_IO;
	return 0;
}



/* dio_alloc */

int
dio_alloc
	P_B(( io, size, byteorder ))
	P_1		struct _IO __F* io
	P_N		long size
	P_N		unsigned byteorder
	P_E
{
	_diosize diosize = size;
	if( size<0 ) size = -size;
	return ioWrite( io, &diosize, 1, byteorder|sizeof(_diosize) )==1
		&& ioSeek( io, size, NULL, IO_SEEK_CURRENT )!=FAIL_IO
		&& ioWrite( io, &diosize, 1, byteorder|sizeof(_diosize) )==1 ?
		0 : FAIL_IO;
}

/*------------------------------------------------------------------------*
	API
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "dioinifl.c" )
/* dioInitFreeLink */

int __API
dioInitFreeLink
	P_B(( dio ))
	P_1		struct _DIO __F* dio
	P_E
{
	_DIOHANDLE freelink;
	x_store( freelink, x_const(0,0) );
	return ioWrite( dio->io, &freelink,
		1, dio->byteorder|sizeof(_DIOHANDLE) )==1? 0 : FAIL_IO;
}
__MODULEEND__



__MODULEBEGIN__( "dioalloc.c" )
/* dioAlloc

         ___ allocated pointer
        |
 | size | ... | size |

*/

int __API
dioAlloc
	P_B(( dio, handle, size ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* handle
	P_N		long _size
	P_E
{
	struct _IO __F* io;
	_DIOHANDLE freelink;
	unsigned byteorder;
	int result;
	long nodesize;

	/* align the given size */
	if( _size==0 ) return FAIL_UNAVAILABLE;
	if( _size<dio->minsize ) _size = dio->minsize;
	nodesize = _size = DIO_ALIGN( _size, dio->align );

	/* seek to the freelink header */
	io = dio->io;
	byteorder = dio->byteorder;

	if( dio_get_free_link( dio, &freelink )!=0 )
		return FAIL_IO;

	/* search free link */
	result = dio_search_free_link( dio, &freelink, &nodesize, handle );
	if( result==0 ) { /* free link not found */
		long offset_high=0, offset=ioSeek( io, 0, &offset_high, IO_SEEK_END );
		if( ( offset==FAIL_IO && offset_high==FAIL_IO )
			|| dio_alloc(io,nodesize,byteorder)!=0 )
			return FAIL_IO;
		/* *handle = offset + sizeof(_diosize) */
		x_store( *handle, x_add( x_const(offset_high,offset),
			sizeof(_diosize) ) );
	}
	else if( result>0 ) {

		if( nodesize>=dio->minsize+_size+(long)sizeof(_diosize[2]) ) {
			/* divide free node */
			nodesize = _size+sizeof(_diosize[2])-nodesize;
			x_store( freelink, x_sub(*handle,sizeof(_diosize) ) );
			x_adds( *handle, x_short(sizeof(_diosize[2])-nodesize) );
			if( dioSeek( dio, &freelink )!=0 ||
				dio_alloc( io, nodesize, byteorder )!=0 ||
				dio_alloc( io, _size, byteorder )!=0 )
				return FAIL_IO;
			result = 0;
		}
		else {
			result = dio_cut_free_node( dio, &freelink, handle );
			if( result>0 )
				result = dio_set_free_link( dio, &freelink );
			if( result>=0 ) {
				/* freelink = *handle - sizeof(_diosize) */
				x_store( freelink, x_sub( *handle, sizeof(_diosize) ) );
				if( dioSeek( dio, &freelink )!=0 ||
					dio_alloc( io, nodesize, byteorder )!=0 )
					return FAIL_IO;
				result = 0;
			}
		}
	}

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "diofree.c" )
/* dioFree */
/*                     _ h
             nodesize |
           size0 size1|           nextsize
 |size| ... |size|size| ... |size|size|...|size|
                 |_______________|
                  node to free
            |_________|     |_________|
    case A. sizeof(_diosize[2]) bytes are
    additionally added for previous
    node merge operation.
*/

int __API
dioFree
	P_B(( dio, handle ))
	P_1		struct _DIO __F* dio
	P_N		const _DIOHANDLE __F* handle
	P_E
{
	_DIOHANDLE freeh, endh, freelink;
	_diosize nodesize[2], totalsize, nextsize;

	int result = 0;
	struct _IO* io = dio->io;
	unsigned byteorder = dio->byteorder;

	/* XXX Handle existence test is needed ? */

	/* check the given handle is the first handle */
	if( x_gt(*handle, x_short(dio->offset+(sizeof(_DIOHANDLE)+sizeof(_diosize))) ) ) {
		/* get the previous and the current node size */
		x_store( freeh, x_sub( *handle, sizeof(nodesize) ) );
		dioRead( dio, &freeh, nodesize, 2, byteorder|sizeof(_diosize) );
	}
	else {
		nodesize[0] = 1; /* force previous node is not a free node */
		/* get current node size */
		x_store( freeh, x_sub( *handle, sizeof(_diosize) ) );
		dioRead( dio, &freeh, nodesize+1, 1, byteorder|sizeof(_diosize) );
	}

	if( nodesize[0]<0 ) { /* see diagram, case A */
		register _diosize newsize = nodesize[0] - sizeof(nodesize);
		totalsize = newsize - nodesize[1];
		x_store( freeh, x_add( *handle, x_short(newsize) ) );

		if( dio_get_free_link( dio, &freelink )!=0
			|| dio_cut_free_node( dio, &freelink, &freeh )<0 )
			return FAIL_IO;
	}
	else {
		totalsize = -nodesize[1];
		x_store( freeh, *handle );
	}

	x_store( endh, x_add(*handle, x_short(nodesize[1]+sizeof(_diosize))) );
	if( x_eq( endh, ioSeek(io,0,NULL,IO_SEEK_END) ) ) {
		/* XXX if total size is relatively small, truncation operation can be
		  reserved */

		x_subs( freeh, sizeof(_diosize) );
		dioSeek( dio, &freeh );
		dio->seteof( io );
	}
	else {
		dioRead( dio, &endh, &nextsize, 1, byteorder|sizeof(_diosize) );

		if( nodesize[0]>=0 )
			dio_get_free_link( dio, &freelink );

		if( nextsize<0 ) { /* Next node is free node */
			x_adds( endh, sizeof(_diosize) );
			dio_cut_free_node( dio, &freelink, &endh );

			/* Merge */
			totalsize -= sizeof(nodesize)-nextsize;
		}

		x_store( endh, x_sub( freeh, x_short(sizeof(_diosize)) ) );
		dioSeek( dio, &endh );
		dio_alloc( io, totalsize, byteorder );

		result = dio_add_free_node( dio, &freelink, &freeh );
		if( result>0 )
			dio_set_free_link( dio, &freelink );
	}
	return result;
}
__MODULEEND__



__MODULEBEGIN__( "dioseek.c" )
int __API
dioSeek
	P_B(( dio, handle ))
	P_1		struct _DIO __F* dio
	P_N		const _DIOHANDLE __F* handle
	P_E
{
	long offset_high = x_high(*handle);
	return ioSeek( dio->io, x_low(*handle),
			&offset_high, IO_SEEK_BEGIN )==FAIL_IO && offset_high==FAIL_IO ?
		FAIL_IO : 0;
}
__MODULEEND__



__MODULEBEGIN__( "diord.c" )
/* dioRead */

int __API
dioRead
	P_B(( dio, handle, p, size ))
	P_1		struct _DIO __F* dio
	P_N		const _DIOHANDLE __F* handle
	P_N		void __F* p
	P_N		long n
	P_N		unsigned flag
	P_E
{
	return dioSeek( dio, handle )==0 && ioRead( dio->io, p, n, flag )==n ?
		0 : FAIL_IO;
}
__MODULEEND__



__MODULEBEGIN__( "diowr.c" )
/* dioWrite */

int __API
dioWrite
	P_B(( dio, handle, p, size ))
	P_1		struct _DIO __F* dio
	P_N		const _DIOHANDLE __F* handle
	P_N		void __F* p
	P_N		long n
	P_N		unsigned flag
	P_E
{
	return dioSeek( dio, handle )==0 && ioWrite( dio->io, p, n, flag )==n ?
		0 : FAIL_IO;
}

__MODULEEND__



__MODULEBEGIN__( "diowalk.c" )
/* dioWalk */

int __API
dioWalk
	P_B(( dio, p_size, used ))
	P_1		struct _DIO __F* dio
	P_N		long __F* p_size
	P_N		long used
	P_E
{
	struct _IO __F* io = dio->io;
	long size = *p_size;
	if( size==0 ) {
		if( ioSeek( io, dio->offset+sizeof(_DIOHANDLE),
				NULL, IO_SEEK_BEGIN )==FAIL_IO )
			return FAIL_IO;
	}
	else {
		if( ioSeek( io, (size>0?size:-size)-used+sizeof(_diosize), NULL,
				IO_SEEK_CURRENT )==FAIL_IO )
			return FAIL_IO;
	}
	if( ioRead(io,p_size,1,dio->byteorder|sizeof(_diosize))!=1 )
		return FAIL_UNAVAILABLE;

	return 0;
}
__MODULEEND__


__MODULEBEGIN__( "diowfree.c" )
int __API
dioWalkFree
	P_B(( dio, walk, param ))
	P_1		struct _DIO __F* dio
	P_N		int (__API* walk) P_A(( long size, void __F* param ))
	P_N		void __F* param
	P_E
{
	_DIOHANDLE freelink;
	long size;
	struct _IO __F* io = dio->io;
	unsigned byteorder = dio->byteorder;
	int result = dio_get_free_link( dio, &freelink );
	if( result==0 ) {
		while( x_nonzero(freelink) ) {
			int result;
			x_subs( freelink, x_short(sizeof(_diosize)) );
			if( dioSeek( dio, &freelink )!=0
				|| ioRead(io,&size,1,byteorder|sizeof(_diosize) )!=1 )
				return FAIL_IO;
			result = walk( size, param );
			if( result!=0 ) break;
			result = dio_get_next_free_link( dio, &freelink );
			if( result!=0 ) break;
		}
	}
	return result;
}
__MODULEEND__



/* EOF */

