/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */



#include <coralsrc/integer.h>
#include <coralsrc/dio.h>
#include <coralsrc/list.h>



__MODULEBEGIN__( "dldnadd.c" )
/* dioListDNAdd */

int __API
dioListDNAdd
	P_B(( dio, head, p_where, what, flag ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* head
	P_N		const _DIOHANDLE __F* p_where
	P_N		const _DIOHANDLE __F* what
	P_N		unsigned flag
	P_E
{
	if( x_nonzero(*what) ) {
		_DIOHANDLE link[2], where, handle;
		if( x_nonzero(*head) ) {
	//		union {
	//			struct { _DIOHANDLE where, whereprev; } a;
	//			struct { _DIOHANDLE wherenext, where; } b;
	//			DIOLIST c;
	//		} nextprev;
			unsigned byteorder = dio->byteorder;

			if( p_where==NULL || x_zero(*p_where) ) {
				x_store( where, *head );
				goto l_add_prev;
			}
			else {
				x_store( where, *p_where );
			}

			if( flag&LIST_FLAG_ADD_PREV ) {
				l_add_prev:
				/* what->next = where; what->prev = where->prev; */
				x_store( link[0], where );
				x_store( handle, x_add( where, sizeof(_DIOHANDLE) ) );
	// XXX EXCEPTION
				dioRead( dio, &handle, &link[1],
					1, byteorder|sizeof(_DIOHANDLE) ); /* where->prev */
				dioWrite( dio, what, link, 2, byteorder|sizeof(_DIOHANDLE) );

				if( x_nonzero(link[1]) )
					dioWrite( dio, &link[1], (void __F*)what, 1,
						byteorder|sizeof(_DIOHANDLE) ); /* where->prev->next = what */

				/* where->prev = what */
				dioWrite( dio, &handle, (void __F*)what,
					1, byteorder|sizeof(_DIOHANDLE) );

				if( (flag&LIST_FLAG_QUEUE) || x_ne(where,*head) )
					return 0;
			}
			else {
				/* what->next = where->next; what->prev = where; */
				x_store( link[1], where );
	//XXX EXCEPTION
				dioRead( dio, &where, &link[0], 1, sizeof(_DIOHANDLE) );
				dioWrite( dio, what, link, 2, sizeof(_DIOHANDLE) );

				if( x_nonzero(link[0]) ) {
					/* where->next->prev = what */
					x_adds( link[0], sizeof(_DIOHANDLE) );
					dioWrite( dio, &link[0], (void __F*)what,
						1, sizeof(_DIOHANDLE) );
				}

				/* where->next = what */
				dioWrite( dio, &where, (void __F*)what, 1, sizeof(_DIOHANDLE) );
				return 0;
			}
		}
		else {
			/* what->next = what->prev = flag&LIST_FLAG_CIRCULAR? what:NULL */
			x_store( link[1],
				flag&LIST_FLAG_CIRCULAR? *what : x_const(0,0) );
			x_store( link[0], link[1] );
			dioWrite( dio, what, link, 2, sizeof(_DIOHANDLE) );
		}
		x_store( *head, *what );
	}
	return 0;
}



__MODULEBEGIN__( "dldncut.c" )
/* dioListDNCut */

int __API
dioListDNCut
	P_B(( diol, head, what ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* head
	P_N		const _DIOHANDLE __F* what
	P_E
{
	unsigned byteorder = dio->byteorder|sizeof(_DIOHANDLE);
	_DIOHANDLE whatlink[2];
// XXX exception
	int result = dioRead( dio, what, whatlink, 2, byteorder );
	if( result==0 ) {
		if( x_eq(*head,*what) )
			x_store( *head,
				x_eq(*head,whatlink[0]) ? x_const(0,0) : whatlink[0] );

		if( whatlink[1] ) {
			/* what->prev->next = what->next */
			dioWrite( dio, &whatlink[1], &whatlink[0], 1, byteorder );
		}

		if( whatlink[0] ) {
			/* what->next->prev = what->prev */
			x_adds( whatlink[0], x_short(sizeof(_DIOHANDLE)) );
			dioWrite( dio, &whatlink[0], &whatlink[1], 1, byteorder );
		}

	} /* if( result==0 ) */

	return result;
}


__MODULEBEGIN__( "dldncr.c" )
/* dioListDNCreate */

int __API
dioListDNCreate
	P_B(( dio, handle, head, where, size, flag ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* handle
	P_N		_DIOHANDLE __F* head
	P_N		const _DIOHANDLE __F* where
	P_N		long size
	P_N		unsigned flag
	P_E
{
	_DIOHANDLE allocated;
	int result = dioAlloc( dio, &allocated, size );
	if( result==0 ) {
		if( flag&FLAG_CLEAR ) {
 /*  while( size-- )
    dioWrite(
   memShortSet( newnode, 0, size ); */
		}
		result = dioListDNAdd( dio, head, where, &allocated, flag );
		if( result==0 )
			x_store( *handle, allocated );
	}

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "dldnrm.c" )
/* dioListDNRemove */

int __API
dioListDNRemove
	P_B(( dio, head, what ))
	P_1		struct _DIO __F* dio
	P_N		_DIOHANDLE __F* head
	P_N		const _DIOHANDLE __F* what
	P_E
{
	dioListDNCut( dio, head, what );
	dioFree( dio, what );
	return 0;
}
__MODULEEND__



/* EOF */

