/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Dynamic Loader Extension */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <coralsrc/ext/mutex.h> /* must be here because it includes windows.h */
#include <coralsrc/memory.h>
#include <coralsrc/dl.h>
#include <coralsrc/file.h>
#include <coralsrc/string.h>
#include <coral/compat/string.h>



EXTERN_RECURSIVE_MUTEX( dl_mutex );



__MODULEEND__



__MODULEBEGIN__( "dlopen.c" )

/* dlOpen */

void __F* __API
dlOpen( name, flag )
	const char __F* name;
	unsigned flag;
{
	void __F* handle;
	char buf[256], __F* p=buf;
	if( !(flag&DL_FLAG_PURE_NAME) ) {
		int len = strlen(name) + _DL_DEFAULT_NAME_EXTRA + 1;
		if( sizeof(buf)<len ) {
			p = (char __F*)memAlloc( len );
			if( p==NULL ) return NULL;
		}
		sprintf( p, _DL_DEFAULT_NAME_FORMAT, name );
		name = p;
	}

	handle = dl_open( name, flag&DL_FLAG_IMMEDIATE? _DL_NOW : _DL_LAZY );

	if( p!=buf ) memFree( p );
	return handle;
}



/* dlClose */

void __API
dlClose( handle )
	void __F* handle;
{
	dl_close( handle );
}




/* dlSymbol */

void __F* __API
dlSymbol( handle, name )
	void __F* handle;
	const char __F* name;
{

#ifdef __UNIX__
#ifdef __hpux
	void* symbol = NULL;
	shl_findsym( (shl_t*)&handle, name, TYPE_PROCEDURE, &symbol ) ;
	return symbol;
#else
	return dl_symbol( handle, name );
#endif
#else /* !__UNIX__ */
#ifdef __WINDOWS__
	void __F* symbol;
	char buf[128], __F* p = buf;

	int paramsize_index = strBackwardTokenIndex(name,0,"@",0);
	if( paramsize_index>0 ) {
		int len = strlen(name)+1;
		if( sizeof(buf)<len ) {
			p = memAlloc( len );
			if( p==NULL ) return NULL;
		}
		strcpy( p, name );
		name = p;
	}

	while( 1 ) {
		symbol = dl_symbol( handle, name );
		if( symbol || paramsize_index<0 ) break;
		p[paramsize_index] = 0;
		name = p + 1;
		paramsize_index = -1;
	}

	if( buf!=p ) memFree( p );
	return symbol;
#else /* !__WINDOWS__ */
	...
#endif /* __WINDOWS__ */
#endif /* __UNIX__ */
}



__MODULEBEGIN__( "dlinit.c" )



/* Globals for dlInitialize */
int dl_init_count;
DEFINE_RECURSIVE_MUTEX( dl_mutex );



/* dlInitialize */

_bool __API
dlInitialize
	P_V
{
	if( dl_init_count<=0 ) {
		dl_init_count++;
		RECURSIVE_MUTEX_INIT( dl_mutex );
	}
	else {
		dlBegin();
		dl_init_count++;
		dlEnd();
	}

	return 1;
}



/* dlFinalize */

void __API
dlFinalize
P_V
{
	dlBegin();
	dl_init_count--;
	dlEnd();

	if( dl_init_count==0 ) /* only do finalization when counter varies from 1 to 0 */
		RECURSIVE_MUTEX_CLEAR( dl_mutex );
	else if( dl_init_count<0 )
		dl_init_count = 0;
}



/* dlBegin */

void __API
dlBegin
P_V
{
	RECURSIVE_MUTEX_LOCK( dl_mutex );
}



/* dlEnd */

void __API
dlEnd
P_V
{
	RECURSIVE_MUTEX_UNLOCK( dl_mutex );
}

__MODULEEND__



/* end of dl.c */

