/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Compatibility for the dynamic loader of a shared library */


#ifndef __CORALSRC_DL_H
#	define __CORALSRC_DL_H



#include <coralsrc/def.h>
#include <coral/dl.h>



#ifdef __UNIX__


#	ifdef __hpux
		/* -ldld needed */
#		include <dl.h>
#		define _DL_LAZY BIND_DEFERRED
#		define _DL_NOW  BIND_IMMEDIATE
#		define dl_open(name,mode) shl_load(name,mode,0L)
#		define dl_close(handle) shl_unload(handle)

#	else /* !__hpux */
		/* -ldl needed */
#		define _DL_LAZY RTLD_LAZY
#		define _DL_NOW  RTLD_NOW
#		include <dlfcn.h>
#		define dl_open dlopen
#		define dl_close dlclose
#		define dl_symbol dlsym
#	endif /* __hpux */

#define _DL_DEFAULT_NAME_EXTRA 6
#define _DL_DEFAULT_NAME_FORMAT   "lib%s.so"

#else /* !__UNIX__ */
#ifdef __WINDOWS__

#	ifndef WINVER

#		ifdef __WIN32__
			extern void __F* __API LoadLibraryA( const char __F* name );
#			define dl_open(name,mode) LoadLibraryA(name)
#		else
			void __F* __API LoadLibrary( const char __F* name );
#			define dl_open(name,mode) LoadLibrary(name)
#		endif
		extern _bool __API FreeLibrary( void __F* handle );
		extern void __F* __API GetProcAddress( void __F* handle, const char __F* name );
#		define dl_close(handle) FreeLibrary(handle)
#		define dl_symbol(handle,symbol) GetProcAddress(handle,symbol)
#	else /* !WINVER */
#		define dl_open(name,mode) ((void*)LoadLibrary(name))
#		define dl_close(handle) FreeLibrary((HINSTANCE)handle)
#		define dl_symbol(handle,symbol) GetProcAddress((HINSTANCE)handle,symbol)
#	endif /* WINVER */

#define _DL_DEFAULT_NAME_EXTRA  4
#define _DL_DEFAULT_NAME_FORMAT "%s"

#else /* !__WINDOWS__ */

...

#endif /* __WINDOWS__ */
#endif /* __UNIX__ */



#ifndef _DL_LAZY
# define _DL_LAZY 0x0000
# define _DL_NOW  0x0001
#endif



__CBEGIN__
__CEND__



#endif /* __CORALSRC_DL_H */

