/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* mutex definitions */



#ifndef __CORAL_EXT_MUTEX_H
# define __CORAL_EXT_MUTEX_H



#ifdef __UNIX__

#include <sys/sem.h>
# define DEFINE_MUTEX(name)  int name;
# define EXTERN_MUTEX(name)  extern int name
# define MUTEX_INIT(cs)      semctl
# define MUTEX_CLEAR(cs)     semctl
# define MUTEX_LOCK(cs)      semget
# define MUTEX_UNLOCK(cs)    semget

/* -lpthread is needed */
# include <pthread.h>
# define DEFINE_FAST_MUTEX(name)  pthread_mutex_t name;
# define EXTERN_FAST_MUTEX(name)  extern pthread_mutex_t name;
# define FAST_MUTEX_INIT(name)    pthread_mutex_init(&name,attr)
# define FAST_MUTEX_CLEAR(name)   pthread_mutex_destroy(&name)
# define FAST_MUTEX_LOCK(name)    pthread_mutex_lock(&name)
# define FAST_MUTEX_UNLOCK(name)  pthread_mutex_unlock(&name)


#else /* !__UNIX__ */
#ifdef __WIN32__

# include <windows.h>
# define DEFINE_MUTEX(name) CRITICAL_SECTION name;
# define EXTERN_MUTEX(name) extern CRITICAL_SECTION name;
# define MUTEX_INIT(name)   InitializeCriticalSection(&name);
# define MUTEX_CLEAR(name)  DeleteCriticalSection(&name);
# define MUTEX_LOCK(name)   EnterCriticalSection(&name);
# define MUTEX_UNLOCK(name) LeaveCriticalSection(&name);

#else /* __WIN32__ */


# define DEFINE_MUTEX(name)
# define EXTERN_MUTEX(name)
# define MUTEX_INIT(cs)
# define MUTEX_CLEAR(cs)
# define MUTEX_LOCK(cs)
# define MUTEX_UNLOCK(cs)


#endif /* __WIN32__ */
#endif /* __UNIX__ */



#ifndef DEFINE_FAST_MUTEX
# define DEFINE_FAST_MUTEX DEFINE_MUTEX
# define EXTERN_FAST_MUTEX EXTERN_MUTEX
# define FAST_MUTEX_INIT   MUTEX_INIT
# define FAST_MUTEX_CLEAR  MUTEX_CLEAR
# define FAST_MUTEX_LOCK   MUTEX_LOCK
# define FAST_MUTEX_UNLOCK MUTEX_UNLOCK
#endif /* DEFINE_FAST_MUTEX */

#ifndef DEFINE_RECURSIVE_MUTEX
# define DEFINE_RECURSIVE_MUTEX DEFINE_MUTEX
# define EXTERN_RECURSIVE_MUTEX EXTERN_MUTEX
# define RECURSIVE_MUTEX_INIT   MUTEX_INIT
# define RECURSIVE_MUTEX_CLEAR  MUTEX_CLEAR
# define RECURSIVE_MUTEX_LOCK   MUTEX_LOCK
# define RECURSIVE_MUTEX_UNLOCK MUTEX_UNLOCK
#endif /* DEFINE_RECURSIVE_MUTEX */

#endif /* __CORAL_EXT_MUTEX_H */
