/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Extended Integer Instructions */
/*
	1999.04.25 module created
	2000.06.05 module renamed
*/

#include <coralsrc/memory.h>
#include <coralsrc/integer.h>
#include <coral/inline.h>



/*
 *
 *  Extended Integer Instructions
 *
 */
/*
  These instructions are defined to port extended integer operation codes
 to long long supported compilers and less concerned on their execution
 speed.
*/



#ifndef __INTEGERX__



__MODULEBEGIN__( "x_sign.c" )
/* x_sign */

__integerx __API
x_sign
	P_B(( a ))
	P_1		__integerx a
	P_E
{
	/* see OP_X_SUBS, replace a of it to 0 */
	a.s.high = a.s.low!=0 ? -a.s.high-1 : -a.s.high;
	a.s.low = (_uhalfx)-(_halfx)a.s.low;

	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_add.c" )
/* x_add */

__integerx __API
x_add
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	a.s.high += b.s.low>OP_MAXUINT(_uhalfx)-a.s.low ? b.s.high+1 : b.s.high;
	a.s.low += b.s.low;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_add1.c" )
/* x_add1 */

__integerx __API
x_add1
	P_B(( a ))
	P_1		__integerx a
	P_E
{
	if( a.s.low==OP_MAXUINT(_uhalfx) ) a.s.high++;
	a.s.low++;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_sub.c" )
/* x_sub */

__integerx __API
x_sub
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	a.s.high -= a.s.low<b.s.low ? b.s.high+1 : b.s.high;
	a.s.low -= b.s.low;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_sub1.c" )
/* x_sub1 */

__integerx __API
x_sub1
	P_B(( a ))
	P_1		__integerx a
	P_E
{
	if( a.s.low==0 ) a.s.high--;
	a.s.low--;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_mul.c" )
/* x_mul

		Evidently this routine can be optimized using edx:eax combinated multiplication
	in i386 instruction, but c can give the portability.
*/

__integerx __API
	x_mul
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	register _uhalfx val, overflow, hlowlow,
		lowlow = (_uhalfx)a.q.lowlow * (_uhalfx)b.q.lowlow,
		lowhigh = (_uhalfx)a.q.lowlow * (_uhalfx)b.q.lowhigh,
		highlow = (_uhalfx)a.q.lowhigh * (_uhalfx)b.q.lowlow;

	hlowlow = lowlow >> (sizeof(_uquarterx)*8);
	val = lowhigh + highlow;

	/* following overflow test is based on x_add */
	overflow = highlow>OP_MAXUINT(_uhalfx)-lowhigh?
			hlowlow>OP_MAXUINT(_uhalfx)-val?
		2<<(sizeof(_uquarterx)*8) : 1<<(sizeof(_uquarterx)*8) : 0;

	val += hlowlow;

	/* don't care high part is overflowed, a.high*b.high need not to be
		considered because they will be always placed in the overflowed part */
	a.s.high = a.s.high*(_halfx)b.s.low + (_halfx)a.s.low*b.s.high +
		(_uhalfx)a.q.lowhigh*(_uhalfx)b.q.lowhigh + overflow +
		(val>>(sizeof(_uquarterx)*8));

	a.q.lowhigh = (_uquarterx)val;
	a.q.lowlow = (_uquarterx)lowlow;

	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_div.c" )
/* x_divide */

#define _INTX_FLAG_NEGATIVE 0x0004
__integerx __API
x_divide
	P_B(( a, b, flag ))
	P_1		__integerx a
	P_N		__integerx b
	P_N		unsigned flag
	P_E
{
	if( b.s.high==0 && ( a.s.high==0 || b.s.low==0 ) )
	{
		/* if b==0, zero devision exception will be occurred */
		if( flag&INTX_FLAG_MODULUS ) a.s.low%=b.s.low;
		else a.s.low/=b.s.low;
	}
	else
	{
		unsigned i;
		__integerx mod;

		/* if not unsigned case, sign evaluation is needed */
		if( !(flag&INTX_FLAG_UNSIGNED) )
		{
			if( a.s.high<0 )
			{
				/* neg dx(a.s.high)
				   neg ax(a.s.low)
				   sbb dx(a.s.high),0000H */
				a.s.high = a.s.low ? -a.s.high-1 : -a.s.high;
				a.s.low = (_uhalfx)-(_halfx)a.s.low;
				/* enable sign bits */
				flag|=((_INTX_FLAG_NEGATIVE<<1)|_INTX_FLAG_NEGATIVE);
			}
			if( b.s.high<0 )
			{
				/* neg cx(b.s.high)
				   neg bx(b.s.low)
				   sbb cx(b.s.high),0000H */
				b.s.high = b.s.low ? -b.s.high-1 : -b.s.high;
				b.s.low = (_uhalfx)-(_halfx)b.s.low; /* neg  */
				flag ^= _INTX_FLAG_NEGATIVE; /* invert the 3rd bit, the sign flag */
			}
		}

		/* for each bits */
		mod.s.high = mod.s.low = 0;
		i = sizeof(__integerx)*8;
		while( i-- )
		{
			mod.s.high <<= 1;
			if( mod.s.low&OP_MININT(_halfx) ) mod.s.high|=0x1;
			mod.s.low <<= 1;
			if( a.s.high&OP_MININT(_halfx) ) mod.s.low|=0x1;
			a.s.high <<= 1;
			if( a.s.low&OP_MININT(_halfx) ) a.s.high|=0x1;
			a.s.low <<= 1;

			/* modulus cannot exceed divider */
			if( mod.u.high>b.u.high ||
				( mod.u.high==b.u.high && mod.u.low>=b.u.low ) )
			{
				/* sub si(mod.s.low),bx(b.s.low)
				   sbb di(mod.s.high),bp(b.s.high)
				   inc ax(a.s.low) */
				mod.s.high-= mod.u.low<b.u.low ? b.s.high+1 : b.s.high;
				mod.s.low -= b.s.low;
				a.s.low++;
			}
		} /* end while( for each bits ) */

		if( flag&INTX_FLAG_MODULUS )
		{
			a.s.low = mod.s.low;
			a.s.high = mod.s.high;
			/* translate the 4th bit to sign bit
				following sign bit test will be true if above
				if( a.s.high<0 ) section was executed */
			flag>>=1;
		}

		if( flag&_INTX_FLAG_NEGATIVE )
		{
			/* neg dx(a.s.high)
			   neg ax(a.s.low)
			   sbb dx(a.s.high),0000H */
			a.s.high = a.s.low ? -a.s.high-1 : -a.s.high;
			a.s.low = (_uhalfx)-(_halfx)a.s.low;
		}
	}

	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_and.c" )
/* x_and */

__integerx __API
x_and
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	a.s.high&=b.s.high;
	a.s.low&=b.s.low;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_or.c" )
/* x_or */

__integerx __API
x_or
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	a.s.high|=b.s.high;
	a.s.low|=b.s.low;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_xor.c" )
/* x_xor */

__integerx __API
x_xor
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	a.s.high^=b.s.high;
	a.s.low^=b.s.low;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_not.c" )
/* x_not */

__integerx __API
x_not
	P_B(( a ))
	P_1		__integerx a
	P_E
{
	a.s.high=~a.s.high;
	a.s.low=~a.s.low;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_shr.c" )
/* x_shr */

__integerx __API
x_shr
	P_B(( a, cnt ))
	P_1		__integerx a
	P_N		unsigned cnt
	P_E
{
	if( cnt>=sizeof(_halfx)*8 )
	{
		a.s.low = a.s.high>>(cnt-sizeof(_halfx)*8);
	}
	else
	{
		a.s.low = (a.s.low>>cnt) | (a.s.high<<(sizeof(_halfx)*8-cnt));
	}
	a.s.high >>= cnt;

	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_ushr.c" )
/* x_ushr */

__integerx __API
x_ushr
	P_B(( a, cnt ))
	P_1		__integerx a
	P_N		unsigned cnt
	P_E
{
	if( cnt>=sizeof(_halfx)*8 )
	{
		a.u.low = a.u.high>>(cnt-sizeof(_halfx)*8);
		/* following code cannot be a.u.high>>=cnt because first 5 bit of
			cnt is used in 32 bit instruction. */
		a.u.high = 0;
	}
	else
	{
		a.u.low = (a.u.low>>cnt) | (a.u.high<<(sizeof(_halfx)*8-cnt));
		a.u.high >>= cnt;
	}

	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_zero.c" )
/* x_zero */

_bool __API
x_zero
	P_B(( a ))
	P_1		__integerx a
	P_E
{
	return a.s.high==0 && a.s.low==0;
}
__MODULEEND__



__MODULEBEGIN__( "x_nzero.c" )
/* x_nonzero */

_bool __API
x_nonzero
	P_B(( a ))
	P_1		__integerx a
	P_E
{
	return a.s.high!=0 || a.s.low!=0;
}
__MODULEEND__



__MODULEBEGIN__( "x_eq.c" )
/* x_eq */

_bool __API
x_eq
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	return a.s.high==b.s.high && a.s.low==b.s.low;
}
__MODULEEND__



__MODULEBEGIN__( "x_ne.c" )
/* x_ne */

_bool __API
x_ne
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	return a.s.high!=b.s.high || a.s.low!=b.s.low;
}
__MODULEEND__



__MODULEBEGIN__( "x_gt.c" )
/* x_gt */

_bool __API
x_gt
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	return a.s.high==b.s.high? a.s.low>b.s.low : a.s.high>b.s.high;
}
__MODULEEND__



__MODULEBEGIN__( "x_ge.c" )
/* x_ge */

_bool __API
x_ge
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	return a.s.high==b.s.high? a.s.low>=b.s.low : a.s.high>b.s.high;
}
__MODULEEND__



__MODULEBEGIN__( "x_lt.c" )
/* x_lt */

_bool __API
x_lt
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	return a.s.high==b.s.high? a.s.low<b.s.low : a.s.high<b.s.high;
}
__MODULEEND__



__MODULEBEGIN__( "x_le.c" )
/* x_le */

_bool __API
x_le
	P_B(( a, b ))
	P_1		__integerx a
	P_N		__integerx b
	P_E
{
	return a.s.high==b.s.high? a.s.low<=b.s.low : a.s.high<b.s.high;
}
__MODULEEND__



__MODULEBEGIN__( "x_ugt.c" )
/* x_ugt */

_bool __API
x_ugt
	P_B(( a, b ))
	P_1		__unsignedx a
	P_N		__unsignedx b
	P_E
{
	return a.u.high==b.u.high? a.u.low>b.u.low : a.u.high>b.u.high;
}
__MODULEEND__



__MODULEBEGIN__( "x_uge.c" )
/* x_uge */

_bool __API
x_uge
	P_B(( a, b ))
	P_1		__unsignedx a
	P_N		__unsignedx b
	P_E
{
	return a.u.high==b.u.high? a.u.low>=b.u.low : a.u.high>b.u.high;
}
__MODULEEND__



__MODULEBEGIN__( "x_ult.c" )
/* x_ult */

_bool __API
x_ult
	P_B(( a, b ))
	P_1		__unsignedx a
	P_N		__unsignedx b
	P_E
{
	return a.u.high==b.u.high? a.u.low<b.u.low : a.u.high<b.u.high;
}
__MODULEEND__



__MODULEBEGIN__( "x_ule.c" )
/* x_ule */

_bool __API
x_ule
	P_B(( a, b ))
	P_1		__unsignedx a
	P_N		__unsignedx b
	P_E
{
	return a.u.high==b.u.high? a.u.low<=b.u.low : a.u.high<b.u.high;
}
__MODULEEND__



__MODULEBEGIN__( "x_const.c" )
/* x_const */

__integerx __API
x_const
	P_B(( high, low ))
	P_1		_halfx high
	P_N		_halfx low
	P_E
{
	register __integerx a;
	a.s.high = high;
	a.s.low = low;
	return a;
}
__MODULEEND__



__MODULEBEGIN__( "x_short.c" )
/* x_short */

__integerx __API
x_short
	P_B(( low ))
	P_1		_halfx low
	P_E
{
	register __integerx a;
	a.s.high = 0;
	a.s.low = low;
	return a;
}

__MODULEEND__


__MODULEBEGIN__( "x_conv.c" )
/* x_conv_x */

long __API
x_conv_x
	P_B(( a ))
	P_1		__integerx a
	P_E
{
	return (long) a.s.low;
}
__MODULEEND__

#endif /* __LONGLONG__ */



__MODULEBEGIN__( "x_string.c" )
/* x_string

	radix need to be 2 to 36 because 0-9 and A-Z can represent a number. */

unsigned __API
x_string
	P_B(( val, buf, n, radix ))
	P_1		__integerx val
	P_N		char __F* buf
	P_N		unsigned n
	P_N		unsigned radix
	P_E
{
	char __F* s=buf;
	unsigned len=0;

	/* split flags and radix */
	unsigned flag = radix&0xff00;
	radix &= 0xff;

	if( !--n ) return 0;

	/* signed case for radix==10 */
	if( radix==10 )
	{
		/* signed and val<0 */
		if( !(flag&INTX_FLAG_UNSIGNED) && x_lt(val,x_short(0)) )
		{
			x_store( val, x_sign(val) ); /* val = -val */
			*s++ = '-';
			len++;
			n--;
		}
	}
	else
	{
		/* always unsigned if radix is not 10 */
		OP_ENABLE(flag,INTX_FLAG_UNSIGNED);
	}

	do
	{
		int mod;

		if( !n-- ) return 0;
		len++;

		/* (int)(val%radix) */
		mod = x_conv( int, flag&INTX_FLAG_UNSIGNED ?
			x_umod(val,x_short(radix)) : x_mod(val,x_short(radix)) );
		*s++ = (char) ( mod<10 ? '0'+mod :
			mod + ( flag&INTX_FLAG_UPPERCASE? 'A'-10:'a'-10 )  );

		if( flag&INTX_FLAG_UNSIGNED )
			x_udivs( val, x_short(radix) );
		else
			x_divs( val, x_short(radix) );

	} while( x_nonzero(val) );

	/* terminate string */
	*s=0;

	/* invert result */
	if( buf[0]=='-' )
		memInvert( &buf[1], len-1 );
	else
		memInvert( buf, len );

	return len;
}
__MODULEEND__



__MODULEBEGIN__( "x_value.c" )
/* x_value

	if radix is 0, it is internally detected */

__integerx __API
x_value
	P_B(( s, radix ))
	P_1		const char __F* s
	P_N		unsigned radix
	P_E
{
	__integerx a;
	char c;

	/* skip white spaces */
	while( (c=*s)!=0 )
	{
		if( c!='\t' && c!='\n' && c!='\r' && c!=' ' )
			break;
		s++;
	}

	x_store( a, x_short(0) );

	/* check string was not terminated */
	if( c )
	{
		__integerx m;
		unsigned count;
		_bool sign=0;

		/* check c is a sign specifier */
		if( c=='+' || c=='-' )
		{
			if( c=='-' ) sign=1;
			s++; /* ignore sign */
		}

		/* get real radix, compatible with x_string */
		radix&=0xff;

		if( !radix )
		{
			if( *s=='0' )
			{
				c=*++s;
				if( c=='X' || c=='x' )
				{
					radix=16; /* hexadecimal */
					s++; /* ignore 'x' or 'X' */
				}
				else
					radix = 8; /* octal */
			}
			else radix = 10; /* decimal */
		}

		/* count digits */
		count=0;
		while( (c=*s)!=0 )
		{
			if( c<'0' || '9'<c ) /* not a digit */
			{
				if( radix<=10 )
					break;
				if( 'A'<=c && c<='Z' )
				{
					if( (unsigned)(c-'A')>=radix ) break;
				}
				else if( 'a'<=c && c<='z' )
				{
					if( (unsigned)(c-'a') >=radix ) break;
				}
				else break;
			}
			else
			{
				/* character is a digit, check the value is valid */
				if( (unsigned)(c-'0')>=radix ) break;
			}
			count++;
			s++;
		}

		/* convert string to value */
		x_store( m, x_short(1) ); /* m = 1 */
		while( count-- )
		{
			c=*--s;
			/* a += c*m */
			x_adds( a, x_mul( '0'<=c && c<='9'?
				x_short(c-'0'):'A'<=c && c<='Z'?x_short(c-'A'):x_short(c-'a'), m ) );
			/* m *= radix */
			x_muls( m, x_short(radix) );
		}

		if( sign )
			x_store( a, x_sign(a) ); /* a = -a */

	}

	return a;
}
__MODULEEND__



/* EOF */

