/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */



#ifndef __CORALSRC_MEMORY_H
# define __CORALSRC_MEMORY_H



#include <coralsrc/def.h>
#include <coral/memory.h>

/* @cmt0007 */
#ifdef __MACOS__

# include <stdlib.h>

#else /* !__MACOS__ */
#ifdef __PALMOS__

# include <System/MemoryMgr.h>
# define mem_alloc   MemPtrNew
# define mem_free    MemPtrFree

#else /* !__PALMOS__ */
#ifdef __L77__

  extern void* _NAME(lf_malloc)( unsigned size );
  extern void* _NAME(lf_realloc)( void* p, unsigned size );
  extern void  _NAME(lf_free)( void* p );
# define mem_alloc   _NAME(lf_malloc)
# define mem_realloc _NAME(lf_realloc)
# define mem_free    _NAME(lf_free)

#else /* !__L77__ */
#if __ARCH__<=16

# ifdef __TURBOC__
#  include <alloc.h> /* old version turbo c compatible */
#  define mem_alloc   farmalloc
#  define mem_realloc farrealloc
#  define mem_free    farfree
# endif /* !__TURBOC__ */

#else /* __ARCH__>16 */

# include <malloc.h>
# define mem_alloc   malloc
# define mem_realloc realloc
# define mem_free    free

#endif /* __ARCH__<=16 */
#endif /* __L77__ */
#endif /* __PALMOS__ */
#endif /* __MACOS__ */



DECLARE_BALANCED_BINTREE( _mem_name )
	unsigned count;
	char name[1];
DECLARE_END;



__CBEGIN__



/* Following number must not be defined to 65535U because that number is -1 in
 short integer, and ioEnhance can cause some error because IO_ERROR is -1L. */
#define _MEM_LONGITER ((unsigned)64000)



long __API mem_win_iter P_A(( void __F*__F*, long __F* )); /* only for win16 */
_unsigned2 __API mem_long_iter P_A(( char __H* __F*, long __F*, _unsigned2 ));
long __API mem_iterate_2d P_A(( unsigned itemsize, const void __F*__F* pcolrow, const void __F*__F* pcolcol, unsigned variable ));

struct _mem_skip_pad
{
 /* following entries need to be initialized before using mem_ignore_pad */
 unsigned current; /* initially zero */
 _unsigned1 __F* p;          /* start address of a class object */
 _unsigned1 __F* ci;         /* class member instructions */

 /* following entries will be reported for the current operation */
 unsigned item; /* items */
 _unsigned1   size; /* base class size */
};

_unsigned1 __F* __API mem_skip_pad P_A(( struct _mem_skip_pad __F* skip, unsigned flag ));


__CEND__



#endif /* __CORALSRC_MEMORY_H */

