/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Operation ( Undo & Redo ) log */



#include <memory.h>
#include <coralsrc/memory.h>
#include <coralsrc/list.h>
#include <coralsrc/oplog.h>
#include <coral/compat/string.h>
#include <coral/inline.h>




__MODULEBEGIN__( "oplgetr.c" )
/* oplog_get_root */

static struct _OPLOG __F*__F* oplog_get_root P_A(( OPLOGCONTEXT oplc, unsigned mode ));


static struct _OPLOG __F*__F*
oplog_get_root( oplc, mode )
	OPLOGCONTEXT oplc;
	unsigned mode;
{
	return mode==OPLOG_UNDO? &oplc->undo : &oplc->redo;
}
__MODULEEND__



__MODULEBEGIN__( "oplpop.c" )
/* oplog_pop */

static void oplog_pop P_A(( OPLOGCONTEXT oplc, struct _OPLOG __F*__F* root ));

static void
oplog_pop( oplc, root )
	OPLOGCONTEXT oplc;
	struct _OPLOG __F*__F* root;
{
	struct _OPLOG_CLEAR_PARAM clear;
	clear.oplog = *root;
	oplc->callback
		( oplc, OPLOG_CLEAR, (union _OPLOG_PARAM __F*)&clear );
	listDNRemove( (struct _LIST __F*__F*)root, (struct _LIST __F*)clear.oplog );
}
__MODULEEND__



__MODULEBEGIN__( "oplst.c" )
/* oplog_single_transaction */

static int oplog_single_transaction P_A(( OPLOGCONTEXT oplc, struct _OPLOG __F*__F* root ));

static
int
oplog_single_transaction( oplc, root )
	OPLOGCONTEXT oplc;
	struct _OPLOG __F*__F* root;
{
	int result;
	struct _OPLOG_EXECUTE_PARAM execute;
/*	struct _OPLOG __F* oplog = *root; */

	execute.oplog = *root;
	result = oplc->callback
		( oplc, OPLOG_EXECUTE, (union _OPLOG_PARAM __F*)&execute );
	if( result>=0 && !(result&OPLOG_FLAG_RESERVE) )
		oplog_pop( oplc, root );
	return result;
}
__MODULEEND__



__MODULEBEGIN__( "oplmt.c" )

static int oplog_multi_transaction P_A(( OPLOGCONTEXT oplc, unsigned mode, unsigned cmd ));
static int
oplog_multi_transaction( oplc, mode, cmd )
	OPLOGCONTEXT oplc;
	unsigned mode;
	unsigned cmd;
{
	struct _OPLOG __F*__F* root, __F*__F* opposite;/*, __F*__F* temp=NULL;*/
	int result = 0;

	if( mode==OPLOG_UNDO ) {
		root = &oplc->undo;
		opposite = &oplc->redo;
	}
	else {
		root = &oplc->redo;
		opposite = &oplc->undo;
	}
	
	if( *root ) {
		struct _OPLOG __F* mark = *opposite;

		/* reserve and modify flags */
		unsigned old_flag = oplc->flag;
		if( mode==OPLOG_UNDO ) {
			OP_DISABLE( oplc->flag, _OPLOG_FLAG_UNDO );
		}
		else {
			OP_DISABLE( oplc->flag, _OPLOG_FLAG_REDO );
			OP_ENABLE( oplc->flag, _OPLOG_FLAG_EXECUTE_REDO );
		}

		OP_ENABLE( oplc->flag, _OPLOG_FLAG_MULTI_TRANSACTION );

		/* execute op logs */
		if( (*root)->flag&_OPLOG_FLAG_MULTI_TRANSACTION ) {
			/* execution of op logs which returns _OPLOG_FLAG_SKIP */
			while( 1 ) {
				result = oplog_single_transaction(oplc,root);
				if( result<0 ) goto l_terminate;
				if( !(result&OPLOG_FLAG_SKIP)||
				    (*root)->flag&_OPLOG_FLAG_MULTI_TRANSACTION )
					break;
			}

			/* ignore multi-transaction flag for subsequent logs */
			OP_DISABLE( oplc->flag, _OPLOG_FLAG_MULTI_TRANSACTION );

			while( !( (*root)->flag&_OPLOG_FLAG_MULTI_TRANSACTION ) ) {
				/* @cmt0006 */
				if( oplog_single_transaction(oplc,root)<0 ) goto l_terminate;
			}
		}

		/* transact single op log */
		if( oplog_single_transaction(oplc,root)>=0 ) {
			struct _OPLOG __F* current = *opposite;
			if( current!=mark ) { /* it includes 'current!=NULL' and this means
				at least one log was added */
				if( current->next==mark ) /* just one log was added */
					OP_DISABLE( current->flag, _OPLOG_FLAG_MULTI_TRANSACTION );
				else
					OP_ENABLE( current->flag, _OPLOG_FLAG_MULTI_TRANSACTION );
			}
			result = 0;
		}
		l_terminate:
		/* recover flags */
		OP_ENABLE( oplc->flag, old_flag&(_OPLOG_FLAG_UNDO|_OPLOG_FLAG_REDO) );
		OP_DISABLE( oplc->flag,
			_OPLOG_FLAG_MULTI_TRANSACTION|_OPLOG_FLAG_EXECUTE_REDO );
	}

	return result;
}



__MODULEBEGIN__( "oplinit.c" )
/* oplogCreate */

OPLOGCONTEXT __API
oplogCreate( dio, callback, param )
	struct _DIO __F* dio;
	OPLOGCALLBACK callback;
	void __F* param;
{
	OPLOGCONTEXT oplc =
		(OPLOGCONTEXT)memCAlloc( sizeof(struct _oplog_context) );
	if( oplc ) {
		oplc->undo = oplc->redo = NULL;
		oplc->dio = dio;
		oplc->callback = callback;
		oplc->param = param;
	}
	return oplc;
}



/* oplogRemove */

void __API
oplogRemove( oplc )
	OPLOGCONTEXT oplc;
{
	if( oplc ) {
		oplogClear( oplc, OPLOG_UNDO );
		oplogClear( oplc, OPLOG_REDO );
		memFree( oplc );
	}
}
__MODULEEND__



__MODULEBEGIN__( "oplclear.c" )
/* oplogClear */

void __API
oplogClear( oplc, mode )
	OPLOGCONTEXT oplc;
	unsigned mode;
{
	struct _OPLOG __F*__F* root = oplog_get_root( oplc, mode );
	while( *root ) oplog_pop( oplc, root );

	/* XXX log io should be cleared */
}
__MODULEEND__



__MODULEBEGIN__( "oplsetst.c" )
/* oplogSetState */

void __API
oplogSetState( oplc, mode, enable )
	OPLOGCONTEXT oplc;
	unsigned mode;
	_bool enable;
{
	if( enable )
		OP_ENABLE( oplc->flag,
			mode==OPLOG_UNDO?_OPLOG_FLAG_UNDO:_OPLOG_FLAG_REDO );
	else {
		OP_DISABLE( oplc->flag,
			mode==OPLOG_UNDO?_OPLOG_FLAG_UNDO:_OPLOG_FLAG_REDO );
		oplogClear( oplc, mode );
	}
}
__MODULEEND__



__MODULEBEGIN__( "oplsetcn.c" )
void __API
oplogSetCount( oplc, cnt )
	OPLOGCONTEXT oplc;
	unsigned cnt;
{
//	const struct _OPLOG* oplog = oplc->undo;
//	oplog->next = NULL;
}
__MODULEEND__



__MODULEBEGIN__( "oplgetr.c" )
/* oplogGetRoot */

struct _OPLOG __F* __API
oplogGetRoot( oplc, mode )
	OPLOGCONTEXT oplc;
	unsigned mode;
{
	return *oplog_get_root(oplc,mode);
}
__MODULEEND__



__MODULEBEGIN__( "oplnull.c" )
/* oplogNull */

_bool __API
oplogNull( oplc, mode )
	OPLOGCONTEXT oplc;
	unsigned mode;
{
	/* XXX 8 byte integer comparison */
	return mode==OPLOG_UNDO ?
		oplc->undo==NULL && oplc->undo_dioh==0 :
		oplc->redo==NULL && oplc->redo_dioh==0;
}
__MODULEEND__



__MODULEBEGIN__( "oplgetp.c" )
/* oplogGetParam */

void __F* __API
oplogGetParam( oplc )
	OPLOGCONTEXT oplc;
{
	return oplc->param;
}

__MODULEEND__



__MODULEBEGIN__( "opladd.c" )
/* oplogAdd */

struct _OPLOG __F* __API
oplogAdd( oplc, mode, size, ucode, scode )
	OPLOGCONTEXT oplc;
	unsigned mode;
	unsigned size;
	unsigned ucode;
	const char __F* scode;
{
	struct _OPLOG __F* oplog = (struct _OPLOG*)
//		listPush( (struct _LIST __F*__F*)oplog_get_root(oplc,mode),
//			scode? size+strlen(scode)+1 : size );
		listDNCreate( (struct _LIST __F*__F*)oplog_get_root(oplc,mode),
			NULL, scode? size+strlen(scode)+1 : size,
			LIST_FLAG_ADD_PREV|LIST_FLAG_CIRCULAR );
	if( oplog ) {

		/* XXX temporary codes : modify it after verifying solid model
			undo log implementation */
		if( listCNCount((struct _LIST __F*)oplog)>10 ) {//++oplog->current_count>10 ) {
//			struct _OPLOG __F*
			struct _LIST __F*__F* root = (struct _LIST __F*__F*)oplog_get_root(oplc,mode);
			listDNRemove( root, (*root)->prev );
		}

		/* redo log should be cleared if an undo log is added */
		if( mode==OPLOG_UNDO )
			oplogClear( oplc, OPLOG_REDO );

		if( oplc->flag&_OPLOG_FLAG_MULTI_TRANSACTION ) {
			oplog->flag = _OPLOG_FLAG_MULTI_TRANSACTION;
			OP_DISABLE( oplc->flag, _OPLOG_FLAG_MULTI_TRANSACTION );
		}
		else {
			oplog->flag = 0;
		}

		if( scode ) {
			OP_ENABLE( oplog->flag, _OPLOG_FLAG_STRING_CODE );
			strcpy( oplog->code.s=(_integer1 __F*)oplog+size, scode );
		}
		else {
			oplog->code.u = ucode;
		}

	} /* end if( oplog ) */

	return oplog;
}
__MODULEEND__



__MODULEBEGIN__( "oplexec.c" )
/* oplogExecute */

int __API
oplogExecute( oplc, mode )
	OPLOGCONTEXT oplc;
	unsigned mode;
{
	return oplog_multi_transaction( oplc, mode, OPLOG_EXECUTE );
}
__MODULEEND__



/* EOF */

