/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* String operations, Hee S. Chung, internal source copyright

2001.02.01 Module is created */



#include <coralsrc/memory.h>
#include <coralsrc/string.h>
#include <coralsrc/io.h>
#include <coral/inline.h>
#include <coral/compat/string.h>
#include <coral/compat/ctype.h>

__MODULEEND__



/*------------------------------------------------------------------------*

   Globals

 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "strdata.c" )
const char str_white_char[5] = STR_WHITE_CHAR;
__MODULEEND__



#ifdef __MACOS__

#   define _STR_DIRSEP ':'

#else /* !__MACOS__ */
#if defined(__DOS__) || defined(__WINDOWS__)

#	define _STR_DOS_PATH
static const char str_fileext_delim[] = ".:\\/";

#else /* !__DOS__ && !__WINDOWS__ */

#   define _STR_DIRSEP '/'

#endif /* __DOS__ || __WINDOWS__ */
#endif /* __MACOS__ */


#ifndef _STR_DOS_PATH
static const char str_fileext_delim[] = {
	'.', _STR_DIRSEP, 0
};
#endif



#define str_dirsep_delim (str_fileext_delim+1) /* exclude comma */


#ifdef _STR_DOS_PATH
#	define str_pure_dirsep_delim (str_fileext_delim+2) /* exclude ':' */
#else
#	define str_pure_dirsep_delim str_dirsep_delim
#endif



/*------------------------------------------------------------------------*

   Methods for compatibility

 *------------------------------------------------------------------------*/



#ifdef DEFINE_stricmp
__MODULEBEGIN__( "stricmp.c" )
/* stricmp */

int
stricmp
	P_B(( s1, s2 ))
	P_1		const char __F* s1
	P_N		const char __F* s2
	P_E
{
	register char c1, c2;
	while( (c1=toupper(*s1++))==(c2=toupper(*s2++)) ) {
		if( c1==0 ) return 0;
	}

	return c1>c2? 1 : -1;
}

__MODULEEND__
#endif /* DEFINE_stricmp */



#ifdef DEFINE_strnicmp
__MODULEBEGIN__( "strnicmp.c" );
/* strnicmp */

int
strnicmp
	P_B(( s1, s2, count ))
	P_1		const char __F* s1
	P_N		const char __F* s2
	P_N		size_t count
	P_E
{
	while( count-- ) {
		char c1, c2;
		if( (c1=toupper(*s1++))==(c2=toupper(*s2++)) ) {
			if( c1==0 ) break;
		}
		else {
			return c1>c2? 1 : -1;
		}
	}
	return 0;
}

__MODULEEND__
#endif /* DEFINE_strnicmp */



#ifdef DEFINE_strlwr
__MODULEBEGIN__( "strlwr.c" )
/* strlwr */

char __F*
strlwr
	P_B(( s ))
	P_1		char __F* s
	P_E
{
	char __F* i = s;
	while( *i ) {
		*i = tolower(*i);
		i++;
	}
	return s;
}

__MODULEEND__
#endif /* DEFINE_strlwr */



#ifdef DEFINE_strupr
__MODULEBEGIN__( "strupr.c" )
/* strupr */

char __F*
strupr
	P_B(( s ))
	P_1		char __F* s
	P_E
{
	char __F* i = s;
	while( *i ) {
		*i = toupper(*i);
		i++;
	}
	return s;
}

__MODULEEND__
#endif /* DEFINE_strupr */



/*------------------------------------------------------------------------*

   Internal methods

 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "scmpclss.c" )
/* str_compare_class_str

  It's an internal function to support string search for several data structure
 manipulation functions.
  Return 0 if two strings are the same, otherwise return 1 */

int __API
str_compare_class_str
	P_B(( object, str ))
	P_1		void __F* object
	P_N		struct _str_compare_class_str __F* str
	P_E
{
	/* get the string pointer in the given object of a class */
	char __F* s = (char __F*)object + str->offset;
	register unsigned flag = str->flag;

	if( flag&FLAG_STRING_POINTER )
		s = *(char __F*__F*)s;

	/* compare string */
	if( ( strCompare(s,str->s,flag) )==0 ) {
		str->object = object; /* set the current object */
		return 1;
	}

	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "strcomp.c" )
/* strCompare */

int __API
strCompare
	P_B(( a, b, flag ))
	P_1		const char __F* a
	P_N		const char __F* b
	P_N		unsigned flag
	P_E
{
	return ( flag&FLAG_CASE_INSENSITIVE ? strcmpi : strcmp )( a, b );
}
__MODULEEND__



__MODULEBEGIN__( "strcopy.c" )
/* strCopy
If dest==NULL, 'size' should extra size to be allocated.
If dest!=NULL, 'size' should be the give buffer size. */

char __F* __API
strCopy
	P_B(( dest, src, size ))
	P_1		char __F* dest
	P_N		const char __F* src
	P_N		int size
	P_E
{
	int len;

	if( !src ) return NULL;

	len = strlen(src)+1;
	if( dest ) {
		if( size<len ) return NULL;
	}
	else {
		dest = (char __F*)memAlloc( len+size );
		if( !dest ) return NULL;
	}

	strcpy( dest, src );

	return dest;
}

__MODULEEND__



__MODULEBEGIN__( "strunion.c" )
/* strUnion */

char __F* __API
strUnion
	P_B(( dest, s1, s2, size ))
	P_1		char __F* dest
	P_N		const char __F* s1
	P_N		const char __F* s2
	P_N		int size
	P_E
{
	int len1, len;

	if( !s1 && !s2 ) return NULL;

	len1 = s1? strlen(s1) : 0;
	len = len1 + (s2?strlen(s2):0) + 1;

	if( dest ) {
		if( size<len ) return NULL;
	}
	else {
		dest = (char __F*)memAlloc( len+size );
		if( !dest ) return NULL;
	}

	if( s1 )
		strcpy( dest, s1 );

	if( s2 )
		strcpy( dest+len1, s2 );

	return dest;
}

__MODULEEND__



__MODULEBEGIN__( "strapnd.c" )
/* strAppend */

char __F* __API
strAppend
	P_B(( dest, src, size, resize ))
	P_1		char __F* dest
	P_N		const char __F* src
	P_N		int size
	P_N		int __F* resize
	P_E
{
	return strInsert( dest, src, -1, size, resize );
}

__MODULEEND__



__MODULEBEGIN__( "strins.c" )
/* strInsert
	'dest' must be allocated pointer if resize!=NULL.
	'size' can be zero if the space for 'dest' is strlen(dest)+1.
	NULL is returned if this function is failed and supplied
	pointers are reserved. */

char __F* __API
strInsert
	P_B(( dest, src, i, size, resize ))
	P_1		char __F* dest
	P_N		const char __F* src
	P_N		int i
	P_N		int size
	P_N		int __F* resize
	P_E
{
	int extra = 0; /* this code is redundant but some compiler says
						it is not initialized but used */
	if( resize ) {
		extra = *resize;
		*resize = 0; /* assume reallocation is not occurred */
	}

	if( dest ) {
		if( src ) {
			register char __F* p;
			int destlen, srclen, lack;

			if( size<0 ) size=OP_MAXINT(int); /* buffer is sufficient */

			destlen=strlen(dest);
			srclen=strlen(src);
			lack = size-destlen-srclen;

			if( lack<=0 ) { /* buffer is insufficient */
				char __F* newdest;
				if( !resize ||
					( newdest = strCopy
						( NULL, dest, *resize=extra+size-lack+1) )==NULL )
					return NULL;
				memFree( dest );
				dest = newdest;
			} /* end if( lack<=0 ) */

			/* translate characters */
			if( i<0 || i>destlen ) i = destlen;
			p = dest + i;
			memmove( p+srclen, p, destlen-i+1 );
			memcpy( p, src, srclen );
		} /* end if( src ) */
	}
	else {
  /*XXX????????	if( src ) {
		return strCopy( NULL, src,
	  } */
	} /* end if( dest ) */

	return dest;
}

__MODULEEND__



__MODULEBEGIN__( "strgcpl.c" )
/* strGetCompatiblePathLength */

unsigned __API
strGetCompatiblePathLength
	P_B(( path ))
	P_1		const char __F* path
	P_E
{
	return strlen( path ); // XXX
}

__MODULEEND__



__MODULEBEGIN__( "strgcp.c" )

#ifdef __TURBOC__
# pragma argsused
#endif

char __F* __API
strGetCompatiblePath
	P_B(( path ))
	P_1		char __F* path
	P_E
{
#ifdef _STR_DOS_PATH
	register char c, __F* p = path;
  // XXX exception is not fully handled
	if( p[0]=='/' && p[1]=='/' ) {
		/* replace //x/ to x:/ */
		p[0] = p[2];
		p[1] = ':';
		p[2] = '\\';
		p = &p[3];
		strcpy( p, &p[1] ); /* kill a character */
	}
	while( (c=*p)!=0 ) {
		if( c=='/' ) {
			*p='\\';
		}
		else if( c=='\\' ) {
			switch( c = p[1] ) {
				case '\\':
				case ' ':
				/* list other special characters to be accepted here */
				*p++ = c;
			}
		}
		p++;
	}
#endif /* _STR_DOS_PATH */
	return path;
}
__MODULEEND__



__MODULEBEGIN__( "strsds.c" )
/* strStoreDirectorySeparator */

int __API
strStoreDirectorySeparator
	P_B(( name, index ))
	P_1		char __F* name
	P_N		int index
	P_E
{
	if( index>0 ) {
		/* duplicated directory separator is not allowed :
			it is assumed that characters contained in 'name' which
			are stored from name[0] to name[index] are valid,
			and only a previous character is tested  */
#if defined(_STR_DOS_PATH)
		if( strchr(str_pure_dirsep_delim,name[index-1])!=NULL )
#else
		if( _STR_DIRSEP!=name[index-1] )
#endif
			return index;
	}
#if defined(_STR_DOS_PATH)
	{
		char sep='\\';
		int l = index;
		while( --l>=0 ) {
			char c = name[l];
			if( c=='/' || c=='\\' ) {
				sep = c;
				break;
			}
		} /* end while( --len>=0 ) */
		name[index] = sep;
	}
#else /* !_STR_DOS_PATH */
	name[index] = _STR_DIRSEP;
#endif /* _STR_DOS_PATH */

	return index+1;
}
__MODULEEND__



__MODULEBEGIN__( "strads.c" )
/* strAppendDirectorySeparator

 1998.09.08
 Append trailing slash at the end of given path name if needed

 2000.04.17
 Select more preferable character between '/' and '\\' for DOS.

 2001.08.19 Macintosh support

  All directory separator related APIs are standardized for Unix except
 __MACOS__, __DOS__, and __WINDOWS__ */

int __API
strAppendDirectorySeparator
	P_B(( name ))
	P_1		char __F* name
	P_E
{
	int len = strlen(name);

	/* if directory separator is appended to zero length string, the string
	 will indicate the root directory ( or some other directory which the
	 system recognizes, so zero length string should be excluded here */
	if( len>0 ) {
		len = strStoreDirectorySeparator(name,len);
		name[len] = 0;

	} /* end if( len>0 ) */

	return len;
}
__MODULEEND__



__MODULEBEGIN__( "strgds.c" )
/* strGetDirectorySeparator */

const char __F* __API
strGetDirectorySeparator
	P_V
{
	return str_dirsep_delim;
}
__MODULEEND__



__MODULEBEGIN__( "strgfe.c" )
/* strGetFileExtension */

const char __F* __API
strGetFileExtension
	P_B(( path ))
	P_1		const char __F* path
	P_E
{
	const char __F* s = strBackwardToken
		( (char __F*)path, 0, str_fileext_delim, STR_FLAG_FOUND_DELIMITER );
	if( s ) {
		if( *(s-1)=='.' ) return s;
	}
	return NULL;
}

__MODULEEND__



__MODULEBEGIN__( "strrfec" )
/* strReplaceFileExtension
 Extra size 'extlen-orig_extlen' is needed if there is comma in the filename.
 Extra size 'extlen-orig_extlen+1' is needed if comma can be absent
 in the filename */

char __F* __API
strReplaceFileExtension
	P_B(( path, ext ))
	P_1		char __F* path
	P_N		const char __F* ext
	P_E
{
	char __F* s = (char __F*)strGetFileExtension(path);
	if( s ) { /* extension was found */
		if( ext==NULL ) {
			*(s-1)=0; /* cut extension */
			return path;
		}
	}
	else {
		if( ext==NULL ) return path;
		s = path + strlen(path);
		strcpy( s++, "." );
	}
	strcpy( s, ext );

	return path;
}
__MODULEEND__



__MODULEBEGIN__( "strgfn.c" )
/* strGetFileName

 This function gets file name with extension */

const char __F* __API
strGetFileName
	P_B(( path ))
	P_1		const char __F* path
	P_E
{
	const char __F* s = strBackwardToken
		( (char __F*)path, 0, str_dirsep_delim, STR_FLAG_FOUND_DELIMITER );

	return s? s : path;
}

__MODULEEND__



__MODULEBEGIN__( "strgpn.c" )

char __F* __API
strGetPathName
	P_B(( path ))
	P_1		char __F* path
	P_E
{
	char first;
	if( path && (first=path[0])!=0 ) {
		char __F* s = strBackwardToken( path, 0, str_dirsep_delim, 0 );
		if( s==NULL ) { /* directory separator was not found */
			path[0] = '.';
			path[1] = '\0';
		}
		else if( path[0]==0 ) {
			path[0] = first;
			path[1] = '\0';
		}
		return path;
	}
	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "strsplit.c" )
/* strSplitPath
  If only file name is supplied for this function, '*path' will not be
 a part of 'filename'. */

void __API
strSplitPath
	P_B(( dir, name, ext, path ))
	P_1		char __F*__F* dir
	P_N		char __F*__F* name
	P_N		char __F*__F* ext
	P_N		char __F* path
	P_E
{
	char __F* s;
	if( ext ) {
		int len = strlen(path);

		/* find extension */
		s = (char __F*)strGetFileExtension(path);

		if( s ) *(s-1) = 0;
		/* make ext to null string if extension is not present */
		if( ext ) *ext = s ? s : &path[len];
	}

	/* find file name */
	s = (char __F*)strGetFileName(path);

	if( s ) {
		*(s-1) = 0;
		if( dir ) *dir = path;
		if( name ) *name = s;
	}
	else {
		/* zero length string is excluded in strAppendDirectorySeparator */
		if( dir ) *dir = "";
		if( name ) *name = path;
	}
}
__MODULEEND__



__MODULEBEGIN__( "strabsp.c" )
/* strAbsolutePath */

_bool __API
strAbsolutePath
	P_B(( path ))
	P_1		const char __F* path
	P_E
{
	register char first = path[0];
#ifdef _STR_DOS_PATH
	register char third;
	if( isalpha(first) && path[1]==':' &&
		( (third=path[2])=='/' || third=='\\' ) )
		return 1;
#else
	if( first=='/' || first=='\\' ) return 1;
#endif
	return 0;
}

__MODULEEND__



__MODULEBEGIN__( "strrelp.c" )
/* strGetRelativePath

	Canonical path names should be supplied for both 'base' and 'path'.

	If the final character in the given path is not a directory separator,
	the final part of it is regarded as a file name.

	Sufficient buffer size can be calculated by 'strlen(base)+strlen(path)+1'.
 */

char __F* __API
strGetRelativePath
	P_B(( base, path, flag ))
	P_1		const char __F* base
	P_N		char __F* path
	P_N		unsigned flag
	P_E
{
	char __F* p = path;
	if( p ) {
		if( strAbsolutePath( p ) ) {
			char sep, bc, pc;
			const char __F* last=NULL;
			unsigned nprev = 0;
			int len;

	/* actually the following initialization is not needed */
	#ifdef _STR_DOS_PATH
			sep = '\\';
	#else
			sep = '/';
	#endif

			while( (bc=*base)!=0 ) {

	#ifdef _STR_DOS_PATH
				char* pc_delim;
	#endif

				pc = *p;

	#ifdef _STR_DOS_PATH
				pc_delim = strchr(str_pure_dirsep_delim,pc);
				if( ( pc_delim==NULL
					|| strchr(str_pure_dirsep_delim,bc)==NULL )
					&& tolower(bc)!=tolower(pc) )
	#else
				if( bc!=pc )
	#endif
				{
					if( last==NULL )
						return flag&STR_FLAG_ALLOCATE?
							strCopy(NULL,path,0) : path;

					nprev = strCountChar(base,str_pure_dirsep_delim);
					break;
				}

	#ifdef _STR_DOS_PATH
				if( pc_delim )
	#else
				if( pc=='/' )
	#endif
				{
					last = p+1;
					sep = pc;
				}
				p++;
				base++;
			}

			len = strlen(last)+1;

			/* XXX consider mac path */
			if( flag&STR_FLAG_ALLOCATE ) {
				path = (char __F*)memAlloc( nprev*3 + len );
				if( path==NULL ) return NULL;
			}

			memmove( path+nprev*3, last, len );
			p = path;
			while( nprev-->0 ) {
				p[0] = p[1] = '.';
				p[2] = sep;
				p+=3;
			}
		}
		else {
			if( flag&STR_FLAG_ALLOCATE )
				return strCopy(NULL,path,0);
		}
	} /* if( p ) */
	return path;
}

__MODULEEND__



__MODULEBEGIN__( "strabsp.c" )
/* strGetAbsolutePath

	The given path should be canonicalized. */

char __F* __API
strGetAbsolutePath
	P_B(( base, path, flag ))
	P_1		const char __F* base
	P_N		char __F* path
	P_N		unsigned flag
	P_E
{
	if( path ) {
		if( strAbsolutePath(path) ) {
			if( flag&STR_FLAG_ALLOCATE )
				return strCopy( NULL, path, 0 );
		}
		else {
			int i, plen, len=0; /* actually len=0 is not needed */
			char __F* p = path;
			unsigned depth = 1;
			while( (i=strTokenIndex(p,str_pure_dirsep_delim,0))>=0 ) {
				if( p[0]!='.' || p[1]!='.' ) break;
				p = p + i + 1;
				depth ++;
			}

			i = strlen(base);
			while( depth-->0 ) {
				i =  strBackwardTokenIndex( base, i, str_pure_dirsep_delim, 0 );
				if( i<0 ) break;
				len = i;
			}

			plen = strlen(p)+1;

			if( flag&STR_FLAG_ALLOCATE )
				path = (char __F*)memAlloc
					( len + plen + strGetDirectorySeparatorSize() );

			if( path ) {
				memmove( path+len+strGetDirectorySeparatorSize(), p, plen );
				memcpy( path, base, len );
				strStoreDirectorySeparator( path, len );
			}
		} /* end if( strAbsolutePath(path) ) */
	}
	return path;
}
__MODULEEND__


__MODULEBEGIN__( "strcanon.c" )
/* strCanonicalPath ???*/

#if 0
char __F* __API
strCanonicalPath
 ( char __F* result, const char __F* current, char __F* relative )
{
 char c, __F* dest;
 int len, curlen, curfree=0;

 /* check relative path is really relative path */
 if( strAbsolutePath(relative) )
 {
  /* it is absolute path */
  curlen = 0;
  len = strlen(relative);
 }
 else
 {
  if( !current )
  {
   current = getcwd( NULL, 0 );
   if( !current ) return NULL;
   curfree = 1;
  }

  curlen = strlen(current);
  len = curlen + strlen(relative);
 }

 len++;

 /* allocate memory if it was not supplied */
 if( !result )
 {
  result = (char __F*)memAlloc( len );
  if( !result ) return NULL;
 }

 /* union two given strings */
 if( curlen==0 )
 {
  strcpy( result, relative );
 }
 else
 {
  strcpy( result, current );
  strcat( result, relative );
 }

 /* assume the given or generated current path is canonical */

 dest = relative = &result[curlen];
 while( (c=*relative++)!=0 )
 {
  ///if( c=='/'
 }

 if( curfree )
  free( current );

 return result;
}
#endif
__MODULEEND__



__MODULEBEGIN__( "strtok.c" )
/* strToken
 1998.03.07 Written
 2001.03.13 Reviewed and renamed from strSub to strToken

  while( (next=strToken(s,...))!=NULL ) {
	// if STR_FLAG_FOUND_DELIMITER was specified
	// char found = *next;
	// ...;
	// *next++ = 0;
	puts( s );
	s = next;
  }

  If STR_FLAG_EXCLUSIVE is used, this function returns the first position where
 the supplied delimiters was not found. For example, '.' is specified delimiter
 for string s : s = "@....." will return s+1 for exclusive mode. This rule is
 applied to strBackwardToken too.
*/

char __F* __API
strToken
	P_B(( s, delimiter, flag ))
	P_1		char __F* s
	P_N		const char __F* delimiter
	P_N		unsigned flag
	P_E
{
	char __F* t = s;
	if( t && *t ) {

		char c;

		if( delimiter==NULL ) delimiter = str_white_char;

		while( (c=*t)!=0 ) {
			register char __F* found = strchr(delimiter,c);

			if( flag&STR_FLAG_EXCLUSIVE ) {
				if( found==NULL ) {
					/* null termination is unavailable */
					if( t==s ) OP_ENABLE( flag, STR_FLAG_FOUND_DELIMITER );
					else t--;
					break;
				}
			}
			else if( found!=NULL ) {
				break;
			}

			/* translate string pointer */
			t++;
		}

		if( c ) {
			if( !(flag&STR_FLAG_FOUND_DELIMITER) ) *t++ = 0;
			return t;
		}

	} /* end if( s ) */

	return NULL;
}

__MODULEEND__



__MODULEBEGIN__( "strbtok.c" )
/* strBackwardToken

 1998.03.07 Written
 2001.03.13 Reviewed and renamed from strSubBackward to strBackwardToken
 Return substring of parameter s, and s is modified for next iteration.
 ( Modification means termination marked on the place where a
   delimiter was found. ) NULL if there are no more delimiters.

 Typical iteration form is :

  len = strlen(s); // or length can be NULL to be internally calculated
  while( (current=strBackwardToken(s,len,...))!=NULL ) {
	puts( current );
	len = current-s-1;
	// if STR_FLAG_FOUND_DELIMITER is specified
	// char found=*--current;
	// *current = 0;
  }

 s, A string pointer
 len, The length of the string. Zero to get it internally.
 STR_FLAG_END_OF_STRING : s must point the end of string and len must be the
  valid string length


*/

char __F* __API
strBackwardToken
	P_B(( s, len, delimiter, flag ))
	P_1		char __F* s
	P_N		int len
	P_N		const char __F* delimiter
	P_N		unsigned flag
	P_E
{
	if( s ) {

		if( delimiter==NULL ) delimiter = str_white_char;

		/* if string pointer was not translated to the end of string,
		   translate it  : caller must guarantee 's' is the end of string
		   and 'len' is valid string length if STR_FLAG_END_OF_STRING
		   is specified. */
		if( !(flag&STR_FLAG_END_OF_STRING) ) {
			/* if string length was not specified, calculate it */
			if( len<=0 )
				len = strlen(s);

			s = &s[len];
		}

		/* at least a character must be present */
		if( len>0 ) {

			/* start from the last meaningful character : by this,
			  we can share the code to return the result for the two cases,
			  a delimiter was found or not found */
			s--;

			while( --len>=0 ) {
				register char* found = strchr(delimiter,*s);

				if( flag&STR_FLAG_EXCLUSIVE ? found==NULL : found!=NULL )
					break;

				s--;
			}

			if( len>=0 || (flag&STR_FLAG_EXCLUSIVE) ) {
				/* exclusive mode so reserve the character not contained
				  in delimiters */
				if( (flag&STR_FLAG_EXCLUSIVE) && !*++s )
					return NULL;

				/* the found delimiter was not wanted */
				if( !(flag&STR_FLAG_FOUND_DELIMITER) )
					*s = 0;

				return s+1;
			}

		} /* end if( len>0 ) */
	} /* end if( s ) */

	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "strtoki.c" )
/* strTokenIndex */

int __API
strTokenIndex
	P_B(( s, delimiter, flag ))
	P_1		const char __F* s
	P_N		const char __F* delimiter
	P_N		unsigned flag
	P_E
{
	if( s ) {
		char __F* found = strToken
			( (char __F*)s, delimiter, flag|STR_FLAG_FOUND_DELIMITER );
		if( found )
			return found-s;
	}

	return -1;
}



__MODULEBEGIN__( "strbtoki.c" )
/* strBackwardTokenIndex */

int __API
strBackwardTokenIndex
	P_B(( s, len, delimiter, flag ))
	P_1		const char __F* s
	P_N		int len
	P_N		const char __F* delimiter
	P_N		unsigned flag
	P_E
{
	char __F* found = strBackwardToken
		( (char __F*)s, len, delimiter, flag|STR_FLAG_FOUND_DELIMITER );

	/* according to strBackwardToken, if len>0, 's' must point the end of
	  a string so, 's-len' will be the start address of the string. if len==0,
	  's' must point the start adress of a string and s-len==s */

	return found? found-(flag&STR_FLAG_END_OF_STRING?s-len:s)-1 : -1;
}
__MODULEEND__



__MODULEBEGIN__( "strreplc.c" )
/* strReplaceChar */

char __F* __API
strReplaceChar
	P_B(( str, oldchr, newchr, n ))
	P_1		char __F* str
	P_N		const char __F* oldchr
	P_N		const char __F* newchr
	P_N		int n
	P_E
{
	int inc, len = strlen(str), ncend = strlen(newchr)-1;
	char __F* s=str, cncend = newchr[ncend];

	if( n>0 ) { /* forward */
		inc = 1;
	}
	else { /* backward */
		inc = -1;
		n = -n;
		s += len-1;
	}

	while( len-->0 ) {
		int i;
		char c, sc=*s;
		const char __F* oc=oldchr;

		i = 0;
		while( (c=*oc++)!=0 ) {
			if( sc==c ) { /* delimiter is found */
				*s = i>ncend? cncend : newchr[i];
				if( n>0 ) {
					if( --n==0 ) return str;
				}
				break;
			}
			i++;
		}

		s += inc;
	} /* end while( len-->0 ) */ 

	return str;
}

__MODULEEND__



__MODULEBEGIN__( "strcntc.c" )
/* strCountChar */

unsigned __API
strCountChar
	P_B(( str, chr ))
	P_1		const char __F* str
	P_N		const char __F* chr
	P_E
{
	char c;
	unsigned count = 0;

	while( (c=*str++)!=0 ) {
		if( strchr(chr,c) )
			count++;
	}

	return count;
}
__MODULEEND__



__MODULEBEGIN__( "strrc.c" )
/* strReduceChar */


char __F* __API
strReduceChar
	P_B(( str, chr ))
	P_1		char __F* str
	P_N		const char __F* chr
	P_E
{
	char c, __F* src = str, __F* dest = str;

	while( (c=*src++)!=0 ) {
		if( strchr(chr,c)==NULL )
			*dest++= c;
	}
	*dest = 0;

	return str;
}
__MODULEEND__



__MODULEBEGIN__( "strrz.c" )

/* strReduceZero
high byte of low word : levels for exponent
0 = remove 'e' or 'E' if possible
1 = strip + sign and zeros
2 = no modification

low byte of low word : floating point
0 = reduce all zeros and decimal point
1 = reduce all zeros and decimal point is remained
2 = reserve one zero after decimal point */

char __F* __API
strReduceZero
	P_B(( s, level ))
	P_1		char __F* s
	P_N		unsigned level
	P_E
{
	char c;

	int len = strlen(s);
	char __F* exp = strBackwardToken( s, len, "eE", STR_FLAG_FOUND_DELIMITER );
	if( exp ) len = exp - s - 1;

	if( len>0 ) {

		/* backward iteration to find trailing useless zeros */
		char __F* zero=NULL, __F* iter = &s[len-1];

		while( iter!=s ) {

			c = *iter;

			/* beginning position of zero and nonzero digit is found */
			if( !zero && c!='0' )
				zero = iter+1;

			if( c=='.' ) {
				if( zero ) {

					/* zero is followed immediately */
					if( iter+1==zero ) {
						/* evaulate reduction level */
						switch( level&0xff ) {
							case 0:
							len = iter - s;
							break;

							case 1:
							len = iter - s + 1;
							break;

							default:
							if( zero[0]=='0' ) len = zero - s + 1;
							break;
						} /* end switch( level ) */
					}
					else {
						len = zero - s;
					}

				} /* end if( zero ) */
				break;
			} /* end if( c=='.' ) */
		
			iter--;
		} /* end while( iter!=s ) */

		/* remove -0 caused by numerical error */
		if( s[0]=='-' && s[1]=='0' ) {
			if( s[2]==0 || /* -0 */
				( s[2]=='.' &&
				  ( s[3]==0 || /* -0. */
					( s[3]=='0' && s[4]==0 /* -0.0 */ ) ) ) )
				strcpy( s, &s[1] );
		}

	} /* end if( len ) */


	/* dispatch exponent */
	if( exp ) {
		char expchar = *(exp-1);
		level &= 0xff00;
		/* sign and zeros need to be stripped */
		if( level<0x0200 ) {
			_bool minus = 0;
			while( (c=*exp)!=0 ) {
				if( c!='+' && c!='-' && c!='0' ) {
					/* dispatch minus sign */
					if( minus ) *--exp = '-';
					goto l_append_exp;
				}
				/* check minus sign was found */
				if( c=='-' ) minus = 1;
				exp++;
			} /* end while( *exp ) */

			/* exponent is meaningless for level 0 */
			if( level==0 ) goto l_terminate_string;

			/* append zero after 'e' for level 1 */
			exp--;
		} /* end if( (level&0xff00)<0x0200 ) */

		l_append_exp:

		/* append exponent */
		s[len] = expchar;
		strcpy( s+len+1, exp );
	} /* end if( exp ) */
	else {
		l_terminate_string:
		s[len] = 0;
	}

	return s;
}
__MODULEEND__



__MODULEBEGIN__( "strfind.c" )
/* strFind */

int __API
strFind
	P_B(( s, key, flag ))
	P_1		const char __F* s
	P_N		const char __F* key
	P_N		unsigned flag
	P_E
{
	int i = 0;
	while( *s ) {
		if( strCompare(s,key,flag)==0 ) {
		   return i;
		}
		i++;
		s += strlen(s)+1;
	}
	return -1;
}

__MODULEEND__



__MODULEBEGIN__( "soargv.c" )
/* stroptArgv */

const char __F* __API
stroptArgv
	P_B(( stropt ))
	P_1		struct _STROPT __F* stropt
	P_E
{
	const char __F*__F* argv = (const char __F*__F*)stropt->context;
	if( argv && *argv ) {
		stropt->context = (void __F*)(argv+1);
		return argv[1];
	}
	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "solinear.c" )
_API( char __F*, stroptLinear, ( struct _STROPT __F* stropt ) );
__MODULEEND__



__MODULEBEGIN__( "soiob.c" )
char __F* __API
stroptIOTextBuffer
	P_B(( stropt ))
	P_1		struct _STROPT __F* stropt
	P_E
{
//	struct _IOBUFFER __F* iob = &((struct _IOTEXTBUFFER __F*)stropt->context)->iob;
//	struct _IOTEXT __F* iot = &((struct _IOTEXTBUFFER __F*)stropt->context)->iot;
	return NULL;
}
__MODULEEND__



__MODULEBEGIN__( "sogetopt.c" )

/* stroptGetOption

 1. hint to evaluate return value

  for depth 2 switch

  switch( OP_HIGH4(result) ) {
	case ... :
	switch( OP_LOW4(result) ) {
	}
  }

  for depth 1 switch

  switch( result ) {
	case STROPT_FLAG... | ... :
  }

 2. hint for the return value STROPT_IGNORED

  Option indicators can be defined like this : "- @"
 General parsing routines are ignored if indicators appended after the blank
 ( '@' for above sample ) are found.  */



struct _str_option_compare {
	struct _STROPT __F* stropt;
	const char __F* delimiter;
	const char __F* arg;
};



static _bool __API
str_option_find_delimiter
	P_B(( s, ac ))
	P_1		const char __F* s
	P_N		char ac
	P_E
{
	char sc = *s;

	if( sc==' ' ) return 1; /* no explicit delimiter was specified, so allow all */
	do {
		if( sc==ac ) return 1;
		sc = *++s;
	} while( sc!=' ' );

	return 0;
}



static int __API
str_option_compare
	P_B(( p, key ))
	P_1		const void __F* p
	P_N		const void __F* key
	P_E
{
	char ac, sc;
	int index, len;
	const char __F* str, __F* origstr = ((const struct _STROPT_LONG __F*)p)->name;
	struct _STROPT __F* stropt=((struct _str_option_compare __F*)key)->stropt;
	const char __F* arg = ((struct _str_option_compare __F*)key)->arg;

	index = strTokenIndex(origstr," ",0) + 1;
	str = origstr + index;
	len = strlen(str);

	while( (ac=*arg++)!=0 ) {
		sc=*str++;
		if( ac!=sc ) { /* the position where the first different character is found during string comparison */
			_bool delimiter_found = 0;
			const char __F* const delimiter =
				((struct _str_option_compare __F*)key)->delimiter;
			if( delimiter ) {
				/* check the first different character is a delimiter */
				if( strchr(delimiter,ac) ) {
					if( sc!=0 ) return 1;
					delimiter_found = 1;
				}
				else {
					/* assume the given string doesn't have its parameter
					 specifier not to call str_option_find_delimiter */
					index = 0;
				}
			}

			if( index!=0 && str_option_find_delimiter(origstr,ac) ) { /* the found delimiter is specified for this option? */
				if( sc==0 ) goto l_get_param; /* source characters were terminated so get appended parameter */
				return 1;
			}

			return delimiter_found || ac>sc? -1 : 1;
		}
	}
	if( *str!=0 ) return 1;

	if( index!=0 ) { /* the given option string has its parameter */

	  l_get_param:

	  if( ac==0 && stropt->flag&STROPT_FLAG_ALLOW_BLANK ) {
		  OP_ENABLE( stropt->flag, STROPT_FLAG_BLANK_ALLOWED );
		  stropt->append = NULL;
	  }

	  if( ac!=0 ) { /* ac is a delimiter */
		  stropt->append = arg;
		  if( !*arg ) {
			  if( stropt->flag&STROPT_FLAG_ALLOW_BLANK ) {
				  OP_ENABLE( stropt->flag, STROPT_FLAG_BLANK_ALLOWED );
			  }
			  else {
				  stropt->append = NULL;
			  }
		  }
	  }

	} /* end if( index!=-1 ) */

	return 0;
}



static const char __F*
str_option_getarg
	P_B(( stropt ))
	P_1		struct _STROPT __F* stropt
	P_E
{
	const char __F* nextarg = stropt->nextarg;
	if( nextarg ) {
		/* already end of option was reached */
		if( nextarg==(char __F*)-1 ) return NULL;
		stropt->nextarg = NULL;
		return nextarg;
	}
	return stropt->getarg(stropt);
}


_unsigned4
stropt_dispatch_string
	P_B(( soc, lopt, nlopt ))
	P_1		struct _str_option_compare* soc
	P_N		const struct _STROPT_LONG __F* lopt
	P_N		int nlopt
	P_E
{
	unsigned index;
	register struct _STROPT* stropt = soc->stropt;

	lopt = (struct _STROPT_LONG __F*)
		( (stropt->flag&STROPT_FLAG_BINARY_SEARCH) && soc->delimiter ?
			memBinarySearch : memLinearSearch ) /* function */
		( lopt, nlopt, sizeof(struct _STROPT_LONG),
			soc, str_option_compare, &index ); /* parameters */
	if( lopt ) {
		if( stropt->flag&STROPT_FLAG_BLANK_ALLOWED ) {
			OP_DISABLE( stropt->flag, STROPT_FLAG_BLANK_ALLOWED );
			if( !stropt->append ) { /* a delimiter was not found */
				const char __F* arg;
				stropt->append = NULL;
				arg = str_option_getarg(stropt);
				if( arg ) {
					if( str_option_find_delimiter(lopt->name,arg[0]) ) {
						if( arg[1]==0 ) goto l_get_arg;
						stropt->append = &arg[1];
					}
					else {
						/* reserve the next argument */
						stropt->nextarg = arg;
					}
				}
				else {
					stropt->nextarg = (char __F*)-1;
				} /* end if( arg ) */
			} /* end if( !stropt->append ) */
			else {
			  l_get_arg:
			  if( (stropt->append=str_option_getarg(stropt))==NULL ) {
				  /* end of option is reached */
				  stropt->nextarg = (char __F*)-1;
			  }
			}
		} /* end if( stropt->flag&STROPT_FLAG_BLANK_ALLOWED ) */
		return STROPT_STRING | lopt->value;
	}
	return 0;
}


_unsigned4
stropt_dispatch_char
	P_B(( stropt, arg ))
	P_1		struct _STROPT __F* stropt
	P_N		const char __F* arg
	P_E
{
	unsigned flag = stropt->flag;
	char ac = arg[0];

	if( !ac )
		return flag&STROPT_FLAG_INDICATOR_ONLY? 0 : STROPT_UNKNOWN;

	/* if current arg string is terminated, no more combinated characters */
	if( (stropt->append=arg[1]? &arg[1]:NULL)!=NULL && (flag&STROPT_FLAG_COMBINATION) )
		OP_ENABLE( stropt->flag, STROPT_FLAG_COMBINATED );
	else
		OP_DISABLE( stropt->flag, STROPT_FLAG_COMBINATED );

	return ac;
}


_unsigned4 __API
stroptGetOption
	P_B(( stropt, indicator, delimiter, lopt, nlopt ))
	P_1		struct _STROPT __F* stropt
	P_N		const char __F* indicator
	P_N		const char __F* delimiter
	P_N		const struct _STROPT_LONG __F* lopt
	P_N		unsigned nlopt
	P_E
{
	const char __F* arg; /* current argument */

	_bool ignore=0; /* true to ignore parsing */
	_unsigned4 result = STROPT_RESERVED;

	unsigned flag = stropt->flag;

	/* default indicator */
	if( flag&STROPT_FLAG_DEFAULT_INDICATOR )
		indicator = "-";

	/* default delimiter */
	if( (flag&STROPT_FLAG_DEFAULT_DELIMITER) && !delimiter )
		delimiter = "=";

	if( flag&STROPT_FLAG_COMBINATED ) {
		/* parse appended characters as combinated options */
		arg = stropt->append;
		/* not all combinated character options were evaluated */
		goto l_character_option;
	}

	/* stropt->arg==NULL means current process is not in the combinated
	  character option loop so get an argument from array, and
	  ++stropt->argv will ignore the first argument at the first time
	  - it is valid to be applied for the main arguments and to maintain
	  current argv in stropt to be referred outside of this function */
	stropt->arg = arg = str_option_getarg(stropt);
	if( !arg ) return STROPT_TERMINATED;

	/* reset appended string */
	stropt->append = NULL;

	/* indicator can be NULL if indicator is initially NULL and
	  STROPT_FLAG_DEFAULT_INDICATOR was not specified */
	if( indicator ) {
		char i; /* current indicator cache */
		struct _str_option_compare soc;
		soc.stropt = stropt;
		soc.delimiter = delimiter;

		/* indicator iteration loop */
		while( (i=*indicator++)!=0 ) {

			/* check there are some indicators to ignore parsing */
			if( i==' ' ) {
				ignore = 1;
				continue;
			}

			/* arg begins with a indicator ? */
			if( i==arg[0] ) {
				/* translate pointer to keep consistency with the case when
				  OPTION_FLAG_ALWAYS is enabled */
				arg++;
				if( ignore ) {
					stropt->append = arg;
					/* low word is found indicator if parsing was ignored */
					result = STROPT_IGNORED | i;
					goto l_terminate;
				}
				else {
					if( lopt==NULL ) goto l_character_option;

					/* a hint to detect string option was given */
					if( flag&STROPT_FLAG_DOUBLE_INDICATOR ) {
						if( i==arg[0] ) arg++;
						else goto l_character_option;
					}

					l_string_option:
					soc.arg = arg;
					result = stropt_dispatch_string( &soc, lopt, nlopt );
					if( result ) goto l_terminate;

					/* double indicator mode but string was not found */
					if( flag&STROPT_FLAG_DOUBLE_INDICATOR ) {
						stropt->append = arg;
						result = STROPT_UNKNOWN;
						goto l_terminate;
					}
					else {
						l_character_option :
						/* evidently this case is only option indicator is
						 specified and no appended characters */
						result = stropt_dispatch_char( stropt, arg );
						goto l_terminate;
					} /* end if( flag&STROPT_FLAG_DOUBLE_INDICATOR ) */
				} /* end if( ignore ) */
			} /* end if( t==arg[0] ) */
		} /* end while( indicator iteration loop ) */
	} /* end if( indicator ) */

	/* indicator not found :
	  though OPTION_FLAG_ALWAYS can be detected by indicator==NULL for
	  some cases, not for all, i.e. both tar xzvf and tar -xzvf are valid
	  notify this function is evidently returned if parsing is begun
	  so never visit here again */
	if( !indicator || (flag&STROPT_FLAG_ALWAYS) ) {
		/* double indicator is useless if there is no option indicator */
		if( flag&STROPT_FLAG_DOUBLE_INDICATOR ) {
			if( indicator )
				goto l_character_option;
			else
				OP_DISABLE( flag, STROPT_FLAG_DOUBLE_INDICATOR );
		}
		goto l_string_option;
	}

	l_terminate:

	/* always parse only for the first argument : following routine is commonly
	  used and directly jumped here except for the case during parsing
	  combinated character options */
	if( flag&STROPT_FLAG_FIRST )
		OP_DISABLE( stropt->flag, STROPT_FLAG_FIRST|STROPT_FLAG_ALWAYS );

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "sogetstr.c" )
/* stroptGetString */

const char __F* __API
stroptGetString
	P_B(( stropt ))
	P_1		struct _STROPT __F* stropt
	P_E
{
	if( stropt->append )
		return stropt->append;

	/* string was not appended directly after a indicator, so return the next
	  argument and translate the current pointer */

	return stropt->arg = str_option_getarg(stropt);
}
__MODULEEND__



/* end of string.c */

